use std::fmt::{self, Display};

/// A NTSTATUS wrapper that gives information on the value
#[repr(transparent)]
#[derive(Copy, Clone, Debug)]
pub struct NTSTATUS(pub i32);

impl NTSTATUS {
    //
    // https://learn.microsoft.com/en-us/windows-hardware/drivers/kernel/using-ntstatus-values
    //

    // NT_SUCCESS
    pub fn is_success(&self) -> bool {
        let val = bytemuck::cast::<_, u32>(self.0);
        // NT_SUCCESS | NT_INFORMATION
        matches!(val, 0..=0x3FFFFFFF | 0x40000000..=0x7FFFFFFF)
    }

    // NT_INFORMATION
    pub fn is_information(&self) -> bool {
        let val = bytemuck::cast::<_, u32>(self.0);
        matches!(val, 0x40000000..=0x7FFFFFFF)
    }

    // NT_WARNING
    pub fn is_warning(&self) -> bool {
        let val = bytemuck::cast::<_, u32>(self.0);
        matches!(val, 0x80000000..=0xBFFFFFFF)
    }

    // NT_ERROR
    pub fn is_error(&self) -> bool {
        let val = bytemuck::cast::<_, u32>(self.0);
        matches!(val, 0xC0000000..=0xFFFFFFFF)
    }
}

impl From<()> for NTSTATUS {
    fn from(value: ()) -> Self {
        Self(0)
    }
}

impl From<i32> for NTSTATUS {
    fn from(value: i32) -> Self {
        Self(value)
    }
}

impl From<u32> for NTSTATUS {
    fn from(value: u32) -> Self {
        let value = bytemuck::cast(value);
        Self(value)
    }
}

impl From<NTSTATUS> for u32 {
    fn from(value: NTSTATUS) -> Self {
        bytemuck::cast(value.0)
    }
}

impl Display for NTSTATUS {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
        write!(f, "0x{:X}", self.0)
    }
}

#[rustfmt::skip]
impl NTSTATUS {
    pub const STATUS_ABANDONED: NTSTATUS = NTSTATUS(128i32);
    pub const STATUS_ABANDONED_WAIT_0: NTSTATUS = NTSTATUS(128i32);
    pub const STATUS_ABANDONED_WAIT_63: NTSTATUS = NTSTATUS(191i32);
    pub const STATUS_ABANDON_HIBERFILE: NTSTATUS = NTSTATUS(1073741875i32);
    pub const STATUS_ABIOS_INVALID_COMMAND: NTSTATUS = NTSTATUS(-1073741549i32);
    pub const STATUS_ABIOS_INVALID_LID: NTSTATUS = NTSTATUS(-1073741548i32);
    pub const STATUS_ABIOS_INVALID_SELECTOR: NTSTATUS = NTSTATUS(-1073741546i32);
    pub const STATUS_ABIOS_LID_ALREADY_OWNED: NTSTATUS = NTSTATUS(-1073741551i32);
    pub const STATUS_ABIOS_LID_NOT_EXIST: NTSTATUS = NTSTATUS(-1073741552i32);
    pub const STATUS_ABIOS_NOT_LID_OWNER: NTSTATUS = NTSTATUS(-1073741550i32);
    pub const STATUS_ABIOS_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073741553i32);
    pub const STATUS_ABIOS_SELECTOR_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073741547i32);
    pub const STATUS_ACCESS_AUDIT_BY_POLICY: NTSTATUS = NTSTATUS(1073741874i32);
    pub const STATUS_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073741790i32);
    pub const STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT: NTSTATUS = NTSTATUS(-1073740959i32);
    pub const STATUS_ACCESS_DISABLED_BY_POLICY_OTHER: NTSTATUS = NTSTATUS(-1073740956i32);
    pub const STATUS_ACCESS_DISABLED_BY_POLICY_PATH: NTSTATUS = NTSTATUS(-1073740958i32);
    pub const STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER: NTSTATUS = NTSTATUS(-1073740957i32);
    pub const STATUS_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY: NTSTATUS = NTSTATUS(-1073740942i32);
    pub const STATUS_ACCESS_VIOLATION: NTSTATUS = NTSTATUS(-1073741819i32);
    pub const STATUS_ACPI_ACQUIRE_GLOBAL_LOCK: NTSTATUS = NTSTATUS(-1072431086i32);
    pub const STATUS_ACPI_ADDRESS_NOT_MAPPED: NTSTATUS = NTSTATUS(-1072431092i32);
    pub const STATUS_ACPI_ALREADY_INITIALIZED: NTSTATUS = NTSTATUS(-1072431085i32);
    pub const STATUS_ACPI_ASSERT_FAILED: NTSTATUS = NTSTATUS(-1072431101i32);
    pub const STATUS_ACPI_FATAL: NTSTATUS = NTSTATUS(-1072431098i32);
    pub const STATUS_ACPI_HANDLER_COLLISION: NTSTATUS = NTSTATUS(-1072431090i32);
    pub const STATUS_ACPI_INCORRECT_ARGUMENT_COUNT: NTSTATUS = NTSTATUS(-1072431093i32);
    pub const STATUS_ACPI_INVALID_ACCESS_SIZE: NTSTATUS = NTSTATUS(-1072431087i32);
    pub const STATUS_ACPI_INVALID_ARGTYPE: NTSTATUS = NTSTATUS(-1072431096i32);
    pub const STATUS_ACPI_INVALID_ARGUMENT: NTSTATUS = NTSTATUS(-1072431099i32);
    pub const STATUS_ACPI_INVALID_DATA: NTSTATUS = NTSTATUS(-1072431089i32);
    pub const STATUS_ACPI_INVALID_EVENTTYPE: NTSTATUS = NTSTATUS(-1072431091i32);
    pub const STATUS_ACPI_INVALID_INDEX: NTSTATUS = NTSTATUS(-1072431100i32);
    pub const STATUS_ACPI_INVALID_MUTEX_LEVEL: NTSTATUS = NTSTATUS(-1072431083i32);
    pub const STATUS_ACPI_INVALID_OBJTYPE: NTSTATUS = NTSTATUS(-1072431095i32);
    pub const STATUS_ACPI_INVALID_OPCODE: NTSTATUS = NTSTATUS(-1072431103i32);
    pub const STATUS_ACPI_INVALID_REGION: NTSTATUS = NTSTATUS(-1072431088i32);
    pub const STATUS_ACPI_INVALID_SUPERNAME: NTSTATUS = NTSTATUS(-1072431097i32);
    pub const STATUS_ACPI_INVALID_TABLE: NTSTATUS = NTSTATUS(-1072431079i32);
    pub const STATUS_ACPI_INVALID_TARGETTYPE: NTSTATUS = NTSTATUS(-1072431094i32);
    pub const STATUS_ACPI_MUTEX_NOT_OWNED: NTSTATUS = NTSTATUS(-1072431082i32);
    pub const STATUS_ACPI_MUTEX_NOT_OWNER: NTSTATUS = NTSTATUS(-1072431081i32);
    pub const STATUS_ACPI_NOT_INITIALIZED: NTSTATUS = NTSTATUS(-1072431084i32);
    pub const STATUS_ACPI_POWER_REQUEST_FAILED: NTSTATUS = NTSTATUS(-1072431071i32);
    pub const STATUS_ACPI_REG_HANDLER_FAILED: NTSTATUS = NTSTATUS(-1072431072i32);
    pub const STATUS_ACPI_RS_ACCESS: NTSTATUS = NTSTATUS(-1072431080i32);
    pub const STATUS_ACPI_STACK_OVERFLOW: NTSTATUS = NTSTATUS(-1072431102i32);
    pub const STATUS_ADAPTER_HARDWARE_ERROR: NTSTATUS = NTSTATUS(-1073741630i32);
    pub const STATUS_ADDRESS_ALREADY_ASSOCIATED: NTSTATUS = NTSTATUS(-1073741256i32);
    pub const STATUS_ADDRESS_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1073741302i32);
    pub const STATUS_ADDRESS_CLOSED: NTSTATUS = NTSTATUS(-1073741301i32);
    pub const STATUS_ADDRESS_NOT_ASSOCIATED: NTSTATUS = NTSTATUS(-1073741255i32);
    pub const STATUS_ADMINLESS_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073700348i32);
    pub const STATUS_ADVANCED_INSTALLER_FAILED: NTSTATUS = NTSTATUS(-1072365536i32);
    pub const STATUS_AGENTS_EXHAUSTED: NTSTATUS = NTSTATUS(-1073741691i32);
    pub const STATUS_ALERTED: NTSTATUS = NTSTATUS(257i32);
    pub const STATUS_ALIAS_EXISTS: NTSTATUS = NTSTATUS(-1073741484i32);
    pub const STATUS_ALLOCATE_BUCKET: NTSTATUS = NTSTATUS(-1073741265i32);
    pub const STATUS_ALLOTTED_SPACE_EXCEEDED: NTSTATUS = NTSTATUS(-1073741671i32);
    pub const STATUS_ALL_SIDS_FILTERED: NTSTATUS = NTSTATUS(-1073740962i32);
    pub const STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073740798i32);
    pub const STATUS_ALPC_CHECK_COMPLETION_LIST: NTSTATUS = NTSTATUS(1073741872i32);
    pub const STATUS_ALREADY_COMMITTED: NTSTATUS = NTSTATUS(-1073741791i32);
    pub const STATUS_ALREADY_COMPLETE: NTSTATUS = NTSTATUS(255i32);
    pub const STATUS_ALREADY_DISCONNECTED: NTSTATUS = NTSTATUS(-2147483611i32);
    pub const STATUS_ALREADY_HAS_STREAM_ID: NTSTATUS = NTSTATUS(-1073740530i32);
    pub const STATUS_ALREADY_INITIALIZED: NTSTATUS = NTSTATUS(-1073740528i32);
    pub const STATUS_ALREADY_REGISTERED: NTSTATUS = NTSTATUS(-1073740008i32);
    pub const STATUS_ALREADY_WIN32: NTSTATUS = NTSTATUS(1073741851i32);
    pub const STATUS_AMBIGUOUS_SYSTEM_DEVICE: NTSTATUS = NTSTATUS(-1073740719i32);
    pub const STATUS_APC_RETURNED_WHILE_IMPERSONATING: NTSTATUS = NTSTATUS(-1073740015i32);
    pub const STATUS_APISET_NOT_HOSTED: NTSTATUS = NTSTATUS(-1073740671i32);
    pub const STATUS_APISET_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073740670i32);
    pub const STATUS_APPEXEC_APP_COMPAT_BLOCK: NTSTATUS = NTSTATUS(-1058275320i32);
    pub const STATUS_APPEXEC_CALLER_WAIT_TIMEOUT: NTSTATUS = NTSTATUS(-1058275319i32);
    pub const STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_LICENSING: NTSTATUS = NTSTATUS(-1058275317i32);
    pub const STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_RESOURCES: NTSTATUS = NTSTATUS(-1058275316i32);
    pub const STATUS_APPEXEC_CALLER_WAIT_TIMEOUT_TERMINATION: NTSTATUS = NTSTATUS(-1058275318i32);
    pub const STATUS_APPEXEC_CONDITION_NOT_SATISFIED: NTSTATUS = NTSTATUS(-1058275328i32);
    pub const STATUS_APPEXEC_HANDLE_INVALIDATED: NTSTATUS = NTSTATUS(-1058275327i32);
    pub const STATUS_APPEXEC_HOST_ID_MISMATCH: NTSTATUS = NTSTATUS(-1058275322i32);
    pub const STATUS_APPEXEC_INVALID_HOST_GENERATION: NTSTATUS = NTSTATUS(-1058275326i32);
    pub const STATUS_APPEXEC_INVALID_HOST_STATE: NTSTATUS = NTSTATUS(-1058275324i32);
    pub const STATUS_APPEXEC_NO_DONOR: NTSTATUS = NTSTATUS(-1058275323i32);
    pub const STATUS_APPEXEC_UNEXPECTED_PROCESS_REGISTRATION: NTSTATUS = NTSTATUS(-1058275325i32);
    pub const STATUS_APPEXEC_UNKNOWN_USER: NTSTATUS = NTSTATUS(-1058275321i32);
    pub const STATUS_APPHELP_BLOCK: NTSTATUS = NTSTATUS(-1073740963i32);
    pub const STATUS_APPX_FILE_NOT_ENCRYPTED: NTSTATUS = NTSTATUS(-1073740634i32);
    pub const STATUS_APPX_INTEGRITY_FAILURE_CLR_NGEN: NTSTATUS = NTSTATUS(-1073740673i32);
    pub const STATUS_APP_DATA_CORRUPT: NTSTATUS = NTSTATUS(-1073700221i32);
    pub const STATUS_APP_DATA_EXPIRED: NTSTATUS = NTSTATUS(-1073700222i32);
    pub const STATUS_APP_DATA_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073700220i32);
    pub const STATUS_APP_DATA_NOT_FOUND: NTSTATUS = NTSTATUS(-1073700223i32);
    pub const STATUS_APP_DATA_REBOOT_REQUIRED: NTSTATUS = NTSTATUS(-1073700219i32);
    pub const STATUS_APP_INIT_FAILURE: NTSTATUS = NTSTATUS(-1073741499i32);
    pub const STATUS_ARBITRATION_UNHANDLED: NTSTATUS = NTSTATUS(1073741862i32);
    pub const STATUS_ARRAY_BOUNDS_EXCEEDED: NTSTATUS = NTSTATUS(-1073741684i32);
    pub const STATUS_ASSERTION_FAILURE: NTSTATUS = NTSTATUS(-1073740768i32);
    pub const STATUS_ATTACHED_EXECUTABLE_MEMORY_WRITE: NTSTATUS = NTSTATUS(-1073739995i32);
    pub const STATUS_ATTRIBUTE_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073740532i32);
    pub const STATUS_AUDIO_ENGINE_NODE_NOT_FOUND: NTSTATUS = NTSTATUS(-1069285375i32);
    pub const STATUS_AUDITING_DISABLED: NTSTATUS = NTSTATUS(-1073740970i32);
    pub const STATUS_AUDIT_FAILED: NTSTATUS = NTSTATUS(-1073741244i32);
    pub const STATUS_AUTHIP_FAILURE: NTSTATUS = NTSTATUS(-1073700730i32);
    pub const STATUS_AUTH_TAG_MISMATCH: NTSTATUS = NTSTATUS(-1073700862i32);
    pub const STATUS_BACKUP_CONTROLLER: NTSTATUS = NTSTATUS(-1073741433i32);
    pub const STATUS_BAD_BINDINGS: NTSTATUS = NTSTATUS(-1073740965i32);
    pub const STATUS_BAD_CLUSTERS: NTSTATUS = NTSTATUS(-1073739771i32);
    pub const STATUS_BAD_COMPRESSION_BUFFER: NTSTATUS = NTSTATUS(-1073741246i32);
    pub const STATUS_BAD_CURRENT_DIRECTORY: NTSTATUS = NTSTATUS(1073741831i32);
    pub const STATUS_BAD_DATA: NTSTATUS = NTSTATUS(-1073739509i32);
    pub const STATUS_BAD_DESCRIPTOR_FORMAT: NTSTATUS = NTSTATUS(-1073741593i32);
    pub const STATUS_BAD_DEVICE_TYPE: NTSTATUS = NTSTATUS(-1073741621i32);
    pub const STATUS_BAD_DLL_ENTRYPOINT: NTSTATUS = NTSTATUS(-1073741231i32);
    pub const STATUS_BAD_FILE_TYPE: NTSTATUS = NTSTATUS(-1073739517i32);
    pub const STATUS_BAD_FUNCTION_TABLE: NTSTATUS = NTSTATUS(-1073741569i32);
    pub const STATUS_BAD_IMPERSONATION_LEVEL: NTSTATUS = NTSTATUS(-1073741659i32);
    pub const STATUS_BAD_INHERITANCE_ACL: NTSTATUS = NTSTATUS(-1073741699i32);
    pub const STATUS_BAD_INITIAL_PC: NTSTATUS = NTSTATUS(-1073741814i32);
    pub const STATUS_BAD_INITIAL_STACK: NTSTATUS = NTSTATUS(-1073741815i32);
    pub const STATUS_BAD_KEY: NTSTATUS = NTSTATUS(-1073739510i32);
    pub const STATUS_BAD_LOGON_SESSION_STATE: NTSTATUS = NTSTATUS(-1073741564i32);
    pub const STATUS_BAD_MASTER_BOOT_RECORD: NTSTATUS = NTSTATUS(-1073741655i32);
    pub const STATUS_BAD_MCFG_TABLE: NTSTATUS = NTSTATUS(-1073739512i32);
    pub const STATUS_BAD_NETWORK_NAME: NTSTATUS = NTSTATUS(-1073741620i32);
    pub const STATUS_BAD_NETWORK_PATH: NTSTATUS = NTSTATUS(-1073741634i32);
    pub const STATUS_BAD_REMOTE_ADAPTER: NTSTATUS = NTSTATUS(-1073741627i32);
    pub const STATUS_BAD_SERVICE_ENTRYPOINT: NTSTATUS = NTSTATUS(-1073741230i32);
    pub const STATUS_BAD_STACK: NTSTATUS = NTSTATUS(-1073741784i32);
    pub const STATUS_BAD_TOKEN_TYPE: NTSTATUS = NTSTATUS(-1073741656i32);
    pub const STATUS_BAD_VALIDATION_CLASS: NTSTATUS = NTSTATUS(-1073741657i32);
    pub const STATUS_BAD_WORKING_SET_LIMIT: NTSTATUS = NTSTATUS(-1073741748i32);
    pub const STATUS_BCD_NOT_ALL_ENTRIES_IMPORTED: NTSTATUS = NTSTATUS(-2143748095i32);
    pub const STATUS_BCD_NOT_ALL_ENTRIES_SYNCHRONIZED: NTSTATUS = NTSTATUS(-2143748093i32);
    pub const STATUS_BCD_TOO_MANY_ELEMENTS: NTSTATUS = NTSTATUS(-1070006270i32);
    pub const STATUS_BEGINNING_OF_MEDIA: NTSTATUS = NTSTATUS(-2147483617i32);
    pub const STATUS_BEYOND_VDL: NTSTATUS = NTSTATUS(-1073740750i32);
    pub const STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT: NTSTATUS = NTSTATUS(-1073741458i32);
    pub const STATUS_BIZRULES_NOT_ENABLED: NTSTATUS = NTSTATUS(1073741876i32);
    pub const STATUS_BLOCKED_BY_PARENTAL_CONTROLS: NTSTATUS = NTSTATUS(-1073740664i32);
    pub const STATUS_BLOCK_SHARED: NTSTATUS = NTSTATUS(-1073739499i32);
    pub const STATUS_BLOCK_SOURCE_WEAK_REFERENCE_INVALID: NTSTATUS = NTSTATUS(-1073739501i32);
    pub const STATUS_BLOCK_TARGET_WEAK_REFERENCE_INVALID: NTSTATUS = NTSTATUS(-1073739500i32);
    pub const STATUS_BLOCK_TOO_MANY_REFERENCES: NTSTATUS = NTSTATUS(-1073740660i32);
    pub const STATUS_BLOCK_WEAK_REFERENCE_INVALID: NTSTATUS = NTSTATUS(-1073739502i32);
    pub const STATUS_BREAKPOINT: NTSTATUS = NTSTATUS(-2147483645i32);
    pub const STATUS_BTH_ATT_ATTRIBUTE_NOT_FOUND: NTSTATUS = NTSTATUS(-1069416438i32);
    pub const STATUS_BTH_ATT_ATTRIBUTE_NOT_LONG: NTSTATUS = NTSTATUS(-1069416437i32);
    pub const STATUS_BTH_ATT_INSUFFICIENT_AUTHENTICATION: NTSTATUS = NTSTATUS(-1069416443i32);
    pub const STATUS_BTH_ATT_INSUFFICIENT_AUTHORIZATION: NTSTATUS = NTSTATUS(-1069416440i32);
    pub const STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION: NTSTATUS = NTSTATUS(-1069416433i32);
    pub const STATUS_BTH_ATT_INSUFFICIENT_ENCRYPTION_KEY_SIZE: NTSTATUS = NTSTATUS(-1069416436i32);
    pub const STATUS_BTH_ATT_INSUFFICIENT_RESOURCES: NTSTATUS = NTSTATUS(-1069416431i32);
    pub const STATUS_BTH_ATT_INVALID_ATTRIBUTE_VALUE_LENGTH: NTSTATUS = NTSTATUS(-1069416435i32);
    pub const STATUS_BTH_ATT_INVALID_HANDLE: NTSTATUS = NTSTATUS(-1069416447i32);
    pub const STATUS_BTH_ATT_INVALID_OFFSET: NTSTATUS = NTSTATUS(-1069416441i32);
    pub const STATUS_BTH_ATT_INVALID_PDU: NTSTATUS = NTSTATUS(-1069416444i32);
    pub const STATUS_BTH_ATT_PREPARE_QUEUE_FULL: NTSTATUS = NTSTATUS(-1069416439i32);
    pub const STATUS_BTH_ATT_READ_NOT_PERMITTED: NTSTATUS = NTSTATUS(-1069416446i32);
    pub const STATUS_BTH_ATT_REQUEST_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1069416442i32);
    pub const STATUS_BTH_ATT_UNKNOWN_ERROR: NTSTATUS = NTSTATUS(-1069412352i32);
    pub const STATUS_BTH_ATT_UNLIKELY: NTSTATUS = NTSTATUS(-1069416434i32);
    pub const STATUS_BTH_ATT_UNSUPPORTED_GROUP_TYPE: NTSTATUS = NTSTATUS(-1069416432i32);
    pub const STATUS_BTH_ATT_WRITE_NOT_PERMITTED: NTSTATUS = NTSTATUS(-1069416445i32);
    pub const STATUS_BUFFER_ALL_ZEROS: NTSTATUS = NTSTATUS(279i32);
    pub const STATUS_BUFFER_OVERFLOW: NTSTATUS = NTSTATUS(-2147483643i32);
    pub const STATUS_BUFFER_TOO_SMALL: NTSTATUS = NTSTATUS(-1073741789i32);
    pub const STATUS_BUS_RESET: NTSTATUS = NTSTATUS(-2147483619i32);
    pub const STATUS_BYPASSIO_FLT_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740590i32);
    pub const STATUS_CACHE_PAGE_LOCKED: NTSTATUS = NTSTATUS(277i32);
    pub const STATUS_CALLBACK_BYPASS: NTSTATUS = NTSTATUS(-1073740541i32);
    pub const STATUS_CALLBACK_INVOKE_INLINE: NTSTATUS = NTSTATUS(-1073740661i32);
    pub const STATUS_CALLBACK_POP_STACK: NTSTATUS = NTSTATUS(-1073740765i32);
    pub const STATUS_CALLBACK_RETURNED_LANG: NTSTATUS = NTSTATUS(-1073740001i32);
    pub const STATUS_CALLBACK_RETURNED_LDR_LOCK: NTSTATUS = NTSTATUS(-1073740002i32);
    pub const STATUS_CALLBACK_RETURNED_PRI_BACK: NTSTATUS = NTSTATUS(-1073740000i32);
    pub const STATUS_CALLBACK_RETURNED_THREAD_AFFINITY: NTSTATUS = NTSTATUS(-1073739999i32);
    pub const STATUS_CALLBACK_RETURNED_THREAD_PRIORITY: NTSTATUS = NTSTATUS(-1073740005i32);
    pub const STATUS_CALLBACK_RETURNED_TRANSACTION: NTSTATUS = NTSTATUS(-1073740003i32);
    pub const STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING: NTSTATUS = NTSTATUS(-1073740016i32);
    pub const STATUS_CANCELLED: NTSTATUS = NTSTATUS(-1073741536i32);
    pub const STATUS_CANNOT_ABORT_TRANSACTIONS: NTSTATUS = NTSTATUS(-1072103347i32);
    pub const STATUS_CANNOT_ACCEPT_TRANSACTED_WORK: NTSTATUS = NTSTATUS(-1072103348i32);
    pub const STATUS_CANNOT_BREAK_OPLOCK: NTSTATUS = NTSTATUS(-1073739511i32);
    pub const STATUS_CANNOT_DELETE: NTSTATUS = NTSTATUS(-1073741535i32);
    pub const STATUS_CANNOT_EXECUTE_FILE_IN_TRANSACTION: NTSTATUS = NTSTATUS(-1072103356i32);
    pub const STATUS_CANNOT_GRANT_REQUESTED_OPLOCK: NTSTATUS = NTSTATUS(-2147483602i32);
    pub const STATUS_CANNOT_IMPERSONATE: NTSTATUS = NTSTATUS(-1073741555i32);
    pub const STATUS_CANNOT_LOAD_REGISTRY_FILE: NTSTATUS = NTSTATUS(-1073741288i32);
    pub const STATUS_CANNOT_MAKE: NTSTATUS = NTSTATUS(-1073741078i32);
    pub const STATUS_CANNOT_SWITCH_RUNLEVEL: NTSTATUS = NTSTATUS(-1073700543i32);
    pub const STATUS_CANT_ACCESS_DOMAIN_INFO: NTSTATUS = NTSTATUS(-1073741606i32);
    pub const STATUS_CANT_BREAK_TRANSACTIONAL_DEPENDENCY: NTSTATUS = NTSTATUS(-1072103369i32);
    pub const STATUS_CANT_CLEAR_ENCRYPTION_FLAG: NTSTATUS = NTSTATUS(-1073740616i32);
    pub const STATUS_CANT_CREATE_MORE_STREAM_MINIVERSIONS: NTSTATUS = NTSTATUS(-1072103386i32);
    pub const STATUS_CANT_CROSS_RM_BOUNDARY: NTSTATUS = NTSTATUS(-1072103368i32);
    pub const STATUS_CANT_DISABLE_MANDATORY: NTSTATUS = NTSTATUS(-1073741731i32);
    pub const STATUS_CANT_ENABLE_DENY_ONLY: NTSTATUS = NTSTATUS(-1073741133i32);
    pub const STATUS_CANT_OPEN_ANONYMOUS: NTSTATUS = NTSTATUS(-1073741658i32);
    pub const STATUS_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT: NTSTATUS = NTSTATUS(-1072103387i32);
    pub const STATUS_CANT_RECOVER_WITH_HANDLE_OPEN: NTSTATUS = NTSTATUS(-2145845199i32);
    pub const STATUS_CANT_TERMINATE_SELF: NTSTATUS = NTSTATUS(-1073741605i32);
    pub const STATUS_CANT_WAIT: NTSTATUS = NTSTATUS(-1073741608i32);
    pub const STATUS_CARDBUS_NOT_SUPPORTED: NTSTATUS = NTSTATUS(1073741863i32);
    pub const STATUS_CASE_DIFFERING_NAMES_IN_DIR: NTSTATUS = NTSTATUS(-1073740621i32);
    pub const STATUS_CASE_SENSITIVE_PATH: NTSTATUS = NTSTATUS(-1073740614i32);
    pub const STATUS_CC_NEEDS_CALLBACK_SECTION_DRAIN: NTSTATUS = NTSTATUS(-1073700856i32);
    pub const STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE: NTSTATUS = NTSTATUS(-1073740012i32);
    pub const STATUS_CERTIFICATE_VALIDATION_PREFERENCE_CONFLICT: NTSTATUS = NTSTATUS(-1073741387i32);
    pub const STATUS_CHECKING_FILE_SYSTEM: NTSTATUS = NTSTATUS(1073741844i32);
    pub const STATUS_CHECKOUT_REQUIRED: NTSTATUS = NTSTATUS(-1073739518i32);
    pub const STATUS_CHILD_MUST_BE_VOLATILE: NTSTATUS = NTSTATUS(-1073741439i32);
    pub const STATUS_CHILD_PROCESS_BLOCKED: NTSTATUS = NTSTATUS(-1073740643i32);
    pub const STATUS_CIMFS_IMAGE_CORRUPT: NTSTATUS = NTSTATUS(-1073692671i32);
    pub const STATUS_CIMFS_IMAGE_VERSION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073692670i32);
    pub const STATUS_CLEANER_CARTRIDGE_INSTALLED: NTSTATUS = NTSTATUS(-2147483609i32);
    pub const STATUS_CLIENT_SERVER_PARAMETERS_INVALID: NTSTATUS = NTSTATUS(-1073741277i32);
    pub const STATUS_CLIP_DEVICE_LICENSE_MISSING: NTSTATUS = NTSTATUS(-1058406397i32);
    pub const STATUS_CLIP_KEYHOLDER_LICENSE_MISSING_OR_INVALID: NTSTATUS = NTSTATUS(-1058406395i32);
    pub const STATUS_CLIP_LICENSE_DEVICE_ID_MISMATCH: NTSTATUS = NTSTATUS(-1058406390i32);
    pub const STATUS_CLIP_LICENSE_EXPIRED: NTSTATUS = NTSTATUS(-1058406394i32);
    pub const STATUS_CLIP_LICENSE_HARDWARE_ID_OUT_OF_TOLERANCE: NTSTATUS = NTSTATUS(-1058406391i32);
    pub const STATUS_CLIP_LICENSE_INVALID_SIGNATURE: NTSTATUS = NTSTATUS(-1058406396i32);
    pub const STATUS_CLIP_LICENSE_NOT_FOUND: NTSTATUS = NTSTATUS(-1058406398i32);
    pub const STATUS_CLIP_LICENSE_NOT_SIGNED: NTSTATUS = NTSTATUS(-1058406392i32);
    pub const STATUS_CLIP_LICENSE_SIGNED_BY_UNKNOWN_SOURCE: NTSTATUS = NTSTATUS(-1058406393i32);
    pub const STATUS_CLOUD_FILE_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073688808i32);
    pub const STATUS_CLOUD_FILE_ALREADY_CONNECTED: NTSTATUS = NTSTATUS(-1073688823i32);
    pub const STATUS_CLOUD_FILE_AUTHENTICATION_FAILED: NTSTATUS = NTSTATUS(-1073688817i32);
    pub const STATUS_CLOUD_FILE_CONNECTED_PROVIDER_ONLY: NTSTATUS = NTSTATUS(-1073688819i32);
    pub const STATUS_CLOUD_FILE_DEHYDRATION_DISALLOWED: NTSTATUS = NTSTATUS(-1073688800i32);
    pub const STATUS_CLOUD_FILE_INCOMPATIBLE_HARDLINKS: NTSTATUS = NTSTATUS(-1073688807i32);
    pub const STATUS_CLOUD_FILE_INSUFFICIENT_RESOURCES: NTSTATUS = NTSTATUS(-1073688816i32);
    pub const STATUS_CLOUD_FILE_INVALID_REQUEST: NTSTATUS = NTSTATUS(-1073688821i32);
    pub const STATUS_CLOUD_FILE_IN_USE: NTSTATUS = NTSTATUS(-1073688812i32);
    pub const STATUS_CLOUD_FILE_METADATA_CORRUPT: NTSTATUS = NTSTATUS(-1073688830i32);
    pub const STATUS_CLOUD_FILE_METADATA_TOO_LARGE: NTSTATUS = NTSTATUS(-1073688829i32);
    pub const STATUS_CLOUD_FILE_NETWORK_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073688815i32);
    pub const STATUS_CLOUD_FILE_NOT_IN_SYNC: NTSTATUS = NTSTATUS(-1073688824i32);
    pub const STATUS_CLOUD_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073688822i32);
    pub const STATUS_CLOUD_FILE_NOT_UNDER_SYNC_ROOT: NTSTATUS = NTSTATUS(-1073688813i32);
    pub const STATUS_CLOUD_FILE_PINNED: NTSTATUS = NTSTATUS(-1073688811i32);
    pub const STATUS_CLOUD_FILE_PROPERTY_BLOB_CHECKSUM_MISMATCH: NTSTATUS = NTSTATUS(-2147430656i32);
    pub const STATUS_CLOUD_FILE_PROPERTY_BLOB_TOO_LARGE: NTSTATUS = NTSTATUS(-2147430652i32);
    pub const STATUS_CLOUD_FILE_PROPERTY_CORRUPT: NTSTATUS = NTSTATUS(-1073688809i32);
    pub const STATUS_CLOUD_FILE_PROPERTY_LOCK_CONFLICT: NTSTATUS = NTSTATUS(-1073688806i32);
    pub const STATUS_CLOUD_FILE_PROPERTY_VERSION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073688826i32);
    pub const STATUS_CLOUD_FILE_PROVIDER_NOT_RUNNING: NTSTATUS = NTSTATUS(-1073688831i32);
    pub const STATUS_CLOUD_FILE_PROVIDER_TERMINATED: NTSTATUS = NTSTATUS(-1073688803i32);
    pub const STATUS_CLOUD_FILE_READ_ONLY_VOLUME: NTSTATUS = NTSTATUS(-1073688820i32);
    pub const STATUS_CLOUD_FILE_REQUEST_ABORTED: NTSTATUS = NTSTATUS(-1073688810i32);
    pub const STATUS_CLOUD_FILE_REQUEST_CANCELED: NTSTATUS = NTSTATUS(-1073688805i32);
    pub const STATUS_CLOUD_FILE_REQUEST_TIMEOUT: NTSTATUS = NTSTATUS(-1073688801i32);
    pub const STATUS_CLOUD_FILE_SYNC_ROOT_METADATA_CORRUPT: NTSTATUS = NTSTATUS(-1073688832i32);
    pub const STATUS_CLOUD_FILE_TOO_MANY_PROPERTY_BLOBS: NTSTATUS = NTSTATUS(-2147430651i32);
    pub const STATUS_CLOUD_FILE_UNSUCCESSFUL: NTSTATUS = NTSTATUS(-1073688814i32);
    pub const STATUS_CLOUD_FILE_US_MESSAGE_TIMEOUT: NTSTATUS = NTSTATUS(-1073688799i32);
    pub const STATUS_CLOUD_FILE_VALIDATION_FAILED: NTSTATUS = NTSTATUS(-1073688818i32);
    pub const STATUS_CLUSTER_CAM_TICKET_REPLAY_DETECTED: NTSTATUS = NTSTATUS(-1072496591i32);
    pub const STATUS_CLUSTER_CSV_AUTO_PAUSE_ERROR: NTSTATUS = NTSTATUS(-1072496607i32);
    pub const STATUS_CLUSTER_CSV_INVALID_HANDLE: NTSTATUS = NTSTATUS(-1072496599i32);
    pub const STATUS_CLUSTER_CSV_NOT_REDIRECTED: NTSTATUS = NTSTATUS(-1072496605i32);
    pub const STATUS_CLUSTER_CSV_NO_SNAPSHOTS: NTSTATUS = NTSTATUS(-1072496601i32);
    pub const STATUS_CLUSTER_CSV_READ_OPLOCK_BREAK_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072496608i32);
    pub const STATUS_CLUSTER_CSV_REDIRECTED: NTSTATUS = NTSTATUS(-1072496606i32);
    pub const STATUS_CLUSTER_CSV_SNAPSHOT_CREATION_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072496603i32);
    pub const STATUS_CLUSTER_CSV_SUPPORTED_ONLY_ON_COORDINATOR: NTSTATUS = NTSTATUS(-1072496592i32);
    pub const STATUS_CLUSTER_CSV_VOLUME_DRAINING: NTSTATUS = NTSTATUS(-1072496604i32);
    pub const STATUS_CLUSTER_CSV_VOLUME_DRAINING_SUCCEEDED_DOWNLEVEL: NTSTATUS = NTSTATUS(-1072496602i32);
    pub const STATUS_CLUSTER_CSV_VOLUME_NOT_LOCAL: NTSTATUS = NTSTATUS(-1072496615i32);
    pub const STATUS_CLUSTER_INVALID_NETWORK: NTSTATUS = NTSTATUS(-1072496624i32);
    pub const STATUS_CLUSTER_INVALID_NETWORK_PROVIDER: NTSTATUS = NTSTATUS(-1072496629i32);
    pub const STATUS_CLUSTER_INVALID_NODE: NTSTATUS = NTSTATUS(-1072496639i32);
    pub const STATUS_CLUSTER_INVALID_REQUEST: NTSTATUS = NTSTATUS(-1072496630i32);
    pub const STATUS_CLUSTER_JOIN_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072496637i32);
    pub const STATUS_CLUSTER_JOIN_NOT_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072496625i32);
    pub const STATUS_CLUSTER_LOCAL_NODE_NOT_FOUND: NTSTATUS = NTSTATUS(-1072496635i32);
    pub const STATUS_CLUSTER_NETINTERFACE_EXISTS: NTSTATUS = NTSTATUS(-1072496632i32);
    pub const STATUS_CLUSTER_NETINTERFACE_NOT_FOUND: NTSTATUS = NTSTATUS(-1072496631i32);
    pub const STATUS_CLUSTER_NETWORK_ALREADY_OFFLINE: NTSTATUS = NTSTATUS(-2146238460i32);
    pub const STATUS_CLUSTER_NETWORK_ALREADY_ONLINE: NTSTATUS = NTSTATUS(-2146238461i32);
    pub const STATUS_CLUSTER_NETWORK_EXISTS: NTSTATUS = NTSTATUS(-1072496634i32);
    pub const STATUS_CLUSTER_NETWORK_NOT_FOUND: NTSTATUS = NTSTATUS(-1072496633i32);
    pub const STATUS_CLUSTER_NETWORK_NOT_INTERNAL: NTSTATUS = NTSTATUS(-1072496618i32);
    pub const STATUS_CLUSTER_NODE_ALREADY_DOWN: NTSTATUS = NTSTATUS(-2146238462i32);
    pub const STATUS_CLUSTER_NODE_ALREADY_MEMBER: NTSTATUS = NTSTATUS(-2146238459i32);
    pub const STATUS_CLUSTER_NODE_ALREADY_UP: NTSTATUS = NTSTATUS(-2146238463i32);
    pub const STATUS_CLUSTER_NODE_DOWN: NTSTATUS = NTSTATUS(-1072496628i32);
    pub const STATUS_CLUSTER_NODE_EXISTS: NTSTATUS = NTSTATUS(-1072496638i32);
    pub const STATUS_CLUSTER_NODE_NOT_FOUND: NTSTATUS = NTSTATUS(-1072496636i32);
    pub const STATUS_CLUSTER_NODE_NOT_MEMBER: NTSTATUS = NTSTATUS(-1072496626i32);
    pub const STATUS_CLUSTER_NODE_NOT_PAUSED: NTSTATUS = NTSTATUS(-1072496620i32);
    pub const STATUS_CLUSTER_NODE_PAUSED: NTSTATUS = NTSTATUS(-1072496621i32);
    pub const STATUS_CLUSTER_NODE_UNREACHABLE: NTSTATUS = NTSTATUS(-1072496627i32);
    pub const STATUS_CLUSTER_NODE_UP: NTSTATUS = NTSTATUS(-1072496622i32);
    pub const STATUS_CLUSTER_NON_CSV_PATH: NTSTATUS = NTSTATUS(-1072496616i32);
    pub const STATUS_CLUSTER_NO_NET_ADAPTERS: NTSTATUS = NTSTATUS(-1072496623i32);
    pub const STATUS_CLUSTER_NO_SECURITY_CONTEXT: NTSTATUS = NTSTATUS(-1072496619i32);
    pub const STATUS_CLUSTER_POISONED: NTSTATUS = NTSTATUS(-1072496617i32);
    pub const STATUS_COMMITMENT_LIMIT: NTSTATUS = NTSTATUS(-1073741523i32);
    pub const STATUS_COMMITMENT_MINIMUM: NTSTATUS = NTSTATUS(-1073741112i32);
    pub const STATUS_COMPRESSED_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740677i32);
    pub const STATUS_COMPRESSION_DISABLED: NTSTATUS = NTSTATUS(-1073740762i32);
    pub const STATUS_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION: NTSTATUS = NTSTATUS(-1072103338i32);
    pub const STATUS_COMPRESSION_NOT_BENEFICIAL: NTSTATUS = NTSTATUS(-1073740689i32);
    pub const STATUS_CONFLICTING_ADDRESSES: NTSTATUS = NTSTATUS(-1073741800i32);
    pub const STATUS_CONNECTION_ABORTED: NTSTATUS = NTSTATUS(-1073741247i32);
    pub const STATUS_CONNECTION_ACTIVE: NTSTATUS = NTSTATUS(-1073741253i32);
    pub const STATUS_CONNECTION_COUNT_LIMIT: NTSTATUS = NTSTATUS(-1073741242i32);
    pub const STATUS_CONNECTION_DISCONNECTED: NTSTATUS = NTSTATUS(-1073741300i32);
    pub const STATUS_CONNECTION_INVALID: NTSTATUS = NTSTATUS(-1073741254i32);
    pub const STATUS_CONNECTION_IN_USE: NTSTATUS = NTSTATUS(-1073741560i32);
    pub const STATUS_CONNECTION_REFUSED: NTSTATUS = NTSTATUS(-1073741258i32);
    pub const STATUS_CONNECTION_RESET: NTSTATUS = NTSTATUS(-1073741299i32);
    pub const STATUS_CONTAINER_ASSIGNED: NTSTATUS = NTSTATUS(-1073740536i32);
    pub const STATUS_CONTENT_BLOCKED: NTSTATUS = NTSTATUS(-1073739772i32);
    pub const STATUS_CONTEXT_MISMATCH: NTSTATUS = NTSTATUS(-1073740007i32);
    pub const STATUS_CONTEXT_STOWED_EXCEPTION: NTSTATUS = NTSTATUS(-1073741188i32);
    pub const STATUS_CONTROL_C_EXIT: NTSTATUS = NTSTATUS(-1073741510i32);
    pub const STATUS_CONTROL_STACK_VIOLATION: NTSTATUS = NTSTATUS(-1073741390i32);
    pub const STATUS_CONVERT_TO_LARGE: NTSTATUS = NTSTATUS(-1073741268i32);
    pub const STATUS_COPY_PROTECTION_FAILURE: NTSTATUS = NTSTATUS(-1073741051i32);
    pub const STATUS_CORRUPT_LOG_CLEARED: NTSTATUS = NTSTATUS(-1073739763i32);
    pub const STATUS_CORRUPT_LOG_CORRUPTED: NTSTATUS = NTSTATUS(-1073739766i32);
    pub const STATUS_CORRUPT_LOG_DELETED_FULL: NTSTATUS = NTSTATUS(-1073739764i32);
    pub const STATUS_CORRUPT_LOG_OVERFULL: NTSTATUS = NTSTATUS(-1073739767i32);
    pub const STATUS_CORRUPT_LOG_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073739765i32);
    pub const STATUS_CORRUPT_LOG_UPLEVEL_RECORDS: NTSTATUS = NTSTATUS(-1073739759i32);
    pub const STATUS_CORRUPT_SYSTEM_FILE: NTSTATUS = NTSTATUS(-1073741116i32);
    pub const STATUS_COULD_NOT_INTERPRET: NTSTATUS = NTSTATUS(-1073741639i32);
    pub const STATUS_COULD_NOT_RESIZE_LOG: NTSTATUS = NTSTATUS(-2145845239i32);
    pub const STATUS_CPU_SET_INVALID: NTSTATUS = NTSTATUS(-1073741393i32);
    pub const STATUS_CRASH_DUMP: NTSTATUS = NTSTATUS(278i32);
    pub const STATUS_CRC_ERROR: NTSTATUS = NTSTATUS(-1073741761i32);
    pub const STATUS_CRED_REQUIRES_CONFIRMATION: NTSTATUS = NTSTATUS(-1073740736i32);
    pub const STATUS_CRM_PROTOCOL_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1072103409i32);
    pub const STATUS_CRM_PROTOCOL_NOT_FOUND: NTSTATUS = NTSTATUS(-1072103407i32);
    pub const STATUS_CROSSREALM_DELEGATION_FAILURE: NTSTATUS = NTSTATUS(-1073740789i32);
    pub const STATUS_CROSS_PARTITION_VIOLATION: NTSTATUS = NTSTATUS(-1073740277i32);
    pub const STATUS_CRYPTO_SYSTEM_INVALID: NTSTATUS = NTSTATUS(-1073741069i32);
    pub const STATUS_CSS_AUTHENTICATION_FAILURE: NTSTATUS = NTSTATUS(-1073741050i32);
    pub const STATUS_CSS_KEY_NOT_ESTABLISHED: NTSTATUS = NTSTATUS(-1073741048i32);
    pub const STATUS_CSS_KEY_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073741049i32);
    pub const STATUS_CSS_REGION_MISMATCH: NTSTATUS = NTSTATUS(-1073741046i32);
    pub const STATUS_CSS_RESETS_EXHAUSTED: NTSTATUS = NTSTATUS(-1073741045i32);
    pub const STATUS_CSS_SCRAMBLED_SECTOR: NTSTATUS = NTSTATUS(-1073741047i32);
    pub const STATUS_CSV_IO_PAUSE_TIMEOUT: NTSTATUS = NTSTATUS(-1072496600i32);
    pub const STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE: NTSTATUS = NTSTATUS(-1073740733i32);
    pub const STATUS_CS_ENCRYPTION_FILE_NOT_CSE: NTSTATUS = NTSTATUS(-1073740731i32);
    pub const STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE: NTSTATUS = NTSTATUS(-1073740735i32);
    pub const STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE: NTSTATUS = NTSTATUS(-1073740732i32);
    pub const STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER: NTSTATUS = NTSTATUS(-1073740734i32);
    pub const STATUS_CTLOG_INCONSISTENT_TRACKING_FILE: NTSTATUS = NTSTATUS(-1069940700i32);
    pub const STATUS_CTLOG_INVALID_TRACKING_STATE: NTSTATUS = NTSTATUS(-1069940701i32);
    pub const STATUS_CTLOG_LOGFILE_SIZE_EXCEEDED_MAXSIZE: NTSTATUS = NTSTATUS(-1069940703i32);
    pub const STATUS_CTLOG_TRACKING_NOT_INITIALIZED: NTSTATUS = NTSTATUS(-1069940704i32);
    pub const STATUS_CTLOG_VHD_CHANGED_OFFLINE: NTSTATUS = NTSTATUS(-1069940702i32);
    pub const STATUS_CTL_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073741737i32);
    pub const STATUS_CTX_BAD_VIDEO_MODE: NTSTATUS = NTSTATUS(-1073086440i32);
    pub const STATUS_CTX_CDM_CONNECT: NTSTATUS = NTSTATUS(1074397188i32);
    pub const STATUS_CTX_CDM_DISCONNECT: NTSTATUS = NTSTATUS(1074397189i32);
    pub const STATUS_CTX_CLIENT_LICENSE_IN_USE: NTSTATUS = NTSTATUS(-1073086412i32);
    pub const STATUS_CTX_CLIENT_LICENSE_NOT_SET: NTSTATUS = NTSTATUS(-1073086413i32);
    pub const STATUS_CTX_CLIENT_QUERY_TIMEOUT: NTSTATUS = NTSTATUS(-1073086426i32);
    pub const STATUS_CTX_CLOSE_PENDING: NTSTATUS = NTSTATUS(-1073086458i32);
    pub const STATUS_CTX_CONSOLE_CONNECT: NTSTATUS = NTSTATUS(-1073086424i32);
    pub const STATUS_CTX_CONSOLE_DISCONNECT: NTSTATUS = NTSTATUS(-1073086425i32);
    pub const STATUS_CTX_GRAPHICS_INVALID: NTSTATUS = NTSTATUS(-1073086430i32);
    pub const STATUS_CTX_INVALID_MODEMNAME: NTSTATUS = NTSTATUS(-1073086455i32);
    pub const STATUS_CTX_INVALID_PD: NTSTATUS = NTSTATUS(-1073086462i32);
    pub const STATUS_CTX_INVALID_WD: NTSTATUS = NTSTATUS(-1073086418i32);
    pub const STATUS_CTX_LICENSE_CLIENT_INVALID: NTSTATUS = NTSTATUS(-1073086446i32);
    pub const STATUS_CTX_LICENSE_EXPIRED: NTSTATUS = NTSTATUS(-1073086444i32);
    pub const STATUS_CTX_LICENSE_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073086445i32);
    pub const STATUS_CTX_LOGON_DISABLED: NTSTATUS = NTSTATUS(-1073086409i32);
    pub const STATUS_CTX_MODEM_INF_NOT_FOUND: NTSTATUS = NTSTATUS(-1073086456i32);
    pub const STATUS_CTX_MODEM_RESPONSE_BUSY: NTSTATUS = NTSTATUS(-1073086450i32);
    pub const STATUS_CTX_MODEM_RESPONSE_NO_CARRIER: NTSTATUS = NTSTATUS(-1073086452i32);
    pub const STATUS_CTX_MODEM_RESPONSE_NO_DIALTONE: NTSTATUS = NTSTATUS(-1073086451i32);
    pub const STATUS_CTX_MODEM_RESPONSE_TIMEOUT: NTSTATUS = NTSTATUS(-1073086453i32);
    pub const STATUS_CTX_MODEM_RESPONSE_VOICE: NTSTATUS = NTSTATUS(-1073086449i32);
    pub const STATUS_CTX_NOT_CONSOLE: NTSTATUS = NTSTATUS(-1073086428i32);
    pub const STATUS_CTX_NO_OUTBUF: NTSTATUS = NTSTATUS(-1073086457i32);
    pub const STATUS_CTX_PD_NOT_FOUND: NTSTATUS = NTSTATUS(-1073086461i32);
    pub const STATUS_CTX_RESPONSE_ERROR: NTSTATUS = NTSTATUS(-1073086454i32);
    pub const STATUS_CTX_SECURITY_LAYER_ERROR: NTSTATUS = NTSTATUS(-1073086408i32);
    pub const STATUS_CTX_SHADOW_DENIED: NTSTATUS = NTSTATUS(-1073086422i32);
    pub const STATUS_CTX_SHADOW_DISABLED: NTSTATUS = NTSTATUS(-1073086415i32);
    pub const STATUS_CTX_SHADOW_ENDED_BY_MODE_CHANGE: NTSTATUS = NTSTATUS(-1073086411i32);
    pub const STATUS_CTX_SHADOW_INVALID: NTSTATUS = NTSTATUS(-1073086416i32);
    pub const STATUS_CTX_SHADOW_NOT_RUNNING: NTSTATUS = NTSTATUS(-1073086410i32);
    pub const STATUS_CTX_TD_ERROR: NTSTATUS = NTSTATUS(-1073086448i32);
    pub const STATUS_CTX_WD_NOT_FOUND: NTSTATUS = NTSTATUS(-1073086417i32);
    pub const STATUS_CTX_WINSTATION_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073086421i32);
    pub const STATUS_CTX_WINSTATION_BUSY: NTSTATUS = NTSTATUS(-1073086441i32);
    pub const STATUS_CTX_WINSTATION_NAME_COLLISION: NTSTATUS = NTSTATUS(-1073086442i32);
    pub const STATUS_CTX_WINSTATION_NAME_INVALID: NTSTATUS = NTSTATUS(-1073086463i32);
    pub const STATUS_CTX_WINSTATION_NOT_FOUND: NTSTATUS = NTSTATUS(-1073086443i32);
    pub const STATUS_CURRENT_DOMAIN_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1073741079i32);
    pub const STATUS_CURRENT_TRANSACTION_NOT_VALID: NTSTATUS = NTSTATUS(-1072103400i32);
    pub const STATUS_DATATYPE_MISALIGNMENT: NTSTATUS = NTSTATUS(-2147483646i32);
    pub const STATUS_DATATYPE_MISALIGNMENT_ERROR: NTSTATUS = NTSTATUS(-1073741115i32);
    pub const STATUS_DATA_CHECKSUM_ERROR: NTSTATUS = NTSTATUS(-1073740688i32);
    pub const STATUS_DATA_ERROR: NTSTATUS = NTSTATUS(-1073741762i32);
    pub const STATUS_DATA_LATE_ERROR: NTSTATUS = NTSTATUS(-1073741763i32);
    pub const STATUS_DATA_LOST_REPAIR: NTSTATUS = NTSTATUS(-2147481597i32);
    pub const STATUS_DATA_NOT_ACCEPTED: NTSTATUS = NTSTATUS(-1073741285i32);
    pub const STATUS_DATA_OVERRUN: NTSTATUS = NTSTATUS(-1073741764i32);
    pub const STATUS_DATA_OVERWRITTEN: NTSTATUS = NTSTATUS(304i32);
    pub const STATUS_DAX_MAPPING_EXISTS: NTSTATUS = NTSTATUS(-1073740644i32);
    pub const STATUS_DEBUGGER_INACTIVE: NTSTATUS = NTSTATUS(-1073740972i32);
    pub const STATUS_DEBUG_ATTACH_FAILED: NTSTATUS = NTSTATUS(-1073741287i32);
    pub const STATUS_DECRYPTION_FAILED: NTSTATUS = NTSTATUS(-1073741173i32);
    pub const STATUS_DELAY_LOAD_FAILED: NTSTATUS = NTSTATUS(-1073740782i32);
    pub const STATUS_DELETE_PENDING: NTSTATUS = NTSTATUS(-1073741738i32);
    pub const STATUS_DESTINATION_ELEMENT_FULL: NTSTATUS = NTSTATUS(-1073741180i32);
    pub const STATUS_DEVICE_ALREADY_ATTACHED: NTSTATUS = NTSTATUS(-1073741768i32);
    pub const STATUS_DEVICE_BUSY: NTSTATUS = NTSTATUS(-2147483631i32);
    pub const STATUS_DEVICE_CONFIGURATION_ERROR: NTSTATUS = NTSTATUS(-1073741438i32);
    pub const STATUS_DEVICE_DATA_ERROR: NTSTATUS = NTSTATUS(-1073741668i32);
    pub const STATUS_DEVICE_DOES_NOT_EXIST: NTSTATUS = NTSTATUS(-1073741632i32);
    pub const STATUS_DEVICE_DOOR_OPEN: NTSTATUS = NTSTATUS(-2147482999i32);
    pub const STATUS_DEVICE_ENUMERATION_ERROR: NTSTATUS = NTSTATUS(-1073740954i32);
    pub const STATUS_DEVICE_FEATURE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740701i32);
    pub const STATUS_DEVICE_HARDWARE_ERROR: NTSTATUS = NTSTATUS(-1073740669i32);
    pub const STATUS_DEVICE_HINT_NAME_BUFFER_TOO_SMALL: NTSTATUS = NTSTATUS(-1073740650i32);
    pub const STATUS_DEVICE_HUNG: NTSTATUS = NTSTATUS(-1073740537i32);
    pub const STATUS_DEVICE_INSUFFICIENT_RESOURCES: NTSTATUS = NTSTATUS(-1073740696i32);
    pub const STATUS_DEVICE_IN_MAINTENANCE: NTSTATUS = NTSTATUS(-1073740647i32);
    pub const STATUS_DEVICE_NOT_CONNECTED: NTSTATUS = NTSTATUS(-1073741667i32);
    pub const STATUS_DEVICE_NOT_PARTITIONED: NTSTATUS = NTSTATUS(-1073741452i32);
    pub const STATUS_DEVICE_NOT_READY: NTSTATUS = NTSTATUS(-1073741661i32);
    pub const STATUS_DEVICE_OFF_LINE: NTSTATUS = NTSTATUS(-2147483632i32);
    pub const STATUS_DEVICE_PAPER_EMPTY: NTSTATUS = NTSTATUS(-2147483634i32);
    pub const STATUS_DEVICE_POWERED_OFF: NTSTATUS = NTSTATUS(-2147483633i32);
    pub const STATUS_DEVICE_POWER_CYCLE_REQUIRED: NTSTATUS = NTSTATUS(-2147483599i32);
    pub const STATUS_DEVICE_POWER_FAILURE: NTSTATUS = NTSTATUS(-1073741666i32);
    pub const STATUS_DEVICE_PROTOCOL_ERROR: NTSTATUS = NTSTATUS(-1073741434i32);
    pub const STATUS_DEVICE_REMOVED: NTSTATUS = NTSTATUS(-1073741130i32);
    pub const STATUS_DEVICE_REQUIRES_CLEANING: NTSTATUS = NTSTATUS(-2147483000i32);
    pub const STATUS_DEVICE_RESET_REQUIRED: NTSTATUS = NTSTATUS(-2147483210i32);
    pub const STATUS_DEVICE_SUPPORT_IN_PROGRESS: NTSTATUS = NTSTATUS(-2147483600i32);
    pub const STATUS_DEVICE_UNREACHABLE: NTSTATUS = NTSTATUS(-1073740700i32);
    pub const STATUS_DEVICE_UNRESPONSIVE: NTSTATUS = NTSTATUS(-1073740534i32);
    pub const STATUS_DFS_EXIT_PATH_FOUND: NTSTATUS = NTSTATUS(-1073741669i32);
    pub const STATUS_DFS_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073741203i32);
    pub const STATUS_DIF_BINDING_API_NOT_FOUND: NTSTATUS = NTSTATUS(-1073738625i32);
    pub const STATUS_DIF_IOCALLBACK_NOT_REPLACED: NTSTATUS = NTSTATUS(-1073738634i32);
    pub const STATUS_DIF_LIVEDUMP_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073738633i32);
    pub const STATUS_DIF_VOLATILE_DRIVER_HOTPATCHED: NTSTATUS = NTSTATUS(-1073738631i32);
    pub const STATUS_DIF_VOLATILE_DRIVER_IS_NOT_RUNNING: NTSTATUS = NTSTATUS(-1073738629i32);
    pub const STATUS_DIF_VOLATILE_INVALID_INFO: NTSTATUS = NTSTATUS(-1073738630i32);
    pub const STATUS_DIF_VOLATILE_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1073738626i32);
    pub const STATUS_DIF_VOLATILE_PLUGIN_CHANGE_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1073738627i32);
    pub const STATUS_DIF_VOLATILE_PLUGIN_IS_NOT_RUNNING: NTSTATUS = NTSTATUS(-1073738628i32);
    pub const STATUS_DIF_VOLATILE_SECTION_NOT_LOCKED: NTSTATUS = NTSTATUS(-1073738632i32);
    pub const STATUS_DIRECTORY_IS_A_REPARSE_POINT: NTSTATUS = NTSTATUS(-1073741183i32);
    pub const STATUS_DIRECTORY_NOT_EMPTY: NTSTATUS = NTSTATUS(-1073741567i32);
    pub const STATUS_DIRECTORY_NOT_RM: NTSTATUS = NTSTATUS(-1072103416i32);
    pub const STATUS_DIRECTORY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740676i32);
    pub const STATUS_DIRECTORY_SERVICE_REQUIRED: NTSTATUS = NTSTATUS(-1073741135i32);
    pub const STATUS_DISK_CORRUPT_ERROR: NTSTATUS = NTSTATUS(-1073741774i32);
    pub const STATUS_DISK_FULL: NTSTATUS = NTSTATUS(-1073741697i32);
    pub const STATUS_DISK_OPERATION_FAILED: NTSTATUS = NTSTATUS(-1073741462i32);
    pub const STATUS_DISK_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073739774i32);
    pub const STATUS_DISK_RECALIBRATE_FAILED: NTSTATUS = NTSTATUS(-1073741463i32);
    pub const STATUS_DISK_REPAIR_DISABLED: NTSTATUS = NTSTATUS(-1073739776i32);
    pub const STATUS_DISK_REPAIR_REDIRECTED: NTSTATUS = NTSTATUS(1073743879i32);
    pub const STATUS_DISK_REPAIR_UNSUCCESSFUL: NTSTATUS = NTSTATUS(-1073739768i32);
    pub const STATUS_DISK_RESET_FAILED: NTSTATUS = NTSTATUS(-1073741461i32);
    pub const STATUS_DISK_RESOURCES_EXHAUSTED: NTSTATUS = NTSTATUS(-1073740703i32);
    pub const STATUS_DLL_INIT_FAILED: NTSTATUS = NTSTATUS(-1073741502i32);
    pub const STATUS_DLL_INIT_FAILED_LOGOFF: NTSTATUS = NTSTATUS(-1073741205i32);
    pub const STATUS_DLL_MIGHT_BE_INCOMPATIBLE: NTSTATUS = NTSTATUS(-2147483604i32);
    pub const STATUS_DLL_MIGHT_BE_INSECURE: NTSTATUS = NTSTATUS(-2147483605i32);
    pub const STATUS_DLL_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741515i32);
    pub const STATUS_DM_OPERATION_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1070135808i32);
    pub const STATUS_DOMAIN_CONTROLLER_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741261i32);
    pub const STATUS_DOMAIN_CTRLR_CONFIG_ERROR: NTSTATUS = NTSTATUS(-1073741474i32);
    pub const STATUS_DOMAIN_EXISTS: NTSTATUS = NTSTATUS(-1073741600i32);
    pub const STATUS_DOMAIN_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073741599i32);
    pub const STATUS_DOMAIN_TRUST_INCONSISTENT: NTSTATUS = NTSTATUS(-1073741413i32);
    pub const STATUS_DRIVERS_LEAKING_LOCKED_PAGES: NTSTATUS = NTSTATUS(1073741869i32);
    pub const STATUS_DRIVER_BLOCKED: NTSTATUS = NTSTATUS(-1073740948i32);
    pub const STATUS_DRIVER_BLOCKED_CRITICAL: NTSTATUS = NTSTATUS(-1073740949i32);
    pub const STATUS_DRIVER_CANCEL_TIMEOUT: NTSTATUS = NTSTATUS(-1073741282i32);
    pub const STATUS_DRIVER_DATABASE_ERROR: NTSTATUS = NTSTATUS(-1073740947i32);
    pub const STATUS_DRIVER_ENTRYPOINT_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741213i32);
    pub const STATUS_DRIVER_FAILED_PRIOR_UNLOAD: NTSTATUS = NTSTATUS(-1073740914i32);
    pub const STATUS_DRIVER_FAILED_SLEEP: NTSTATUS = NTSTATUS(-1073741118i32);
    pub const STATUS_DRIVER_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1073741437i32);
    pub const STATUS_DRIVER_ORDINAL_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741214i32);
    pub const STATUS_DRIVER_PROCESS_TERMINATED: NTSTATUS = NTSTATUS(-1073740720i32);
    pub const STATUS_DRIVER_UNABLE_TO_LOAD: NTSTATUS = NTSTATUS(-1073741204i32);
    pub const STATUS_DS_ADMIN_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073741119i32);
    pub const STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER: NTSTATUS = NTSTATUS(-1073740968i32);
    pub const STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS: NTSTATUS = NTSTATUS(-1073741148i32);
    pub const STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED: NTSTATUS = NTSTATUS(-1073741149i32);
    pub const STATUS_DS_BUSY: NTSTATUS = NTSTATUS(-1073741147i32);
    pub const STATUS_DS_CANT_MOD_OBJ_CLASS: NTSTATUS = NTSTATUS(-1073741138i32);
    pub const STATUS_DS_CANT_MOD_PRIMARYGROUPID: NTSTATUS = NTSTATUS(-1073741104i32);
    pub const STATUS_DS_CANT_ON_NON_LEAF: NTSTATUS = NTSTATUS(-1073741140i32);
    pub const STATUS_DS_CANT_ON_RDN: NTSTATUS = NTSTATUS(-1073741139i32);
    pub const STATUS_DS_CANT_START: NTSTATUS = NTSTATUS(-1073741087i32);
    pub const STATUS_DS_CROSS_DOM_MOVE_FAILED: NTSTATUS = NTSTATUS(-1073741137i32);
    pub const STATUS_DS_DOMAIN_NAME_EXISTS_IN_FOREST: NTSTATUS = NTSTATUS(-1073740774i32);
    pub const STATUS_DS_DOMAIN_RENAME_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073739775i32);
    pub const STATUS_DS_DUPLICATE_ID_FOUND: NTSTATUS = NTSTATUS(-1073740795i32);
    pub const STATUS_DS_FLAT_NAME_EXISTS_IN_FOREST: NTSTATUS = NTSTATUS(-1073740773i32);
    pub const STATUS_DS_GC_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073741136i32);
    pub const STATUS_DS_GC_REQUIRED: NTSTATUS = NTSTATUS(-1073741084i32);
    pub const STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER: NTSTATUS = NTSTATUS(-1073741094i32);
    pub const STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER: NTSTATUS = NTSTATUS(-1073741097i32);
    pub const STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER: NTSTATUS = NTSTATUS(-1073741096i32);
    pub const STATUS_DS_GROUP_CONVERSION_ERROR: NTSTATUS = NTSTATUS(-1073740794i32);
    pub const STATUS_DS_HAVE_PRIMARY_MEMBERS: NTSTATUS = NTSTATUS(-1073741092i32);
    pub const STATUS_DS_INCORRECT_ROLE_OWNER: NTSTATUS = NTSTATUS(-1073741143i32);
    pub const STATUS_DS_INIT_FAILURE: NTSTATUS = NTSTATUS(-1073741086i32);
    pub const STATUS_DS_INIT_FAILURE_CONSOLE: NTSTATUS = NTSTATUS(-1073741076i32);
    pub const STATUS_DS_INVALID_ATTRIBUTE_SYNTAX: NTSTATUS = NTSTATUS(-1073741150i32);
    pub const STATUS_DS_INVALID_GROUP_TYPE: NTSTATUS = NTSTATUS(-1073741100i32);
    pub const STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER: NTSTATUS = NTSTATUS(-1073741093i32);
    pub const STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY: NTSTATUS = NTSTATUS(-1073741083i32);
    pub const STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073741081i32);
    pub const STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY: NTSTATUS = NTSTATUS(289i32);
    pub const STATUS_DS_NAME_NOT_UNIQUE: NTSTATUS = NTSTATUS(-1073740796i32);
    pub const STATUS_DS_NO_ATTRIBUTE_OR_VALUE: NTSTATUS = NTSTATUS(-1073741151i32);
    pub const STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS: NTSTATUS = NTSTATUS(-1073741082i32);
    pub const STATUS_DS_NO_MORE_RIDS: NTSTATUS = NTSTATUS(-1073741144i32);
    pub const STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN: NTSTATUS = NTSTATUS(-1073741099i32);
    pub const STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN: NTSTATUS = NTSTATUS(-1073741098i32);
    pub const STATUS_DS_NO_RIDS_ALLOCATED: NTSTATUS = NTSTATUS(-1073741145i32);
    pub const STATUS_DS_OBJ_CLASS_VIOLATION: NTSTATUS = NTSTATUS(-1073741141i32);
    pub const STATUS_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS: NTSTATUS = NTSTATUS(-1073700729i32);
    pub const STATUS_DS_OID_NOT_FOUND: NTSTATUS = NTSTATUS(-1073700728i32);
    pub const STATUS_DS_RIDMGR_DISABLED: NTSTATUS = NTSTATUS(-1073741126i32);
    pub const STATUS_DS_RIDMGR_INIT_ERROR: NTSTATUS = NTSTATUS(-1073741142i32);
    pub const STATUS_DS_SAM_INIT_FAILURE: NTSTATUS = NTSTATUS(-1073741109i32);
    pub const STATUS_DS_SAM_INIT_FAILURE_CONSOLE: NTSTATUS = NTSTATUS(-1073741075i32);
    pub const STATUS_DS_SENSITIVE_GROUP_VIOLATION: NTSTATUS = NTSTATUS(-1073741107i32);
    pub const STATUS_DS_SHUTTING_DOWN: NTSTATUS = NTSTATUS(1073742704i32);
    pub const STATUS_DS_SRC_SID_EXISTS_IN_FOREST: NTSTATUS = NTSTATUS(-1073740775i32);
    pub const STATUS_DS_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073741146i32);
    pub const STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER: NTSTATUS = NTSTATUS(-1073741095i32);
    pub const STATUS_DS_VERSION_CHECK_FAILURE: NTSTATUS = NTSTATUS(-1073740971i32);
    pub const STATUS_DUPLICATE_NAME: NTSTATUS = NTSTATUS(-1073741635i32);
    pub const STATUS_DUPLICATE_OBJECTID: NTSTATUS = NTSTATUS(-1073741270i32);
    pub const STATUS_DUPLICATE_PRIVILEGES: NTSTATUS = NTSTATUS(-1073741402i32);
    pub const STATUS_DYNAMIC_CODE_BLOCKED: NTSTATUS = NTSTATUS(-1073740284i32);
    pub const STATUS_EAS_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073741745i32);
    pub const STATUS_EA_CORRUPT_ERROR: NTSTATUS = NTSTATUS(-1073741741i32);
    pub const STATUS_EA_LIST_INCONSISTENT: NTSTATUS = NTSTATUS(-2147483628i32);
    pub const STATUS_EA_TOO_LARGE: NTSTATUS = NTSTATUS(-1073741744i32);
    pub const STATUS_EFS_ALG_BLOB_TOO_BIG: NTSTATUS = NTSTATUS(-1073740974i32);
    pub const STATUS_EFS_NOT_ALLOWED_IN_TRANSACTION: NTSTATUS = NTSTATUS(-1072103362i32);
    pub const STATUS_ELEVATION_REQUIRED: NTSTATUS = NTSTATUS(-1073740756i32);
    pub const STATUS_EMULATION_BREAKPOINT: NTSTATUS = NTSTATUS(1073741880i32);
    pub const STATUS_EMULATION_SYSCALL: NTSTATUS = NTSTATUS(1073741881i32);
    pub const STATUS_ENCLAVE_FAILURE: NTSTATUS = NTSTATUS(-1073740657i32);
    pub const STATUS_ENCLAVE_IS_TERMINATING: NTSTATUS = NTSTATUS(-1073740526i32);
    pub const STATUS_ENCLAVE_NOT_TERMINATED: NTSTATUS = NTSTATUS(-1073740527i32);
    pub const STATUS_ENCLAVE_VIOLATION: NTSTATUS = NTSTATUS(-1073740638i32);
    pub const STATUS_ENCOUNTERED_WRITE_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073740749i32);
    pub const STATUS_ENCRYPTED_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740605i32);
    pub const STATUS_ENCRYPTED_IO_NOT_POSSIBLE: NTSTATUS = NTSTATUS(-1073739760i32);
    pub const STATUS_ENCRYPTING_METADATA_DISALLOWED: NTSTATUS = NTSTATUS(-1073740617i32);
    pub const STATUS_ENCRYPTION_DISABLED: NTSTATUS = NTSTATUS(-1073740618i32);
    pub const STATUS_ENCRYPTION_FAILED: NTSTATUS = NTSTATUS(-1073741174i32);
    pub const STATUS_END_OF_FILE: NTSTATUS = NTSTATUS(-1073741807i32);
    pub const STATUS_END_OF_MEDIA: NTSTATUS = NTSTATUS(-2147483618i32);
    pub const STATUS_ENLISTMENT_NOT_FOUND: NTSTATUS = NTSTATUS(-1072103344i32);
    pub const STATUS_ENLISTMENT_NOT_SUPERIOR: NTSTATUS = NTSTATUS(-1072103373i32);
    pub const STATUS_ENTRYPOINT_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741511i32);
    pub const STATUS_EOF_ON_GHOSTED_RANGE: NTSTATUS = NTSTATUS(-1073700857i32);
    pub const STATUS_EOM_OVERFLOW: NTSTATUS = NTSTATUS(-1073741449i32);
    pub const STATUS_ERROR_PROCESS_NOT_IN_JOB: NTSTATUS = NTSTATUS(-1073741394i32);
    pub const STATUS_EVALUATION_EXPIRATION: NTSTATUS = NTSTATUS(-1073741208i32);
    pub const STATUS_EVENTLOG_CANT_START: NTSTATUS = NTSTATUS(-1073741425i32);
    pub const STATUS_EVENTLOG_FILE_CHANGED: NTSTATUS = NTSTATUS(-1073741417i32);
    pub const STATUS_EVENTLOG_FILE_CORRUPT: NTSTATUS = NTSTATUS(-1073741426i32);
    pub const STATUS_EVENT_DONE: NTSTATUS = NTSTATUS(1073741842i32);
    pub const STATUS_EVENT_PENDING: NTSTATUS = NTSTATUS(1073741843i32);
    pub const STATUS_EXECUTABLE_MEMORY_WRITE: NTSTATUS = NTSTATUS(-1073739997i32);
    pub const STATUS_EXPIRED_HANDLE: NTSTATUS = NTSTATUS(-1072103328i32);
    pub const STATUS_EXTERNAL_BACKING_PROVIDER_UNKNOWN: NTSTATUS = NTSTATUS(-1073740690i32);
    pub const STATUS_EXTERNAL_SYSKEY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740639i32);
    pub const STATUS_EXTRANEOUS_INFORMATION: NTSTATUS = NTSTATUS(-2147483625i32);
    pub const STATUS_FAILED_DRIVER_ENTRY: NTSTATUS = NTSTATUS(-1073740955i32);
    pub const STATUS_FAILED_STACK_SWITCH: NTSTATUS = NTSTATUS(-1073740941i32);
    pub const STATUS_FAIL_CHECK: NTSTATUS = NTSTATUS(-1073741271i32);
    pub const STATUS_FAIL_FAST_EXCEPTION: NTSTATUS = NTSTATUS(-1073740286i32);
    pub const STATUS_FASTPATH_REJECTED: NTSTATUS = NTSTATUS(-1073700844i32);
    pub const STATUS_FATAL_APP_EXIT: NTSTATUS = NTSTATUS(1073741845i32);
    pub const STATUS_FATAL_MEMORY_EXHAUSTION: NTSTATUS = NTSTATUS(-1073741395i32);
    pub const STATUS_FATAL_USER_CALLBACK_EXCEPTION: NTSTATUS = NTSTATUS(-1073740771i32);
    pub const STATUS_FILEMARK_DETECTED: NTSTATUS = NTSTATUS(-2147483621i32);
    pub const STATUS_FILES_OPEN: NTSTATUS = NTSTATUS(-1073741561i32);
    pub const STATUS_FILE_CHECKED_OUT: NTSTATUS = NTSTATUS(-1073739519i32);
    pub const STATUS_FILE_CLOSED: NTSTATUS = NTSTATUS(-1073741528i32);
    pub const STATUS_FILE_CORRUPT_ERROR: NTSTATUS = NTSTATUS(-1073741566i32);
    pub const STATUS_FILE_DELETED: NTSTATUS = NTSTATUS(-1073741533i32);
    pub const STATUS_FILE_ENCRYPTED: NTSTATUS = NTSTATUS(-1073741165i32);
    pub const STATUS_FILE_FORCED_CLOSED: NTSTATUS = NTSTATUS(-1073741642i32);
    pub const STATUS_FILE_HANDLE_REVOKED: NTSTATUS = NTSTATUS(-1073739504i32);
    pub const STATUS_FILE_IDENTITY_NOT_PERSISTENT: NTSTATUS = NTSTATUS(-1072103370i32);
    pub const STATUS_FILE_INVALID: NTSTATUS = NTSTATUS(-1073741672i32);
    pub const STATUS_FILE_IS_A_DIRECTORY: NTSTATUS = NTSTATUS(-1073741638i32);
    pub const STATUS_FILE_IS_OFFLINE: NTSTATUS = NTSTATUS(-1073741209i32);
    pub const STATUS_FILE_LOCKED_WITH_ONLY_READERS: NTSTATUS = NTSTATUS(298i32);
    pub const STATUS_FILE_LOCKED_WITH_WRITERS: NTSTATUS = NTSTATUS(299i32);
    pub const STATUS_FILE_LOCK_CONFLICT: NTSTATUS = NTSTATUS(-1073741740i32);
    pub const STATUS_FILE_METADATA_OPTIMIZATION_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073741397i32);
    pub const STATUS_FILE_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073740697i32);
    pub const STATUS_FILE_NOT_ENCRYPTED: NTSTATUS = NTSTATUS(-1073741167i32);
    pub const STATUS_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740620i32);
    pub const STATUS_FILE_PROTECTED_UNDER_DPL: NTSTATUS = NTSTATUS(-1073740637i32);
    pub const STATUS_FILE_RENAMED: NTSTATUS = NTSTATUS(-1073741611i32);
    pub const STATUS_FILE_SNAP_INVALID_PARAMETER: NTSTATUS = NTSTATUS(-1073679099i32);
    pub const STATUS_FILE_SNAP_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073679104i32);
    pub const STATUS_FILE_SNAP_IO_NOT_COORDINATED: NTSTATUS = NTSTATUS(-1073679101i32);
    pub const STATUS_FILE_SNAP_MODIFY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073679102i32);
    pub const STATUS_FILE_SNAP_UNEXPECTED_ERROR: NTSTATUS = NTSTATUS(-1073679100i32);
    pub const STATUS_FILE_SNAP_USER_SECTION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073679103i32);
    pub const STATUS_FILE_SYSTEM_LIMITATION: NTSTATUS = NTSTATUS(-1073740761i32);
    pub const STATUS_FILE_SYSTEM_VIRTUALIZATION_BUSY: NTSTATUS = NTSTATUS(-1073689085i32);
    pub const STATUS_FILE_SYSTEM_VIRTUALIZATION_INVALID_OPERATION: NTSTATUS = NTSTATUS(-1073689083i32);
    pub const STATUS_FILE_SYSTEM_VIRTUALIZATION_METADATA_CORRUPT: NTSTATUS = NTSTATUS(-1073689086i32);
    pub const STATUS_FILE_SYSTEM_VIRTUALIZATION_PROVIDER_UNKNOWN: NTSTATUS = NTSTATUS(-1073689084i32);
    pub const STATUS_FILE_SYSTEM_VIRTUALIZATION_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073689087i32);
    pub const STATUS_FILE_TOO_LARGE: NTSTATUS = NTSTATUS(-1073739516i32);
    pub const STATUS_FIRMWARE_IMAGE_INVALID: NTSTATUS = NTSTATUS(-1073740667i32);
    pub const STATUS_FIRMWARE_SLOT_INVALID: NTSTATUS = NTSTATUS(-1073740668i32);
    pub const STATUS_FIRMWARE_UPDATED: NTSTATUS = NTSTATUS(1073741868i32);
    pub const STATUS_FLOATED_SECTION: NTSTATUS = NTSTATUS(-1072103349i32);
    pub const STATUS_FLOAT_DENORMAL_OPERAND: NTSTATUS = NTSTATUS(-1073741683i32);
    pub const STATUS_FLOAT_DIVIDE_BY_ZERO: NTSTATUS = NTSTATUS(-1073741682i32);
    pub const STATUS_FLOAT_INEXACT_RESULT: NTSTATUS = NTSTATUS(-1073741681i32);
    pub const STATUS_FLOAT_INVALID_OPERATION: NTSTATUS = NTSTATUS(-1073741680i32);
    pub const STATUS_FLOAT_MULTIPLE_FAULTS: NTSTATUS = NTSTATUS(-1073741132i32);
    pub const STATUS_FLOAT_MULTIPLE_TRAPS: NTSTATUS = NTSTATUS(-1073741131i32);
    pub const STATUS_FLOAT_OVERFLOW: NTSTATUS = NTSTATUS(-1073741679i32);
    pub const STATUS_FLOAT_STACK_CHECK: NTSTATUS = NTSTATUS(-1073741678i32);
    pub const STATUS_FLOAT_UNDERFLOW: NTSTATUS = NTSTATUS(-1073741677i32);
    pub const STATUS_FLOPPY_BAD_REGISTERS: NTSTATUS = NTSTATUS(-1073741464i32);
    pub const STATUS_FLOPPY_ID_MARK_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741467i32);
    pub const STATUS_FLOPPY_UNKNOWN_ERROR: NTSTATUS = NTSTATUS(-1073741465i32);
    pub const STATUS_FLOPPY_VOLUME: NTSTATUS = NTSTATUS(-1073741468i32);
    pub const STATUS_FLOPPY_WRONG_CYLINDER: NTSTATUS = NTSTATUS(-1073741466i32);
    pub const STATUS_FLT_ALREADY_ENLISTED: NTSTATUS = NTSTATUS(-1071906789i32);
    pub const STATUS_FLT_BUFFER_TOO_SMALL: NTSTATUS = NTSTATUS(-2145648639i32);
    pub const STATUS_FLT_CBDQ_DISABLED: NTSTATUS = NTSTATUS(-1071906802i32);
    pub const STATUS_FLT_CONTEXT_ALLOCATION_NOT_FOUND: NTSTATUS = NTSTATUS(-1071906794i32);
    pub const STATUS_FLT_CONTEXT_ALREADY_DEFINED: NTSTATUS = NTSTATUS(-1071906814i32);
    pub const STATUS_FLT_CONTEXT_ALREADY_LINKED: NTSTATUS = NTSTATUS(-1071906788i32);
    pub const STATUS_FLT_DELETING_OBJECT: NTSTATUS = NTSTATUS(-1071906805i32);
    pub const STATUS_FLT_DISALLOW_FAST_IO: NTSTATUS = NTSTATUS(-1071906812i32);
    pub const STATUS_FLT_DISALLOW_FSFILTER_IO: i32 = -1071906812i32;
    pub const STATUS_FLT_DO_NOT_ATTACH: NTSTATUS = NTSTATUS(-1071906801i32);
    pub const STATUS_FLT_DO_NOT_DETACH: NTSTATUS = NTSTATUS(-1071906800i32);
    pub const STATUS_FLT_DUPLICATE_ENTRY: NTSTATUS = NTSTATUS(-1071906803i32);
    pub const STATUS_FLT_FILTER_NOT_FOUND: NTSTATUS = NTSTATUS(-1071906797i32);
    pub const STATUS_FLT_FILTER_NOT_READY: NTSTATUS = NTSTATUS(-1071906808i32);
    pub const STATUS_FLT_INSTANCE_ALTITUDE_COLLISION: NTSTATUS = NTSTATUS(-1071906799i32);
    pub const STATUS_FLT_INSTANCE_NAME_COLLISION: NTSTATUS = NTSTATUS(-1071906798i32);
    pub const STATUS_FLT_INSTANCE_NOT_FOUND: NTSTATUS = NTSTATUS(-1071906795i32);
    pub const STATUS_FLT_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071906806i32);
    pub const STATUS_FLT_INVALID_ASYNCHRONOUS_REQUEST: NTSTATUS = NTSTATUS(-1071906813i32);
    pub const STATUS_FLT_INVALID_CONTEXT_REGISTRATION: NTSTATUS = NTSTATUS(-1071906793i32);
    pub const STATUS_FLT_INVALID_NAME_REQUEST: NTSTATUS = NTSTATUS(-1071906811i32);
    pub const STATUS_FLT_IO_COMPLETE: NTSTATUS = NTSTATUS(1835009i32);
    pub const STATUS_FLT_MUST_BE_NONPAGED_POOL: NTSTATUS = NTSTATUS(-1071906804i32);
    pub const STATUS_FLT_NAME_CACHE_MISS: NTSTATUS = NTSTATUS(-1071906792i32);
    pub const STATUS_FLT_NOT_INITIALIZED: NTSTATUS = NTSTATUS(-1071906809i32);
    pub const STATUS_FLT_NOT_SAFE_TO_POST_OPERATION: NTSTATUS = NTSTATUS(-1071906810i32);
    pub const STATUS_FLT_NO_DEVICE_OBJECT: NTSTATUS = NTSTATUS(-1071906791i32);
    pub const STATUS_FLT_NO_HANDLER_DEFINED: NTSTATUS = NTSTATUS(-1071906815i32);
    pub const STATUS_FLT_NO_WAITER_FOR_REPLY: NTSTATUS = NTSTATUS(-1071906784i32);
    pub const STATUS_FLT_POST_OPERATION_CLEANUP: NTSTATUS = NTSTATUS(-1071906807i32);
    pub const STATUS_FLT_REGISTRATION_BUSY: NTSTATUS = NTSTATUS(-1071906781i32);
    pub const STATUS_FLT_VOLUME_ALREADY_MOUNTED: NTSTATUS = NTSTATUS(-1071906790i32);
    pub const STATUS_FLT_VOLUME_NOT_FOUND: NTSTATUS = NTSTATUS(-1071906796i32);
    pub const STATUS_FLT_WCOS_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071906780i32);
    pub const STATUS_FORMS_AUTH_REQUIRED: NTSTATUS = NTSTATUS(-1073739515i32);
    pub const STATUS_FOUND_OUT_OF_SCOPE: NTSTATUS = NTSTATUS(-1073741266i32);
    pub const STATUS_FREE_SPACE_TOO_FRAGMENTED: NTSTATUS = NTSTATUS(-1073740645i32);
    pub const STATUS_FREE_VM_NOT_AT_BASE: NTSTATUS = NTSTATUS(-1073741665i32);
    pub const STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY: NTSTATUS = NTSTATUS(294i32);
    pub const STATUS_FS_DRIVER_REQUIRED: NTSTATUS = NTSTATUS(-1073741412i32);
    pub const STATUS_FS_METADATA_INCONSISTENT: NTSTATUS = NTSTATUS(-1073740520i32);
    pub const STATUS_FT_DI_SCAN_REQUIRED: NTSTATUS = NTSTATUS(-1073740692i32);
    pub const STATUS_FT_MISSING_MEMBER: NTSTATUS = NTSTATUS(-1073741473i32);
    pub const STATUS_FT_ORPHANING: NTSTATUS = NTSTATUS(-1073741459i32);
    pub const STATUS_FT_READ_FAILURE: NTSTATUS = NTSTATUS(-1073740629i32);
    pub const STATUS_FT_READ_FROM_COPY: NTSTATUS = NTSTATUS(1073741877i32);
    pub const STATUS_FT_READ_FROM_COPY_FAILURE: NTSTATUS = NTSTATUS(-1073740609i32);
    pub const STATUS_FT_READ_RECOVERY_FROM_BACKUP: NTSTATUS = NTSTATUS(1073741834i32);
    pub const STATUS_FT_WRITE_FAILURE: NTSTATUS = NTSTATUS(-1073740693i32);
    pub const STATUS_FT_WRITE_RECOVERY: NTSTATUS = NTSTATUS(1073741835i32);
    pub const STATUS_FULLSCREEN_MODE: NTSTATUS = NTSTATUS(-1073741479i32);
    pub const STATUS_FVE_ACTION_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1071579127i32);
    pub const STATUS_FVE_AUTH_INVALID_APPLICATION: NTSTATUS = NTSTATUS(-1071579109i32);
    pub const STATUS_FVE_AUTH_INVALID_CONFIG: NTSTATUS = NTSTATUS(-1071579108i32);
    pub const STATUS_FVE_BAD_DATA: NTSTATUS = NTSTATUS(-1071579126i32);
    pub const STATUS_FVE_BAD_INFORMATION: NTSTATUS = NTSTATUS(-1071579134i32);
    pub const STATUS_FVE_BAD_METADATA_POINTER: NTSTATUS = NTSTATUS(-1071579105i32);
    pub const STATUS_FVE_BAD_PARTITION_SIZE: NTSTATUS = NTSTATUS(-1071579131i32);
    pub const STATUS_FVE_CONV_READ_ERROR: NTSTATUS = NTSTATUS(-1071579123i32);
    pub const STATUS_FVE_CONV_RECOVERY_FAILED: NTSTATUS = NTSTATUS(-1071579096i32);
    pub const STATUS_FVE_CONV_WRITE_ERROR: NTSTATUS = NTSTATUS(-1071579122i32);
    pub const STATUS_FVE_DATASET_FULL: NTSTATUS = NTSTATUS(-1071579069i32);
    pub const STATUS_FVE_DEBUGGER_ENABLED: NTSTATUS = NTSTATUS(-1071579107i32);
    pub const STATUS_FVE_DEVICE_LOCKEDOUT: NTSTATUS = NTSTATUS(-1071579077i32);
    pub const STATUS_FVE_DRY_RUN_FAILED: NTSTATUS = NTSTATUS(-1071579106i32);
    pub const STATUS_FVE_EDRIVE_BAND_ENUMERATION_FAILED: NTSTATUS = NTSTATUS(-1071579071i32);
    pub const STATUS_FVE_EDRIVE_DRY_RUN_FAILED: NTSTATUS = NTSTATUS(-1071579080i32);
    pub const STATUS_FVE_ENH_PIN_INVALID: NTSTATUS = NTSTATUS(-1071579087i32);
    pub const STATUS_FVE_FAILED_AUTHENTICATION: NTSTATUS = NTSTATUS(-1071579119i32);
    pub const STATUS_FVE_FAILED_SECTOR_SIZE: NTSTATUS = NTSTATUS(-1071579120i32);
    pub const STATUS_FVE_FAILED_WRONG_FS: NTSTATUS = NTSTATUS(-1071579132i32);
    pub const STATUS_FVE_FS_MOUNTED: NTSTATUS = NTSTATUS(-1071579129i32);
    pub const STATUS_FVE_FS_NOT_EXTENDED: NTSTATUS = NTSTATUS(-1071579130i32);
    pub const STATUS_FVE_FULL_ENCRYPTION_NOT_ALLOWED_ON_TP_STORAGE: NTSTATUS = NTSTATUS(-1071579086i32);
    pub const STATUS_FVE_INVALID_DATUM_TYPE: NTSTATUS = NTSTATUS(-1071579094i32);
    pub const STATUS_FVE_KEYFILE_INVALID: NTSTATUS = NTSTATUS(-1071579116i32);
    pub const STATUS_FVE_KEYFILE_NOT_FOUND: NTSTATUS = NTSTATUS(-1071579117i32);
    pub const STATUS_FVE_KEYFILE_NO_VMK: NTSTATUS = NTSTATUS(-1071579115i32);
    pub const STATUS_FVE_LOCKED_VOLUME: NTSTATUS = NTSTATUS(-1071579136i32);
    pub const STATUS_FVE_METADATA_FULL: NTSTATUS = NTSTATUS(-1071579068i32);
    pub const STATUS_FVE_MOR_FAILED: NTSTATUS = NTSTATUS(-1071579099i32);
    pub const STATUS_FVE_NOT_ALLOWED_ON_CLUSTER: NTSTATUS = NTSTATUS(-1071579083i32);
    pub const STATUS_FVE_NOT_ALLOWED_ON_CSV_STACK: NTSTATUS = NTSTATUS(-1071579084i32);
    pub const STATUS_FVE_NOT_ALLOWED_TO_UPGRADE_WHILE_CONVERTING: NTSTATUS = NTSTATUS(-1071579082i32);
    pub const STATUS_FVE_NOT_DATA_VOLUME: NTSTATUS = NTSTATUS(-1071579124i32);
    pub const STATUS_FVE_NOT_DE_VOLUME: NTSTATUS = NTSTATUS(-1071579075i32);
    pub const STATUS_FVE_NOT_ENCRYPTED: NTSTATUS = NTSTATUS(-1071579135i32);
    pub const STATUS_FVE_NOT_OS_VOLUME: NTSTATUS = NTSTATUS(-1071579118i32);
    pub const STATUS_FVE_NO_AUTOUNLOCK_MASTER_KEY: NTSTATUS = NTSTATUS(-1071579100i32);
    pub const STATUS_FVE_NO_FEATURE_LICENSE: NTSTATUS = NTSTATUS(-1071579098i32);
    pub const STATUS_FVE_NO_LICENSE: NTSTATUS = NTSTATUS(-1071579128i32);
    pub const STATUS_FVE_OLD_METADATA_COPY: NTSTATUS = NTSTATUS(-1071579104i32);
    pub const STATUS_FVE_OSV_KSR_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1071579072i32);
    pub const STATUS_FVE_OVERLAPPED_UPDATE: NTSTATUS = NTSTATUS(-1071579121i32);
    pub const STATUS_FVE_PARTIAL_METADATA: NTSTATUS = NTSTATUS(-2145320959i32);
    pub const STATUS_FVE_PIN_INVALID: NTSTATUS = NTSTATUS(-1071579110i32);
    pub const STATUS_FVE_POLICY_ON_RDV_EXCLUSION_LIST: NTSTATUS = NTSTATUS(-1071579070i32);
    pub const STATUS_FVE_POLICY_USER_DISABLE_RDV_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1071579097i32);
    pub const STATUS_FVE_PROTECTION_CANNOT_BE_DISABLED: NTSTATUS = NTSTATUS(-1071579073i32);
    pub const STATUS_FVE_PROTECTION_DISABLED: NTSTATUS = NTSTATUS(-1071579074i32);
    pub const STATUS_FVE_RAW_ACCESS: NTSTATUS = NTSTATUS(-1071579102i32);
    pub const STATUS_FVE_RAW_BLOCKED: NTSTATUS = NTSTATUS(-1071579101i32);
    pub const STATUS_FVE_REBOOT_REQUIRED: NTSTATUS = NTSTATUS(-1071579103i32);
    pub const STATUS_FVE_SECUREBOOT_CONFIG_CHANGE: NTSTATUS = NTSTATUS(-1071579078i32);
    pub const STATUS_FVE_SECUREBOOT_DISABLED: NTSTATUS = NTSTATUS(-1071579079i32);
    pub const STATUS_FVE_TOO_SMALL: NTSTATUS = NTSTATUS(-1071579133i32);
    pub const STATUS_FVE_TPM_DISABLED: NTSTATUS = NTSTATUS(-1071579114i32);
    pub const STATUS_FVE_TPM_INVALID_PCR: NTSTATUS = NTSTATUS(-1071579112i32);
    pub const STATUS_FVE_TPM_NO_VMK: NTSTATUS = NTSTATUS(-1071579111i32);
    pub const STATUS_FVE_TPM_SRK_AUTH_NOT_ZERO: NTSTATUS = NTSTATUS(-1071579113i32);
    pub const STATUS_FVE_TRANSIENT_STATE: NTSTATUS = NTSTATUS(-2145320958i32);
    pub const STATUS_FVE_VIRTUALIZED_SPACE_TOO_BIG: NTSTATUS = NTSTATUS(-1071579095i32);
    pub const STATUS_FVE_VOLUME_EXTEND_PREVENTS_EOW_DECRYPT: NTSTATUS = NTSTATUS(-1071579076i32);
    pub const STATUS_FVE_VOLUME_NOT_BOUND: NTSTATUS = NTSTATUS(-1071579125i32);
    pub const STATUS_FVE_VOLUME_TOO_SMALL: NTSTATUS = NTSTATUS(-1071579088i32);
    pub const STATUS_FVE_WIPE_CANCEL_NOT_APPLICABLE: NTSTATUS = NTSTATUS(-1071579081i32);
    pub const STATUS_FVE_WIPE_NOT_ALLOWED_ON_TP_STORAGE: NTSTATUS = NTSTATUS(-1071579085i32);
    pub const STATUS_FWP_ACTION_INCOMPATIBLE_WITH_LAYER: NTSTATUS = NTSTATUS(-1071513556i32);
    pub const STATUS_FWP_ACTION_INCOMPATIBLE_WITH_SUBLAYER: NTSTATUS = NTSTATUS(-1071513555i32);
    pub const STATUS_FWP_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1071513591i32);
    pub const STATUS_FWP_BUILTIN_OBJECT: NTSTATUS = NTSTATUS(-1071513577i32);
    pub const STATUS_FWP_CALLOUT_NOTIFICATION_FAILED: NTSTATUS = NTSTATUS(-1071513545i32);
    pub const STATUS_FWP_CALLOUT_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513599i32);
    pub const STATUS_FWP_CANNOT_PEND: NTSTATUS = NTSTATUS(-1071513341i32);
    pub const STATUS_FWP_CONDITION_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513598i32);
    pub const STATUS_FWP_CONNECTIONS_DISABLED: NTSTATUS = NTSTATUS(-1071513535i32);
    pub const STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_CALLOUT: NTSTATUS = NTSTATUS(-1071513553i32);
    pub const STATUS_FWP_CONTEXT_INCOMPATIBLE_WITH_LAYER: NTSTATUS = NTSTATUS(-1071513554i32);
    pub const STATUS_FWP_DROP_NOICMP: NTSTATUS = NTSTATUS(-1071513340i32);
    pub const STATUS_FWP_DUPLICATE_AUTH_METHOD: NTSTATUS = NTSTATUS(-1071513540i32);
    pub const STATUS_FWP_DUPLICATE_CONDITION: NTSTATUS = NTSTATUS(-1071513558i32);
    pub const STATUS_FWP_DUPLICATE_KEYMOD: NTSTATUS = NTSTATUS(-1071513557i32);
    pub const STATUS_FWP_DYNAMIC_SESSION_IN_PROGRESS: NTSTATUS = NTSTATUS(-1071513589i32);
    pub const STATUS_FWP_EM_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071513550i32);
    pub const STATUS_FWP_FILTER_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513597i32);
    pub const STATUS_FWP_IKEEXT_NOT_RUNNING: NTSTATUS = NTSTATUS(-1071513532i32);
    pub const STATUS_FWP_INCOMPATIBLE_AUTH_METHOD: NTSTATUS = NTSTATUS(-1071513552i32);
    pub const STATUS_FWP_INCOMPATIBLE_CIPHER_TRANSFORM: NTSTATUS = NTSTATUS(-1071513542i32);
    pub const STATUS_FWP_INCOMPATIBLE_DH_GROUP: NTSTATUS = NTSTATUS(-1071513551i32);
    pub const STATUS_FWP_INCOMPATIBLE_LAYER: NTSTATUS = NTSTATUS(-1071513580i32);
    pub const STATUS_FWP_INCOMPATIBLE_SA_STATE: NTSTATUS = NTSTATUS(-1071513573i32);
    pub const STATUS_FWP_INCOMPATIBLE_TXN: NTSTATUS = NTSTATUS(-1071513583i32);
    pub const STATUS_FWP_INJECT_HANDLE_CLOSING: NTSTATUS = NTSTATUS(-1071513343i32);
    pub const STATUS_FWP_INJECT_HANDLE_STALE: NTSTATUS = NTSTATUS(-1071513342i32);
    pub const STATUS_FWP_INVALID_ACTION_TYPE: NTSTATUS = NTSTATUS(-1071513564i32);
    pub const STATUS_FWP_INVALID_AUTH_TRANSFORM: NTSTATUS = NTSTATUS(-1071513544i32);
    pub const STATUS_FWP_INVALID_CIPHER_TRANSFORM: NTSTATUS = NTSTATUS(-1071513543i32);
    pub const STATUS_FWP_INVALID_DNS_NAME: NTSTATUS = NTSTATUS(-1071513534i32);
    pub const STATUS_FWP_INVALID_ENUMERATOR: NTSTATUS = NTSTATUS(-1071513571i32);
    pub const STATUS_FWP_INVALID_FLAGS: NTSTATUS = NTSTATUS(-1071513570i32);
    pub const STATUS_FWP_INVALID_INTERVAL: NTSTATUS = NTSTATUS(-1071513567i32);
    pub const STATUS_FWP_INVALID_NET_MASK: NTSTATUS = NTSTATUS(-1071513569i32);
    pub const STATUS_FWP_INVALID_PARAMETER: NTSTATUS = NTSTATUS(-1071513547i32);
    pub const STATUS_FWP_INVALID_RANGE: NTSTATUS = NTSTATUS(-1071513568i32);
    pub const STATUS_FWP_INVALID_TRANSFORM_COMBINATION: NTSTATUS = NTSTATUS(-1071513541i32);
    pub const STATUS_FWP_INVALID_TUNNEL_ENDPOINT: NTSTATUS = NTSTATUS(-1071513539i32);
    pub const STATUS_FWP_INVALID_WEIGHT: NTSTATUS = NTSTATUS(-1071513563i32);
    pub const STATUS_FWP_IN_USE: NTSTATUS = NTSTATUS(-1071513590i32);
    pub const STATUS_FWP_KEY_DICTATION_INVALID_KEYING_MATERIAL: NTSTATUS = NTSTATUS(-1071513536i32);
    pub const STATUS_FWP_KEY_DICTATOR_ALREADY_REGISTERED: NTSTATUS = NTSTATUS(-1071513537i32);
    pub const STATUS_FWP_KM_CLIENTS_ONLY: NTSTATUS = NTSTATUS(-1071513579i32);
    pub const STATUS_FWP_L2_DRIVER_NOT_READY: NTSTATUS = NTSTATUS(-1071513538i32);
    pub const STATUS_FWP_LAYER_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513596i32);
    pub const STATUS_FWP_LIFETIME_MISMATCH: NTSTATUS = NTSTATUS(-1071513578i32);
    pub const STATUS_FWP_MATCH_TYPE_MISMATCH: NTSTATUS = NTSTATUS(-1071513562i32);
    pub const STATUS_FWP_NET_EVENTS_DISABLED: NTSTATUS = NTSTATUS(-1071513581i32);
    pub const STATUS_FWP_NEVER_MATCH: NTSTATUS = NTSTATUS(-1071513549i32);
    pub const STATUS_FWP_NOTIFICATION_DROPPED: NTSTATUS = NTSTATUS(-1071513575i32);
    pub const STATUS_FWP_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513592i32);
    pub const STATUS_FWP_NO_TXN_IN_PROGRESS: NTSTATUS = NTSTATUS(-1071513587i32);
    pub const STATUS_FWP_NULL_DISPLAY_NAME: NTSTATUS = NTSTATUS(-1071513565i32);
    pub const STATUS_FWP_NULL_POINTER: NTSTATUS = NTSTATUS(-1071513572i32);
    pub const STATUS_FWP_OUT_OF_BOUNDS: NTSTATUS = NTSTATUS(-1071513560i32);
    pub const STATUS_FWP_PROVIDER_CONTEXT_MISMATCH: NTSTATUS = NTSTATUS(-1071513548i32);
    pub const STATUS_FWP_PROVIDER_CONTEXT_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513594i32);
    pub const STATUS_FWP_PROVIDER_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513595i32);
    pub const STATUS_FWP_RESERVED: NTSTATUS = NTSTATUS(-1071513559i32);
    pub const STATUS_FWP_SESSION_ABORTED: NTSTATUS = NTSTATUS(-1071513584i32);
    pub const STATUS_FWP_STILL_ON: NTSTATUS = NTSTATUS(-1071513533i32);
    pub const STATUS_FWP_SUBLAYER_NOT_FOUND: NTSTATUS = NTSTATUS(-1071513593i32);
    pub const STATUS_FWP_TCPIP_NOT_READY: NTSTATUS = NTSTATUS(-1071513344i32);
    pub const STATUS_FWP_TIMEOUT: NTSTATUS = NTSTATUS(-1071513582i32);
    pub const STATUS_FWP_TOO_MANY_CALLOUTS: NTSTATUS = NTSTATUS(-1071513576i32);
    pub const STATUS_FWP_TOO_MANY_SUBLAYERS: NTSTATUS = NTSTATUS(-1071513546i32);
    pub const STATUS_FWP_TRAFFIC_MISMATCH: NTSTATUS = NTSTATUS(-1071513574i32);
    pub const STATUS_FWP_TXN_ABORTED: NTSTATUS = NTSTATUS(-1071513585i32);
    pub const STATUS_FWP_TXN_IN_PROGRESS: NTSTATUS = NTSTATUS(-1071513586i32);
    pub const STATUS_FWP_TYPE_MISMATCH: NTSTATUS = NTSTATUS(-1071513561i32);
    pub const STATUS_FWP_WRONG_SESSION: NTSTATUS = NTSTATUS(-1071513588i32);
    pub const STATUS_FWP_ZERO_LENGTH_ARRAY: NTSTATUS = NTSTATUS(-1071513566i32);
    pub const STATUS_GDI_HANDLE_LEAK: NTSTATUS = NTSTATUS(-2143354879i32);
    pub const STATUS_GENERIC_COMMAND_FAILED: NTSTATUS = NTSTATUS(-1072365530i32);
    pub const STATUS_GENERIC_NOT_MAPPED: NTSTATUS = NTSTATUS(-1073741594i32);
    pub const STATUS_GHOSTED: NTSTATUS = NTSTATUS(303i32);
    pub const STATUS_GPIO_CLIENT_INFORMATION_INVALID: NTSTATUS = NTSTATUS(-1073700574i32);
    pub const STATUS_GPIO_INCOMPATIBLE_CONNECT_MODE: NTSTATUS = NTSTATUS(-1073700570i32);
    pub const STATUS_GPIO_INTERRUPT_ALREADY_UNMASKED: NTSTATUS = NTSTATUS(-2147442393i32);
    pub const STATUS_GPIO_INVALID_REGISTRATION_PACKET: NTSTATUS = NTSTATUS(-1073700572i32);
    pub const STATUS_GPIO_OPERATION_DENIED: NTSTATUS = NTSTATUS(-1073700571i32);
    pub const STATUS_GPIO_VERSION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700573i32);
    pub const STATUS_GRACEFUL_DISCONNECT: NTSTATUS = NTSTATUS(-1073741257i32);
    pub const STATUS_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED: NTSTATUS = NTSTATUS(-1071774661i32);
    pub const STATUS_GRAPHICS_ADAPTER_CHAIN_NOT_READY: NTSTATUS = NTSTATUS(-1071774669i32);
    pub const STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE: NTSTATUS = NTSTATUS(-1071774936i32);
    pub const STATUS_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET: NTSTATUS = NTSTATUS(-1071774935i32);
    pub const STATUS_GRAPHICS_ADAPTER_WAS_RESET: NTSTATUS = NTSTATUS(-1071775741i32);
    pub const STATUS_GRAPHICS_ALLOCATION_BUSY: NTSTATUS = NTSTATUS(-1071775486i32);
    pub const STATUS_GRAPHICS_ALLOCATION_CLOSED: NTSTATUS = NTSTATUS(-1071775470i32);
    pub const STATUS_GRAPHICS_ALLOCATION_CONTENT_LOST: NTSTATUS = NTSTATUS(-1071775466i32);
    pub const STATUS_GRAPHICS_ALLOCATION_INVALID: NTSTATUS = NTSTATUS(-1071775482i32);
    pub const STATUS_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION: NTSTATUS = NTSTATUS(-1071774886i32);
    pub const STATUS_GRAPHICS_CANNOTCOLORCONVERT: NTSTATUS = NTSTATUS(-1071775736i32);
    pub const STATUS_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN: NTSTATUS = NTSTATUS(-1071774909i32);
    pub const STATUS_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION: NTSTATUS = NTSTATUS(-1071775479i32);
    pub const STATUS_GRAPHICS_CANT_LOCK_MEMORY: NTSTATUS = NTSTATUS(-1071775487i32);
    pub const STATUS_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION: NTSTATUS = NTSTATUS(-1071775471i32);
    pub const STATUS_GRAPHICS_CHAINLINKS_NOT_ENUMERATED: NTSTATUS = NTSTATUS(-1071774670i32);
    pub const STATUS_GRAPHICS_CHAINLINKS_NOT_POWERED_ON: NTSTATUS = NTSTATUS(-1071774667i32);
    pub const STATUS_GRAPHICS_CHAINLINKS_NOT_STARTED: NTSTATUS = NTSTATUS(-1071774668i32);
    pub const STATUS_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774719i32);
    pub const STATUS_GRAPHICS_CLIENTVIDPN_NOT_SET: NTSTATUS = NTSTATUS(-1071774884i32);
    pub const STATUS_GRAPHICS_COPP_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774463i32);
    pub const STATUS_GRAPHICS_DATASET_IS_EMPTY: NTSTATUS = NTSTATUS(1075708747i32);
    pub const STATUS_GRAPHICS_DDCCI_INVALID_CAPABILITIES_STRING: NTSTATUS = NTSTATUS(-1071774329i32);
    pub const STATUS_GRAPHICS_DDCCI_INVALID_DATA: NTSTATUS = NTSTATUS(-1071774331i32);
    pub const STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM: NTSTATUS = NTSTATUS(-1071774325i32);
    pub const STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND: NTSTATUS = NTSTATUS(-1071774327i32);
    pub const STATUS_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH: NTSTATUS = NTSTATUS(-1071774326i32);
    pub const STATUS_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE: NTSTATUS = NTSTATUS(-1071774330i32);
    pub const STATUS_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774332i32);
    pub const STATUS_GRAPHICS_DEPENDABLE_CHILD_STATUS: NTSTATUS = NTSTATUS(1075708988i32);
    pub const STATUS_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP: NTSTATUS = NTSTATUS(-1071774238i32);
    pub const STATUS_GRAPHICS_DRIVER_MISMATCH: NTSTATUS = NTSTATUS(-1071775735i32);
    pub const STATUS_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION: NTSTATUS = NTSTATUS(-1071774939i32);
    pub const STATUS_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET: NTSTATUS = NTSTATUS(-1071774945i32);
    pub const STATUS_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET: NTSTATUS = NTSTATUS(-1071774947i32);
    pub const STATUS_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774904i32);
    pub const STATUS_GRAPHICS_GPU_EXCEPTION_ON_DEVICE: NTSTATUS = NTSTATUS(-1071775232i32);
    pub const STATUS_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST: NTSTATUS = NTSTATUS(-1071774335i32);
    pub const STATUS_GRAPHICS_I2C_ERROR_RECEIVING_DATA: NTSTATUS = NTSTATUS(-1071774333i32);
    pub const STATUS_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA: NTSTATUS = NTSTATUS(-1071774334i32);
    pub const STATUS_GRAPHICS_I2C_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774336i32);
    pub const STATUS_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT: NTSTATUS = NTSTATUS(-1071774891i32);
    pub const STATUS_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE: NTSTATUS = NTSTATUS(-1071774666i32);
    pub const STATUS_GRAPHICS_INDIRECT_DISPLAY_ABANDON_SWAPCHAIN: NTSTATUS = NTSTATUS(-1071775726i32);
    pub const STATUS_GRAPHICS_INDIRECT_DISPLAY_DEVICE_STOPPED: NTSTATUS = NTSTATUS(-1071775725i32);
    pub const STATUS_GRAPHICS_INSUFFICIENT_DMA_BUFFER: NTSTATUS = NTSTATUS(-1071775743i32);
    pub const STATUS_GRAPHICS_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071774233i32);
    pub const STATUS_GRAPHICS_INVALID_ACTIVE_REGION: NTSTATUS = NTSTATUS(-1071774965i32);
    pub const STATUS_GRAPHICS_INVALID_ALLOCATION_HANDLE: NTSTATUS = NTSTATUS(-1071775468i32);
    pub const STATUS_GRAPHICS_INVALID_ALLOCATION_INSTANCE: NTSTATUS = NTSTATUS(-1071775469i32);
    pub const STATUS_GRAPHICS_INVALID_ALLOCATION_USAGE: NTSTATUS = NTSTATUS(-1071775472i32);
    pub const STATUS_GRAPHICS_INVALID_CLIENT_TYPE: NTSTATUS = NTSTATUS(-1071774885i32);
    pub const STATUS_GRAPHICS_INVALID_COLORBASIS: NTSTATUS = NTSTATUS(-1071774914i32);
    pub const STATUS_GRAPHICS_INVALID_COPYPROTECTION_TYPE: NTSTATUS = NTSTATUS(-1071774897i32);
    pub const STATUS_GRAPHICS_INVALID_DISPLAY_ADAPTER: NTSTATUS = NTSTATUS(-1071775742i32);
    pub const STATUS_GRAPHICS_INVALID_DRIVER_MODEL: NTSTATUS = NTSTATUS(-1071775740i32);
    pub const STATUS_GRAPHICS_INVALID_FREQUENCY: NTSTATUS = NTSTATUS(-1071774966i32);
    pub const STATUS_GRAPHICS_INVALID_GAMMA_RAMP: NTSTATUS = NTSTATUS(-1071774905i32);
    pub const STATUS_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM: NTSTATUS = NTSTATUS(-1071774890i32);
    pub const STATUS_GRAPHICS_INVALID_MONITORDESCRIPTOR: NTSTATUS = NTSTATUS(-1071774933i32);
    pub const STATUS_GRAPHICS_INVALID_MONITORDESCRIPTORSET: NTSTATUS = NTSTATUS(-1071774934i32);
    pub const STATUS_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN: NTSTATUS = NTSTATUS(-1071774889i32);
    pub const STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE: NTSTATUS = NTSTATUS(-1071774948i32);
    pub const STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET: NTSTATUS = NTSTATUS(-1071774949i32);
    pub const STATUS_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT: NTSTATUS = NTSTATUS(-1071774888i32);
    pub const STATUS_GRAPHICS_INVALID_MONITOR_SOURCEMODESET: NTSTATUS = NTSTATUS(-1071774943i32);
    pub const STATUS_GRAPHICS_INVALID_MONITOR_SOURCE_MODE: NTSTATUS = NTSTATUS(-1071774942i32);
    pub const STATUS_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION: NTSTATUS = NTSTATUS(-1071774907i32);
    pub const STATUS_GRAPHICS_INVALID_PATH_CONTENT_TYPE: NTSTATUS = NTSTATUS(-1071774898i32);
    pub const STATUS_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL: NTSTATUS = NTSTATUS(-1071774908i32);
    pub const STATUS_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE: NTSTATUS = NTSTATUS(-1071774324i32);
    pub const STATUS_GRAPHICS_INVALID_PIXELFORMAT: NTSTATUS = NTSTATUS(-1071774915i32);
    pub const STATUS_GRAPHICS_INVALID_PIXELVALUEACCESSMODE: NTSTATUS = NTSTATUS(-1071774913i32);
    pub const STATUS_GRAPHICS_INVALID_POINTER: NTSTATUS = NTSTATUS(-1071774236i32);
    pub const STATUS_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE: NTSTATUS = NTSTATUS(-1071774918i32);
    pub const STATUS_GRAPHICS_INVALID_SCANLINE_ORDERING: NTSTATUS = NTSTATUS(-1071774894i32);
    pub const STATUS_GRAPHICS_INVALID_STRIDE: NTSTATUS = NTSTATUS(-1071774916i32);
    pub const STATUS_GRAPHICS_INVALID_TOTAL_REGION: NTSTATUS = NTSTATUS(-1071774964i32);
    pub const STATUS_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET: NTSTATUS = NTSTATUS(-1071774955i32);
    pub const STATUS_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET: NTSTATUS = NTSTATUS(-1071774954i32);
    pub const STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE: NTSTATUS = NTSTATUS(-1071774972i32);
    pub const STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE: NTSTATUS = NTSTATUS(-1071774960i32);
    pub const STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET: NTSTATUS = NTSTATUS(-1071774971i32);
    pub const STATUS_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE: NTSTATUS = NTSTATUS(-1071774959i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN: NTSTATUS = NTSTATUS(-1071774973i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN_PRESENT_PATH: NTSTATUS = NTSTATUS(-1071774951i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN_SOURCEMODESET: NTSTATUS = NTSTATUS(-1071774968i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN_TARGETMODESET: NTSTATUS = NTSTATUS(-1071774967i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE: NTSTATUS = NTSTATUS(-1071774929i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774976i32);
    pub const STATUS_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON: NTSTATUS = NTSTATUS(-1071774899i32);
    pub const STATUS_GRAPHICS_INVALID_VISIBLEREGION_SIZE: NTSTATUS = NTSTATUS(-1071774917i32);
    pub const STATUS_GRAPHICS_LEADLINK_NOT_ENUMERATED: NTSTATUS = NTSTATUS(-1071774671i32);
    pub const STATUS_GRAPHICS_LEADLINK_START_DEFERRED: NTSTATUS = NTSTATUS(1075708983i32);
    pub const STATUS_GRAPHICS_LINK_CONFIGURATION_IN_PROGRESS: NTSTATUS = NTSTATUS(-2145517568i32);
    pub const STATUS_GRAPHICS_MAX_NUM_PATHS_REACHED: NTSTATUS = NTSTATUS(-1071774887i32);
    pub const STATUS_GRAPHICS_MCA_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071774328i32);
    pub const STATUS_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774237i32);
    pub const STATUS_GRAPHICS_MODE_ALREADY_IN_MODESET: NTSTATUS = NTSTATUS(-1071774956i32);
    pub const STATUS_GRAPHICS_MODE_ID_MUST_BE_UNIQUE: NTSTATUS = NTSTATUS(-1071774940i32);
    pub const STATUS_GRAPHICS_MODE_NOT_IN_MODESET: NTSTATUS = NTSTATUS(-1071774902i32);
    pub const STATUS_GRAPHICS_MODE_NOT_PINNED: NTSTATUS = NTSTATUS(1075708679i32);
    pub const STATUS_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET: NTSTATUS = NTSTATUS(-1071774931i32);
    pub const STATUS_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE: NTSTATUS = NTSTATUS(-1071774930i32);
    pub const STATUS_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET: NTSTATUS = NTSTATUS(-1071774932i32);
    pub const STATUS_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER: NTSTATUS = NTSTATUS(-1071774924i32);
    pub const STATUS_GRAPHICS_MONITOR_NOT_CONNECTED: NTSTATUS = NTSTATUS(-1071774920i32);
    pub const STATUS_GRAPHICS_MONITOR_NO_LONGER_EXISTS: NTSTATUS = NTSTATUS(-1071774323i32);
    pub const STATUS_GRAPHICS_MPO_ALLOCATION_UNPINNED: NTSTATUS = NTSTATUS(-1071775720i32);
    pub const STATUS_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774903i32);
    pub const STATUS_GRAPHICS_NOT_A_LINKED_ADAPTER: NTSTATUS = NTSTATUS(-1071774672i32);
    pub const STATUS_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER: NTSTATUS = NTSTATUS(-1071775744i32);
    pub const STATUS_GRAPHICS_NOT_POST_DEVICE_DRIVER: NTSTATUS = NTSTATUS(-1071774664i32);
    pub const STATUS_GRAPHICS_NO_ACTIVE_VIDPN: NTSTATUS = NTSTATUS(-1071774922i32);
    pub const STATUS_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS: NTSTATUS = NTSTATUS(-1071774892i32);
    pub const STATUS_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET: NTSTATUS = NTSTATUS(-1071774925i32);
    pub const STATUS_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME: NTSTATUS = NTSTATUS(-1071774239i32);
    pub const STATUS_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT: NTSTATUS = NTSTATUS(-1071774911i32);
    pub const STATUS_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE: NTSTATUS = NTSTATUS(-1071774235i32);
    pub const STATUS_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET: NTSTATUS = NTSTATUS(1075708748i32);
    pub const STATUS_GRAPHICS_NO_PREFERRED_MODE: NTSTATUS = NTSTATUS(1075708702i32);
    pub const STATUS_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN: NTSTATUS = NTSTATUS(-1071774941i32);
    pub const STATUS_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774950i32);
    pub const STATUS_GRAPHICS_NO_VIDEO_MEMORY: NTSTATUS = NTSTATUS(-1071775488i32);
    pub const STATUS_GRAPHICS_NO_VIDPNMGR: NTSTATUS = NTSTATUS(-1071774923i32);
    pub const STATUS_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED: NTSTATUS = NTSTATUS(-1071774240i32);
    pub const STATUS_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE: NTSTATUS = NTSTATUS(-1071774440i32);
    pub const STATUS_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071774434i32);
    pub const STATUS_GRAPHICS_OPM_HDCP_SRM_NEVER_SET: NTSTATUS = NTSTATUS(-1071774442i32);
    pub const STATUS_GRAPHICS_OPM_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071774453i32);
    pub const STATUS_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST: NTSTATUS = NTSTATUS(-1071774431i32);
    pub const STATUS_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS: NTSTATUS = NTSTATUS(-1071774461i32);
    pub const STATUS_GRAPHICS_OPM_INVALID_HANDLE: NTSTATUS = NTSTATUS(-1071774452i32);
    pub const STATUS_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST: NTSTATUS = NTSTATUS(-1071774435i32);
    pub const STATUS_GRAPHICS_OPM_INVALID_SRM: NTSTATUS = NTSTATUS(-1071774446i32);
    pub const STATUS_GRAPHICS_OPM_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774464i32);
    pub const STATUS_GRAPHICS_OPM_NO_PROTECTED_OUTPUTS_EXIST: NTSTATUS = NTSTATUS(-1071774459i32);
    pub const STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP: NTSTATUS = NTSTATUS(-1071774444i32);
    pub const STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA: NTSTATUS = NTSTATUS(-1071774443i32);
    pub const STATUS_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP: NTSTATUS = NTSTATUS(-1071774445i32);
    pub const STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS: NTSTATUS = NTSTATUS(-1071774436i32);
    pub const STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS: NTSTATUS = NTSTATUS(-1071774433i32);
    pub const STATUS_GRAPHICS_OPM_PROTECTED_OUTPUT_NO_LONGER_EXISTS: NTSTATUS = NTSTATUS(-1071774438i32);
    pub const STATUS_GRAPHICS_OPM_RESOLUTION_TOO_HIGH: NTSTATUS = NTSTATUS(-1071774441i32);
    pub const STATUS_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774432i32);
    pub const STATUS_GRAPHICS_OPM_SPANNING_MODE_ENABLED: NTSTATUS = NTSTATUS(-1071774449i32);
    pub const STATUS_GRAPHICS_OPM_THEATER_MODE_ENABLED: NTSTATUS = NTSTATUS(-1071774448i32);
    pub const STATUS_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL: NTSTATUS = NTSTATUS(-1071774234i32);
    pub const STATUS_GRAPHICS_PARTIAL_DATA_POPULATED: NTSTATUS = NTSTATUS(1075707914i32);
    pub const STATUS_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774957i32);
    pub const STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED: NTSTATUS = NTSTATUS(1075708753i32);
    pub const STATUS_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774906i32);
    pub const STATUS_GRAPHICS_PATH_NOT_IN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774937i32);
    pub const STATUS_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET: NTSTATUS = NTSTATUS(-1071774958i32);
    pub const STATUS_GRAPHICS_POLLING_TOO_FREQUENTLY: NTSTATUS = NTSTATUS(1075708985i32);
    pub const STATUS_GRAPHICS_PRESENT_BUFFER_NOT_BOUND: NTSTATUS = NTSTATUS(-1071775728i32);
    pub const STATUS_GRAPHICS_PRESENT_DENIED: NTSTATUS = NTSTATUS(-1071775737i32);
    pub const STATUS_GRAPHICS_PRESENT_INVALID_WINDOW: NTSTATUS = NTSTATUS(-1071775729i32);
    pub const STATUS_GRAPHICS_PRESENT_MODE_CHANGED: NTSTATUS = NTSTATUS(-1071775739i32);
    pub const STATUS_GRAPHICS_PRESENT_OCCLUDED: NTSTATUS = NTSTATUS(-1071775738i32);
    pub const STATUS_GRAPHICS_PRESENT_REDIRECTION_DISABLED: NTSTATUS = NTSTATUS(-1071775733i32);
    pub const STATUS_GRAPHICS_PRESENT_UNOCCLUDED: NTSTATUS = NTSTATUS(-1071775732i32);
    pub const STATUS_GRAPHICS_PVP_HFS_FAILED: NTSTATUS = NTSTATUS(-1071774447i32);
    pub const STATUS_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH: NTSTATUS = NTSTATUS(-1071774450i32);
    pub const STATUS_GRAPHICS_RESOURCES_NOT_RELATED: NTSTATUS = NTSTATUS(-1071774928i32);
    pub const STATUS_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS: NTSTATUS = NTSTATUS(-1071774232i32);
    pub const STATUS_GRAPHICS_SKIP_ALLOCATION_PREPARATION: NTSTATUS = NTSTATUS(1075708417i32);
    pub const STATUS_GRAPHICS_SOURCE_ALREADY_IN_SET: NTSTATUS = NTSTATUS(-1071774953i32);
    pub const STATUS_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE: NTSTATUS = NTSTATUS(-1071774927i32);
    pub const STATUS_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774919i32);
    pub const STATUS_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED: NTSTATUS = NTSTATUS(-1071774720i32);
    pub const STATUS_GRAPHICS_STALE_MODESET: NTSTATUS = NTSTATUS(-1071774944i32);
    pub const STATUS_GRAPHICS_STALE_VIDPN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774921i32);
    pub const STATUS_GRAPHICS_START_DEFERRED: NTSTATUS = NTSTATUS(1075708986i32);
    pub const STATUS_GRAPHICS_TARGET_ALREADY_IN_SET: NTSTATUS = NTSTATUS(-1071774952i32);
    pub const STATUS_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE: NTSTATUS = NTSTATUS(-1071774926i32);
    pub const STATUS_GRAPHICS_TARGET_NOT_IN_TOPOLOGY: NTSTATUS = NTSTATUS(-1071774912i32);
    pub const STATUS_GRAPHICS_TOO_MANY_REFERENCES: NTSTATUS = NTSTATUS(-1071775485i32);
    pub const STATUS_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1071774893i32);
    pub const STATUS_GRAPHICS_TRY_AGAIN_LATER: NTSTATUS = NTSTATUS(-1071775484i32);
    pub const STATUS_GRAPHICS_TRY_AGAIN_NOW: NTSTATUS = NTSTATUS(-1071775483i32);
    pub const STATUS_GRAPHICS_UAB_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774462i32);
    pub const STATUS_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1071774896i32);
    pub const STATUS_GRAPHICS_UNKNOWN_CHILD_STATUS: NTSTATUS = NTSTATUS(1075708975i32);
    pub const STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE: NTSTATUS = NTSTATUS(-1071775481i32);
    pub const STATUS_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED: NTSTATUS = NTSTATUS(-1071775480i32);
    pub const STATUS_GRAPHICS_VAIL_STATE_CHANGED: NTSTATUS = NTSTATUS(-1071775727i32);
    pub const STATUS_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES: NTSTATUS = NTSTATUS(-1071774938i32);
    pub const STATUS_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774970i32);
    pub const STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE: NTSTATUS = NTSTATUS(-1071774910i32);
    pub const STATUS_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774974i32);
    pub const STATUS_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071774975i32);
    pub const STATUS_GRAPHICS_WINDOWDC_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1071775731i32);
    pub const STATUS_GRAPHICS_WINDOWLESS_PRESENT_DISABLED: NTSTATUS = NTSTATUS(-1071775730i32);
    pub const STATUS_GRAPHICS_WRONG_ALLOCATION_DEVICE: NTSTATUS = NTSTATUS(-1071775467i32);
    pub const STATUS_GROUP_EXISTS: NTSTATUS = NTSTATUS(-1073741723i32);
    pub const STATUS_GUARD_PAGE_VIOLATION: NTSTATUS = NTSTATUS(-2147483647i32);
    pub const STATUS_GUIDS_EXHAUSTED: NTSTATUS = NTSTATUS(-1073741693i32);
    pub const STATUS_GUID_SUBSTITUTION_MADE: NTSTATUS = NTSTATUS(-2147483636i32);
    pub const STATUS_HANDLES_CLOSED: NTSTATUS = NTSTATUS(-2147483638i32);
    pub const STATUS_HANDLE_NOT_CLOSABLE: NTSTATUS = NTSTATUS(-1073741259i32);
    pub const STATUS_HANDLE_NO_LONGER_VALID: NTSTATUS = NTSTATUS(-1072103384i32);
    pub const STATUS_HANDLE_REVOKED: NTSTATUS = NTSTATUS(-1073700858i32);
    pub const STATUS_HARDWARE_MEMORY_ERROR: NTSTATUS = NTSTATUS(-1073740023i32);
    pub const STATUS_HASH_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073700607i32);
    pub const STATUS_HASH_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700608i32);
    pub const STATUS_HAS_SYSTEM_CRITICAL_FILES: NTSTATUS = NTSTATUS(-1073740611i32);
    pub const STATUS_HDAUDIO_CONNECTION_LIST_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1069285373i32);
    pub const STATUS_HDAUDIO_EMPTY_CONNECTION_LIST: NTSTATUS = NTSTATUS(-1069285374i32);
    pub const STATUS_HDAUDIO_NO_LOGICAL_DEVICES_CREATED: NTSTATUS = NTSTATUS(-1069285372i32);
    pub const STATUS_HDAUDIO_NULL_LINKED_LIST_ENTRY: NTSTATUS = NTSTATUS(-1069285371i32);
    pub const STATUS_HEAP_CORRUPTION: NTSTATUS = NTSTATUS(-1073740940i32);
    pub const STATUS_HEURISTIC_DAMAGE_POSSIBLE: NTSTATUS = NTSTATUS(1075380225i32);
    pub const STATUS_HIBERNATED: NTSTATUS = NTSTATUS(1073741866i32);
    pub const STATUS_HIBERNATION_FAILURE: NTSTATUS = NTSTATUS(-1073740783i32);
    pub const STATUS_HIVE_UNLOADED: NTSTATUS = NTSTATUS(-1073740763i32);
    pub const STATUS_HMAC_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700863i32);
    pub const STATUS_HOPLIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073700846i32);
    pub const STATUS_HOST_DOWN: NTSTATUS = NTSTATUS(-1073740976i32);
    pub const STATUS_HOST_UNREACHABLE: NTSTATUS = NTSTATUS(-1073741251i32);
    pub const STATUS_HUNG_DISPLAY_DRIVER_THREAD: NTSTATUS = NTSTATUS(-1073740779i32);
    pub const STATUS_HV_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1070268410i32);
    pub const STATUS_HV_ACKNOWLEDGED: NTSTATUS = NTSTATUS(-1070268394i32);
    pub const STATUS_HV_CALL_PENDING: NTSTATUS = NTSTATUS(-1070268295i32);
    pub const STATUS_HV_CPUID_FEATURE_VALIDATION_ERROR: NTSTATUS = NTSTATUS(-1070268356i32);
    pub const STATUS_HV_CPUID_XSAVE_FEATURE_VALIDATION_ERROR: NTSTATUS = NTSTATUS(-1070268355i32);
    pub const STATUS_HV_DEVICE_NOT_IN_DOMAIN: NTSTATUS = NTSTATUS(-1070268298i32);
    pub const STATUS_HV_EVENT_BUFFER_ALREADY_FREED: NTSTATUS = NTSTATUS(-1070268300i32);
    pub const STATUS_HV_FEATURE_UNAVAILABLE: NTSTATUS = NTSTATUS(-1070268386i32);
    pub const STATUS_HV_INACTIVE: NTSTATUS = NTSTATUS(-1070268388i32);
    pub const STATUS_HV_INSUFFICIENT_BUFFER: NTSTATUS = NTSTATUS(-1070268365i32);
    pub const STATUS_HV_INSUFFICIENT_BUFFERS: NTSTATUS = NTSTATUS(-1070268397i32);
    pub const STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY: NTSTATUS = NTSTATUS(-1070268299i32);
    pub const STATUS_HV_INSUFFICIENT_CONTIGUOUS_MEMORY_MIRRORING: NTSTATUS = NTSTATUS(-1070268286i32);
    pub const STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY: NTSTATUS = NTSTATUS(-1070268285i32);
    pub const STATUS_HV_INSUFFICIENT_CONTIGUOUS_ROOT_MEMORY_MIRRORING: NTSTATUS = NTSTATUS(-1070268283i32);
    pub const STATUS_HV_INSUFFICIENT_DEVICE_DOMAINS: NTSTATUS = NTSTATUS(-1070268360i32);
    pub const STATUS_HV_INSUFFICIENT_MEMORY: NTSTATUS = NTSTATUS(-1070268405i32);
    pub const STATUS_HV_INSUFFICIENT_MEMORY_MIRRORING: NTSTATUS = NTSTATUS(-1070268287i32);
    pub const STATUS_HV_INSUFFICIENT_ROOT_MEMORY: NTSTATUS = NTSTATUS(-1070268301i32);
    pub const STATUS_HV_INSUFFICIENT_ROOT_MEMORY_MIRRORING: NTSTATUS = NTSTATUS(-1070268284i32);
    pub const STATUS_HV_INVALID_ALIGNMENT: NTSTATUS = NTSTATUS(-1070268412i32);
    pub const STATUS_HV_INVALID_CONNECTION_ID: NTSTATUS = NTSTATUS(-1070268398i32);
    pub const STATUS_HV_INVALID_CPU_GROUP_ID: NTSTATUS = NTSTATUS(-1070268305i32);
    pub const STATUS_HV_INVALID_CPU_GROUP_STATE: NTSTATUS = NTSTATUS(-1070268304i32);
    pub const STATUS_HV_INVALID_DEVICE_ID: NTSTATUS = NTSTATUS(-1070268329i32);
    pub const STATUS_HV_INVALID_DEVICE_STATE: NTSTATUS = NTSTATUS(-1070268328i32);
    pub const STATUS_HV_INVALID_HYPERCALL_CODE: NTSTATUS = NTSTATUS(-1070268414i32);
    pub const STATUS_HV_INVALID_HYPERCALL_INPUT: NTSTATUS = NTSTATUS(-1070268413i32);
    pub const STATUS_HV_INVALID_LP_INDEX: NTSTATUS = NTSTATUS(-1070268351i32);
    pub const STATUS_HV_INVALID_PARAMETER: NTSTATUS = NTSTATUS(-1070268411i32);
    pub const STATUS_HV_INVALID_PARTITION_ID: NTSTATUS = NTSTATUS(-1070268403i32);
    pub const STATUS_HV_INVALID_PARTITION_STATE: NTSTATUS = NTSTATUS(-1070268409i32);
    pub const STATUS_HV_INVALID_PORT_ID: NTSTATUS = NTSTATUS(-1070268399i32);
    pub const STATUS_HV_INVALID_PROXIMITY_DOMAIN_INFO: NTSTATUS = NTSTATUS(-1070268390i32);
    pub const STATUS_HV_INVALID_REGISTER_VALUE: NTSTATUS = NTSTATUS(-1070268336i32);
    pub const STATUS_HV_INVALID_SAVE_RESTORE_STATE: NTSTATUS = NTSTATUS(-1070268393i32);
    pub const STATUS_HV_INVALID_SYNIC_STATE: NTSTATUS = NTSTATUS(-1070268392i32);
    pub const STATUS_HV_INVALID_VP_INDEX: NTSTATUS = NTSTATUS(-1070268402i32);
    pub const STATUS_HV_INVALID_VP_STATE: NTSTATUS = NTSTATUS(-1070268395i32);
    pub const STATUS_HV_INVALID_VTL_STATE: NTSTATUS = NTSTATUS(-1070268335i32);
    pub const STATUS_HV_MSR_ACCESS_FAILED: NTSTATUS = NTSTATUS(-1070268288i32);
    pub const STATUS_HV_NESTED_VM_EXIT: NTSTATUS = NTSTATUS(-1070268297i32);
    pub const STATUS_HV_NOT_ACKNOWLEDGED: NTSTATUS = NTSTATUS(-1070268396i32);
    pub const STATUS_HV_NOT_ALLOWED_WITH_NESTED_VIRT_ACTIVE: NTSTATUS = NTSTATUS(-1070268302i32);
    pub const STATUS_HV_NOT_PRESENT: NTSTATUS = NTSTATUS(-1070264320i32);
    pub const STATUS_HV_NO_DATA: NTSTATUS = NTSTATUS(-1070268389i32);
    pub const STATUS_HV_NO_RESOURCES: NTSTATUS = NTSTATUS(-1070268387i32);
    pub const STATUS_HV_NX_NOT_DETECTED: NTSTATUS = NTSTATUS(-1070268331i32);
    pub const STATUS_HV_OBJECT_IN_USE: NTSTATUS = NTSTATUS(-1070268391i32);
    pub const STATUS_HV_OPERATION_DENIED: NTSTATUS = NTSTATUS(-1070268408i32);
    pub const STATUS_HV_OPERATION_FAILED: NTSTATUS = NTSTATUS(-1070268303i32);
    pub const STATUS_HV_PAGE_REQUEST_INVALID: NTSTATUS = NTSTATUS(-1070268320i32);
    pub const STATUS_HV_PARTITION_TOO_DEEP: NTSTATUS = NTSTATUS(-1070268404i32);
    pub const STATUS_HV_PENDING_PAGE_REQUESTS: NTSTATUS = NTSTATUS(3473497i32);
    pub const STATUS_HV_PROCESSOR_STARTUP_TIMEOUT: NTSTATUS = NTSTATUS(-1070268354i32);
    pub const STATUS_HV_PROPERTY_VALUE_OUT_OF_RANGE: NTSTATUS = NTSTATUS(-1070268406i32);
    pub const STATUS_HV_SMX_ENABLED: NTSTATUS = NTSTATUS(-1070268353i32);
    pub const STATUS_HV_UNKNOWN_PROPERTY: NTSTATUS = NTSTATUS(-1070268407i32);
    pub const STATUS_ILLEGAL_CHARACTER: NTSTATUS = NTSTATUS(-1073741471i32);
    pub const STATUS_ILLEGAL_DLL_RELOCATION: NTSTATUS = NTSTATUS(-1073741207i32);
    pub const STATUS_ILLEGAL_ELEMENT_ADDRESS: NTSTATUS = NTSTATUS(-1073741179i32);
    pub const STATUS_ILLEGAL_FLOAT_CONTEXT: NTSTATUS = NTSTATUS(-1073741494i32);
    pub const STATUS_ILLEGAL_FUNCTION: NTSTATUS = NTSTATUS(-1073741649i32);
    pub const STATUS_ILLEGAL_INSTRUCTION: NTSTATUS = NTSTATUS(-1073741795i32);
    pub const STATUS_ILL_FORMED_PASSWORD: NTSTATUS = NTSTATUS(-1073741717i32);
    pub const STATUS_ILL_FORMED_SERVICE_ENTRY: NTSTATUS = NTSTATUS(-1073741472i32);
    pub const STATUS_IMAGE_ALREADY_LOADED: NTSTATUS = NTSTATUS(-1073741554i32);
    pub const STATUS_IMAGE_ALREADY_LOADED_AS_DLL: NTSTATUS = NTSTATUS(-1073741411i32);
    pub const STATUS_IMAGE_AT_DIFFERENT_BASE: NTSTATUS = NTSTATUS(1073741878i32);
    pub const STATUS_IMAGE_CERT_EXPIRED: NTSTATUS = NTSTATUS(-1073740283i32);
    pub const STATUS_IMAGE_CERT_REVOKED: NTSTATUS = NTSTATUS(-1073740285i32);
    pub const STATUS_IMAGE_CHECKSUM_MISMATCH: NTSTATUS = NTSTATUS(-1073741279i32);
    pub const STATUS_IMAGE_LOADED_AS_PATCH_IMAGE: NTSTATUS = NTSTATUS(-1073740608i32);
    pub const STATUS_IMAGE_MACHINE_TYPE_MISMATCH: NTSTATUS = NTSTATUS(1073741838i32);
    pub const STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE: NTSTATUS = NTSTATUS(1073741859i32);
    pub const STATUS_IMAGE_MP_UP_MISMATCH: NTSTATUS = NTSTATUS(-1073741239i32);
    pub const STATUS_IMAGE_NOT_AT_BASE: NTSTATUS = NTSTATUS(1073741827i32);
    pub const STATUS_IMAGE_SUBSYSTEM_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073741405i32);
    pub const STATUS_IMPLEMENTATION_LIMIT: NTSTATUS = NTSTATUS(-1073740757i32);
    pub const STATUS_INCOMPATIBLE_DRIVER_BLOCKED: NTSTATUS = NTSTATUS(-1073740764i32);
    pub const STATUS_INCOMPATIBLE_FILE_MAP: NTSTATUS = NTSTATUS(-1073741747i32);
    pub const STATUS_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING: NTSTATUS = NTSTATUS(-1073741410i32);
    pub const STATUS_INCORRECT_ACCOUNT_TYPE: NTSTATUS = NTSTATUS(-1073700727i32);
    pub const STATUS_INDEX_OUT_OF_BOUNDS: NTSTATUS = NTSTATUS(-1073740591i32);
    pub const STATUS_INDOUBT_TRANSACTIONS_EXIST: NTSTATUS = NTSTATUS(-1072103366i32);
    pub const STATUS_INFO_LENGTH_MISMATCH: NTSTATUS = NTSTATUS(-1073741820i32);
    pub const STATUS_INSTANCE_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073741653i32);
    pub const STATUS_INSTRUCTION_MISALIGNMENT: NTSTATUS = NTSTATUS(-1073741654i32);
    pub const STATUS_INSUFFICIENT_LOGON_INFO: NTSTATUS = NTSTATUS(-1073741232i32);
    pub const STATUS_INSUFFICIENT_NVRAM_RESOURCES: NTSTATUS = NTSTATUS(-1073740716i32);
    pub const STATUS_INSUFFICIENT_POWER: NTSTATUS = NTSTATUS(-1073741090i32);
    pub const STATUS_INSUFFICIENT_RESOURCES: NTSTATUS = NTSTATUS(-1073741670i32);
    pub const STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE: NTSTATUS = NTSTATUS(-1073740778i32);
    pub const STATUS_INSUFFICIENT_VIRTUAL_ADDR_RESOURCES: NTSTATUS = NTSTATUS(-1073740606i32);
    pub const STATUS_INSUFF_SERVER_RESOURCES: NTSTATUS = NTSTATUS(-1073741307i32);
    pub const STATUS_INTEGER_DIVIDE_BY_ZERO: NTSTATUS = NTSTATUS(-1073741676i32);
    pub const STATUS_INTEGER_OVERFLOW: NTSTATUS = NTSTATUS(-1073741675i32);
    pub const STATUS_INTERMIXED_KERNEL_EA_OPERATION: NTSTATUS = NTSTATUS(-1073740687i32);
    pub const STATUS_INTERNAL_DB_CORRUPTION: NTSTATUS = NTSTATUS(-1073741596i32);
    pub const STATUS_INTERNAL_DB_ERROR: NTSTATUS = NTSTATUS(-1073741480i32);
    pub const STATUS_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1073741595i32);
    pub const STATUS_INTERRUPTED: NTSTATUS = NTSTATUS(-1073740523i32);
    pub const STATUS_INTERRUPT_STILL_CONNECTED: NTSTATUS = NTSTATUS(296i32);
    pub const STATUS_INTERRUPT_VECTOR_ALREADY_CONNECTED: NTSTATUS = NTSTATUS(295i32);
    pub const STATUS_INVALID_ACCOUNT_NAME: NTSTATUS = NTSTATUS(-1073741726i32);
    pub const STATUS_INVALID_ACE_CONDITION: NTSTATUS = NTSTATUS(-1073741406i32);
    pub const STATUS_INVALID_ACL: NTSTATUS = NTSTATUS(-1073741705i32);
    pub const STATUS_INVALID_ADDRESS: NTSTATUS = NTSTATUS(-1073741503i32);
    pub const STATUS_INVALID_ADDRESS_COMPONENT: NTSTATUS = NTSTATUS(-1073741305i32);
    pub const STATUS_INVALID_ADDRESS_WILDCARD: NTSTATUS = NTSTATUS(-1073741304i32);
    pub const STATUS_INVALID_BLOCK_LENGTH: NTSTATUS = NTSTATUS(-1073741453i32);
    pub const STATUS_INVALID_BUFFER_SIZE: NTSTATUS = NTSTATUS(-1073741306i32);
    pub const STATUS_INVALID_CAP: NTSTATUS = NTSTATUS(-1073740539i32);
    pub const STATUS_INVALID_CID: NTSTATUS = NTSTATUS(-1073741813i32);
    pub const STATUS_INVALID_COMPUTER_NAME: NTSTATUS = NTSTATUS(-1073741534i32);
    pub const STATUS_INVALID_CONNECTION: NTSTATUS = NTSTATUS(-1073741504i32);
    pub const STATUS_INVALID_CRUNTIME_PARAMETER: NTSTATUS = NTSTATUS(-1073740777i32);
    pub const STATUS_INVALID_DEVICE_OBJECT_PARAMETER: NTSTATUS = NTSTATUS(-1073740951i32);
    pub const STATUS_INVALID_DEVICE_REQUEST: NTSTATUS = NTSTATUS(-1073741808i32);
    pub const STATUS_INVALID_DEVICE_STATE: NTSTATUS = NTSTATUS(-1073741436i32);
    pub const STATUS_INVALID_DISPOSITION: NTSTATUS = NTSTATUS(-1073741786i32);
    pub const STATUS_INVALID_DOMAIN_ROLE: NTSTATUS = NTSTATUS(-1073741602i32);
    pub const STATUS_INVALID_DOMAIN_STATE: NTSTATUS = NTSTATUS(-1073741603i32);
    pub const STATUS_INVALID_EA_FLAG: NTSTATUS = NTSTATUS(-2147483627i32);
    pub const STATUS_INVALID_EA_NAME: NTSTATUS = NTSTATUS(-2147483629i32);
    pub const STATUS_INVALID_EXCEPTION_HANDLER: NTSTATUS = NTSTATUS(-1073741403i32);
    pub const STATUS_INVALID_FIELD_IN_PARAMETER_LIST: NTSTATUS = NTSTATUS(-1073740683i32);
    pub const STATUS_INVALID_FILE_FOR_SECTION: NTSTATUS = NTSTATUS(-1073741792i32);
    pub const STATUS_INVALID_GROUP_ATTRIBUTES: NTSTATUS = NTSTATUS(-1073741660i32);
    pub const STATUS_INVALID_HANDLE: NTSTATUS = NTSTATUS(-1073741816i32);
    pub const STATUS_INVALID_HW_PROFILE: NTSTATUS = NTSTATUS(-1073741216i32);
    pub const STATUS_INVALID_IDN_NORMALIZATION: NTSTATUS = NTSTATUS(-1073740010i32);
    pub const STATUS_INVALID_ID_AUTHORITY: NTSTATUS = NTSTATUS(-1073741692i32);
    pub const STATUS_INVALID_IMAGE_FORMAT: NTSTATUS = NTSTATUS(-1073741701i32);
    pub const STATUS_INVALID_IMAGE_HASH: NTSTATUS = NTSTATUS(-1073740760i32);
    pub const STATUS_INVALID_IMAGE_LE_FORMAT: NTSTATUS = NTSTATUS(-1073741522i32);
    pub const STATUS_INVALID_IMAGE_NE_FORMAT: NTSTATUS = NTSTATUS(-1073741541i32);
    pub const STATUS_INVALID_IMAGE_NOT_MZ: NTSTATUS = NTSTATUS(-1073741521i32);
    pub const STATUS_INVALID_IMAGE_PROTECT: NTSTATUS = NTSTATUS(-1073741520i32);
    pub const STATUS_INVALID_IMAGE_WIN_16: NTSTATUS = NTSTATUS(-1073741519i32);
    pub const STATUS_INVALID_IMAGE_WIN_32: NTSTATUS = NTSTATUS(-1073740967i32);
    pub const STATUS_INVALID_IMAGE_WIN_64: NTSTATUS = NTSTATUS(-1073740966i32);
    pub const STATUS_INVALID_IMPORT_OF_NON_DLL: NTSTATUS = NTSTATUS(-1073740945i32);
    pub const STATUS_INVALID_INFO_CLASS: NTSTATUS = NTSTATUS(-1073741821i32);
    pub const STATUS_INVALID_INITIATOR_TARGET_PATH: NTSTATUS = NTSTATUS(-1073740681i32);
    pub const STATUS_INVALID_KERNEL_INFO_VERSION: NTSTATUS = NTSTATUS(-1073700860i32);
    pub const STATUS_INVALID_LABEL: NTSTATUS = NTSTATUS(-1073740730i32);
    pub const STATUS_INVALID_LDT_DESCRIPTOR: NTSTATUS = NTSTATUS(-1073741542i32);
    pub const STATUS_INVALID_LDT_OFFSET: NTSTATUS = NTSTATUS(-1073741543i32);
    pub const STATUS_INVALID_LDT_SIZE: NTSTATUS = NTSTATUS(-1073741544i32);
    pub const STATUS_INVALID_LEVEL: NTSTATUS = NTSTATUS(-1073741496i32);
    pub const STATUS_INVALID_LOCK_RANGE: NTSTATUS = NTSTATUS(-1073741407i32);
    pub const STATUS_INVALID_LOCK_SEQUENCE: NTSTATUS = NTSTATUS(-1073741794i32);
    pub const STATUS_INVALID_LOGON_HOURS: NTSTATUS = NTSTATUS(-1073741713i32);
    pub const STATUS_INVALID_LOGON_TYPE: NTSTATUS = NTSTATUS(-1073741557i32);
    pub const STATUS_INVALID_MEMBER: NTSTATUS = NTSTATUS(-1073741445i32);
    pub const STATUS_INVALID_MESSAGE: NTSTATUS = NTSTATUS(-1073740030i32);
    pub const STATUS_INVALID_NETWORK_RESPONSE: NTSTATUS = NTSTATUS(-1073741629i32);
    pub const STATUS_INVALID_OFFSET_ALIGNMENT: NTSTATUS = NTSTATUS(-1073740684i32);
    pub const STATUS_INVALID_OPLOCK_PROTOCOL: NTSTATUS = NTSTATUS(-1073741597i32);
    pub const STATUS_INVALID_OWNER: NTSTATUS = NTSTATUS(-1073741734i32);
    pub const STATUS_INVALID_PACKAGE_SID_LENGTH: NTSTATUS = NTSTATUS(-1073700350i32);
    pub const STATUS_INVALID_PAGE_PROTECTION: NTSTATUS = NTSTATUS(-1073741755i32);
    pub const STATUS_INVALID_PARAMETER: NTSTATUS = NTSTATUS(-1073741811i32);
    pub const STATUS_INVALID_PARAMETER_1: NTSTATUS = NTSTATUS(-1073741585i32);
    pub const STATUS_INVALID_PARAMETER_10: NTSTATUS = NTSTATUS(-1073741576i32);
    pub const STATUS_INVALID_PARAMETER_11: NTSTATUS = NTSTATUS(-1073741575i32);
    pub const STATUS_INVALID_PARAMETER_12: NTSTATUS = NTSTATUS(-1073741574i32);
    pub const STATUS_INVALID_PARAMETER_2: NTSTATUS = NTSTATUS(-1073741584i32);
    pub const STATUS_INVALID_PARAMETER_3: NTSTATUS = NTSTATUS(-1073741583i32);
    pub const STATUS_INVALID_PARAMETER_4: NTSTATUS = NTSTATUS(-1073741582i32);
    pub const STATUS_INVALID_PARAMETER_5: NTSTATUS = NTSTATUS(-1073741581i32);
    pub const STATUS_INVALID_PARAMETER_6: NTSTATUS = NTSTATUS(-1073741580i32);
    pub const STATUS_INVALID_PARAMETER_7: NTSTATUS = NTSTATUS(-1073741579i32);
    pub const STATUS_INVALID_PARAMETER_8: NTSTATUS = NTSTATUS(-1073741578i32);
    pub const STATUS_INVALID_PARAMETER_9: NTSTATUS = NTSTATUS(-1073741577i32);
    pub const STATUS_INVALID_PARAMETER_MIX: NTSTATUS = NTSTATUS(-1073741776i32);
    pub const STATUS_INVALID_PEP_INFO_VERSION: NTSTATUS = NTSTATUS(-1073700859i32);
    pub const STATUS_INVALID_PIPE_STATE: NTSTATUS = NTSTATUS(-1073741651i32);
    pub const STATUS_INVALID_PLUGPLAY_DEVICE_PATH: NTSTATUS = NTSTATUS(-1073741215i32);
    pub const STATUS_INVALID_PORT_ATTRIBUTES: NTSTATUS = NTSTATUS(-1073741778i32);
    pub const STATUS_INVALID_PORT_HANDLE: NTSTATUS = NTSTATUS(-1073741758i32);
    pub const STATUS_INVALID_PRIMARY_GROUP: NTSTATUS = NTSTATUS(-1073741733i32);
    pub const STATUS_INVALID_QUOTA_LOWER: NTSTATUS = NTSTATUS(-1073741775i32);
    pub const STATUS_INVALID_READ_MODE: NTSTATUS = NTSTATUS(-1073741644i32);
    pub const STATUS_INVALID_RUNLEVEL_SETTING: NTSTATUS = NTSTATUS(-1073700542i32);
    pub const STATUS_INVALID_SECURITY_DESCR: NTSTATUS = NTSTATUS(-1073741703i32);
    pub const STATUS_INVALID_SERVER_STATE: NTSTATUS = NTSTATUS(-1073741604i32);
    pub const STATUS_INVALID_SESSION: NTSTATUS = NTSTATUS(-1073740715i32);
    pub const STATUS_INVALID_SID: NTSTATUS = NTSTATUS(-1073741704i32);
    pub const STATUS_INVALID_SIGNATURE: NTSTATUS = NTSTATUS(-1073700864i32);
    pub const STATUS_INVALID_STATE_TRANSITION: NTSTATUS = NTSTATUS(-1073700861i32);
    pub const STATUS_INVALID_SUB_AUTHORITY: NTSTATUS = NTSTATUS(-1073741706i32);
    pub const STATUS_INVALID_SYSTEM_SERVICE: NTSTATUS = NTSTATUS(-1073741796i32);
    pub const STATUS_INVALID_TASK_INDEX: NTSTATUS = NTSTATUS(-1073740543i32);
    pub const STATUS_INVALID_TASK_NAME: NTSTATUS = NTSTATUS(-1073740544i32);
    pub const STATUS_INVALID_THREAD: NTSTATUS = NTSTATUS(-1073740004i32);
    pub const STATUS_INVALID_TOKEN: NTSTATUS = NTSTATUS(-1073740699i32);
    pub const STATUS_INVALID_TRANSACTION: NTSTATUS = NTSTATUS(-1072103422i32);
    pub const STATUS_INVALID_UNWIND_TARGET: NTSTATUS = NTSTATUS(-1073741783i32);
    pub const STATUS_INVALID_USER_BUFFER: NTSTATUS = NTSTATUS(-1073741592i32);
    pub const STATUS_INVALID_USER_PRINCIPAL_NAME: NTSTATUS = NTSTATUS(-1073740772i32);
    pub const STATUS_INVALID_VARIANT: NTSTATUS = NTSTATUS(-1073741262i32);
    pub const STATUS_INVALID_VIEW_SIZE: NTSTATUS = NTSTATUS(-1073741793i32);
    pub const STATUS_INVALID_VOLUME_LABEL: NTSTATUS = NTSTATUS(-1073741690i32);
    pub const STATUS_INVALID_WEIGHT: NTSTATUS = NTSTATUS(-1073740712i32);
    pub const STATUS_INVALID_WORKSTATION: NTSTATUS = NTSTATUS(-1073741712i32);
    pub const STATUS_IN_PAGE_ERROR: NTSTATUS = NTSTATUS(-1073741818i32);
    pub const STATUS_IORING_COMPLETION_QUEUE_TOO_BIG: NTSTATUS = NTSTATUS(-1069154299i32);
    pub const STATUS_IORING_COMPLETION_QUEUE_TOO_FULL: NTSTATUS = NTSTATUS(-1069154296i32);
    pub const STATUS_IORING_CORRUPT: NTSTATUS = NTSTATUS(-1069154297i32);
    pub const STATUS_IORING_REQUIRED_FLAG_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1069154303i32);
    pub const STATUS_IORING_SUBMISSION_QUEUE_FULL: NTSTATUS = NTSTATUS(-1069154302i32);
    pub const STATUS_IORING_SUBMISSION_QUEUE_TOO_BIG: NTSTATUS = NTSTATUS(-1069154300i32);
    pub const STATUS_IORING_SUBMIT_IN_PROGRESS: NTSTATUS = NTSTATUS(-1069154298i32);
    pub const STATUS_IORING_VERSION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1069154301i32);
    pub const STATUS_IO_DEVICE_ERROR: NTSTATUS = NTSTATUS(-1073741435i32);
    pub const STATUS_IO_DEVICE_INVALID_DATA: NTSTATUS = NTSTATUS(-1073741392i32);
    pub const STATUS_IO_OPERATION_TIMEOUT: NTSTATUS = NTSTATUS(-1073740675i32);
    pub const STATUS_IO_PREEMPTED: NTSTATUS = NTSTATUS(-1068433407i32);
    pub const STATUS_IO_PRIVILEGE_FAILED: NTSTATUS = NTSTATUS(-1073741513i32);
    pub const STATUS_IO_REISSUE_AS_CACHED: NTSTATUS = NTSTATUS(-1073479623i32);
    pub const STATUS_IO_REPARSE_DATA_INVALID: NTSTATUS = NTSTATUS(-1073741192i32);
    pub const STATUS_IO_REPARSE_TAG_INVALID: NTSTATUS = NTSTATUS(-1073741194i32);
    pub const STATUS_IO_REPARSE_TAG_MISMATCH: NTSTATUS = NTSTATUS(-1073741193i32);
    pub const STATUS_IO_REPARSE_TAG_NOT_HANDLED: NTSTATUS = NTSTATUS(-1073741191i32);
    pub const STATUS_IO_TIMEOUT: NTSTATUS = NTSTATUS(-1073741643i32);
    pub const STATUS_IO_UNALIGNED_WRITE: NTSTATUS = NTSTATUS(-1073741391i32);
    pub const STATUS_IPSEC_AUTH_FIREWALL_DROP: NTSTATUS = NTSTATUS(-1070202872i32);
    pub const STATUS_IPSEC_BAD_SPI: NTSTATUS = NTSTATUS(-1070202879i32);
    pub const STATUS_IPSEC_CLEAR_TEXT_DROP: NTSTATUS = NTSTATUS(-1070202873i32);
    pub const STATUS_IPSEC_DOSP_BLOCK: NTSTATUS = NTSTATUS(-1070170112i32);
    pub const STATUS_IPSEC_DOSP_INVALID_PACKET: NTSTATUS = NTSTATUS(-1070170110i32);
    pub const STATUS_IPSEC_DOSP_KEYMOD_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1070170107i32);
    pub const STATUS_IPSEC_DOSP_MAX_ENTRIES: NTSTATUS = NTSTATUS(-1070170108i32);
    pub const STATUS_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES: NTSTATUS = NTSTATUS(-1070170106i32);
    pub const STATUS_IPSEC_DOSP_RECEIVED_MULTICAST: NTSTATUS = NTSTATUS(-1070170111i32);
    pub const STATUS_IPSEC_DOSP_STATE_LOOKUP_FAILED: NTSTATUS = NTSTATUS(-1070170109i32);
    pub const STATUS_IPSEC_INTEGRITY_CHECK_FAILED: NTSTATUS = NTSTATUS(-1070202874i32);
    pub const STATUS_IPSEC_INVALID_PACKET: NTSTATUS = NTSTATUS(-1070202875i32);
    pub const STATUS_IPSEC_QUEUE_OVERFLOW: NTSTATUS = NTSTATUS(-1073700848i32);
    pub const STATUS_IPSEC_REPLAY_CHECK_FAILED: NTSTATUS = NTSTATUS(-1070202876i32);
    pub const STATUS_IPSEC_SA_LIFETIME_EXPIRED: NTSTATUS = NTSTATUS(-1070202878i32);
    pub const STATUS_IPSEC_THROTTLE_DROP: NTSTATUS = NTSTATUS(-1070202871i32);
    pub const STATUS_IPSEC_WRONG_SA: NTSTATUS = NTSTATUS(-1070202877i32);
    pub const STATUS_IP_ADDRESS_CONFLICT1: NTSTATUS = NTSTATUS(-1073741228i32);
    pub const STATUS_IP_ADDRESS_CONFLICT2: NTSTATUS = NTSTATUS(-1073741227i32);
    pub const STATUS_ISSUING_CA_UNTRUSTED: NTSTATUS = NTSTATUS(-1073740918i32);
    pub const STATUS_ISSUING_CA_UNTRUSTED_KDC: NTSTATUS = NTSTATUS(-1073740787i32);
    pub const STATUS_JOB_NOT_EMPTY: NTSTATUS = NTSTATUS(-1073740529i32);
    pub const STATUS_JOB_NO_CONTAINER: NTSTATUS = NTSTATUS(-1073740535i32);
    pub const STATUS_JOURNAL_DELETE_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073741129i32);
    pub const STATUS_JOURNAL_ENTRY_DELETED: NTSTATUS = NTSTATUS(-1073741105i32);
    pub const STATUS_JOURNAL_NOT_ACTIVE: NTSTATUS = NTSTATUS(-1073741128i32);
    pub const STATUS_KDC_CERT_EXPIRED: NTSTATUS = NTSTATUS(-1073740786i32);
    pub const STATUS_KDC_CERT_REVOKED: NTSTATUS = NTSTATUS(-1073740785i32);
    pub const STATUS_KDC_INVALID_REQUEST: NTSTATUS = NTSTATUS(-1073741061i32);
    pub const STATUS_KDC_UNABLE_TO_REFER: NTSTATUS = NTSTATUS(-1073741060i32);
    pub const STATUS_KDC_UNKNOWN_ETYPE: NTSTATUS = NTSTATUS(-1073741059i32);
    pub const STATUS_KERNEL_APC: NTSTATUS = NTSTATUS(256i32);
    pub const STATUS_KERNEL_EXECUTABLE_MEMORY_WRITE: NTSTATUS = NTSTATUS(-1073739996i32);
    pub const STATUS_KEY_DELETED: NTSTATUS = NTSTATUS(-1073741444i32);
    pub const STATUS_KEY_HAS_CHILDREN: NTSTATUS = NTSTATUS(-1073741440i32);
    pub const STATUS_LAST_ADMIN: NTSTATUS = NTSTATUS(-1073741719i32);
    pub const STATUS_LICENSE_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073741223i32);
    pub const STATUS_LICENSE_VIOLATION: NTSTATUS = NTSTATUS(-1073741206i32);
    pub const STATUS_LINK_FAILED: NTSTATUS = NTSTATUS(-1073741506i32);
    pub const STATUS_LINK_TIMEOUT: NTSTATUS = NTSTATUS(-1073741505i32);
    pub const STATUS_LM_CROSS_ENCRYPTION_REQUIRED: NTSTATUS = NTSTATUS(-1073741441i32);
    pub const STATUS_LOCAL_DISCONNECT: NTSTATUS = NTSTATUS(-1073741509i32);
    pub const STATUS_LOCAL_POLICY_MODIFICATION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700726i32);
    pub const STATUS_LOCAL_USER_SESSION_KEY: NTSTATUS = NTSTATUS(1073741830i32);
    pub const STATUS_LOCK_NOT_GRANTED: NTSTATUS = NTSTATUS(-1073741739i32);
    pub const STATUS_LOGIN_TIME_RESTRICTION: NTSTATUS = NTSTATUS(-1073741241i32);
    pub const STATUS_LOGIN_WKSTA_RESTRICTION: NTSTATUS = NTSTATUS(-1073741240i32);
    pub const STATUS_LOGON_NOT_GRANTED: NTSTATUS = NTSTATUS(-1073741483i32);
    pub const STATUS_LOGON_SERVER_CONFLICT: NTSTATUS = NTSTATUS(-1073741518i32);
    pub const STATUS_LOGON_SESSION_COLLISION: NTSTATUS = NTSTATUS(-1073741563i32);
    pub const STATUS_LOGON_SESSION_EXISTS: NTSTATUS = NTSTATUS(-1073741586i32);
    pub const STATUS_LOG_APPENDED_FLUSH_FAILED: NTSTATUS = NTSTATUS(-1072037841i32);
    pub const STATUS_LOG_ARCHIVE_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072037855i32);
    pub const STATUS_LOG_ARCHIVE_NOT_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072037856i32);
    pub const STATUS_LOG_BLOCKS_EXHAUSTED: NTSTATUS = NTSTATUS(-1072037882i32);
    pub const STATUS_LOG_BLOCK_INCOMPLETE: NTSTATUS = NTSTATUS(-1072037884i32);
    pub const STATUS_LOG_BLOCK_INVALID: NTSTATUS = NTSTATUS(-1072037878i32);
    pub const STATUS_LOG_BLOCK_VERSION: NTSTATUS = NTSTATUS(-1072037879i32);
    pub const STATUS_LOG_CANT_DELETE: NTSTATUS = NTSTATUS(-1072037871i32);
    pub const STATUS_LOG_CLIENT_ALREADY_REGISTERED: NTSTATUS = NTSTATUS(-1072037852i32);
    pub const STATUS_LOG_CLIENT_NOT_REGISTERED: NTSTATUS = NTSTATUS(-1072037851i32);
    pub const STATUS_LOG_CONTAINER_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1072037870i32);
    pub const STATUS_LOG_CONTAINER_OPEN_FAILED: NTSTATUS = NTSTATUS(-1072037847i32);
    pub const STATUS_LOG_CONTAINER_READ_FAILED: NTSTATUS = NTSTATUS(-1072037849i32);
    pub const STATUS_LOG_CONTAINER_STATE_INVALID: NTSTATUS = NTSTATUS(-1072037846i32);
    pub const STATUS_LOG_CONTAINER_WRITE_FAILED: NTSTATUS = NTSTATUS(-1072037848i32);
    pub const STATUS_LOG_CORRUPTION_DETECTED: NTSTATUS = NTSTATUS(-1072103376i32);
    pub const STATUS_LOG_DEDICATED: NTSTATUS = NTSTATUS(-1072037857i32);
    pub const STATUS_LOG_EPHEMERAL: NTSTATUS = NTSTATUS(-1072037854i32);
    pub const STATUS_LOG_FILE_FULL: NTSTATUS = NTSTATUS(-1073741432i32);
    pub const STATUS_LOG_FULL: NTSTATUS = NTSTATUS(-1072037859i32);
    pub const STATUS_LOG_FULL_HANDLER_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072037850i32);
    pub const STATUS_LOG_GROWTH_FAILED: NTSTATUS = NTSTATUS(-1072103399i32);
    pub const STATUS_LOG_HARD_ERROR: NTSTATUS = NTSTATUS(1073741850i32);
    pub const STATUS_LOG_INCONSISTENT_SECURITY: NTSTATUS = NTSTATUS(-1072037842i32);
    pub const STATUS_LOG_INVALID_RANGE: NTSTATUS = NTSTATUS(-1072037883i32);
    pub const STATUS_LOG_METADATA_CORRUPT: NTSTATUS = NTSTATUS(-1072037875i32);
    pub const STATUS_LOG_METADATA_FLUSH_FAILED: NTSTATUS = NTSTATUS(-1072037843i32);
    pub const STATUS_LOG_METADATA_INCONSISTENT: NTSTATUS = NTSTATUS(-1072037873i32);
    pub const STATUS_LOG_METADATA_INVALID: NTSTATUS = NTSTATUS(-1072037874i32);
    pub const STATUS_LOG_MULTIPLEXED: NTSTATUS = NTSTATUS(-1072037858i32);
    pub const STATUS_LOG_NOT_ENOUGH_CONTAINERS: NTSTATUS = NTSTATUS(-1072037853i32);
    pub const STATUS_LOG_NO_RESTART: NTSTATUS = NTSTATUS(1075445772i32);
    pub const STATUS_LOG_PINNED: NTSTATUS = NTSTATUS(-1072037844i32);
    pub const STATUS_LOG_PINNED_ARCHIVE_TAIL: NTSTATUS = NTSTATUS(-1072037864i32);
    pub const STATUS_LOG_PINNED_RESERVATION: NTSTATUS = NTSTATUS(-1072037840i32);
    pub const STATUS_LOG_POLICY_ALREADY_INSTALLED: NTSTATUS = NTSTATUS(-1072037868i32);
    pub const STATUS_LOG_POLICY_CONFLICT: NTSTATUS = NTSTATUS(-1072037865i32);
    pub const STATUS_LOG_POLICY_INVALID: NTSTATUS = NTSTATUS(-1072037866i32);
    pub const STATUS_LOG_POLICY_NOT_INSTALLED: NTSTATUS = NTSTATUS(-1072037867i32);
    pub const STATUS_LOG_READ_CONTEXT_INVALID: NTSTATUS = NTSTATUS(-1072037881i32);
    pub const STATUS_LOG_READ_MODE_INVALID: NTSTATUS = NTSTATUS(-1072037877i32);
    pub const STATUS_LOG_RECORDS_RESERVED_INVALID: NTSTATUS = NTSTATUS(-1072037862i32);
    pub const STATUS_LOG_RECORD_NONEXISTENT: NTSTATUS = NTSTATUS(-1072037863i32);
    pub const STATUS_LOG_RESERVATION_INVALID: NTSTATUS = NTSTATUS(-1072037872i32);
    pub const STATUS_LOG_RESIZE_INVALID_SIZE: NTSTATUS = NTSTATUS(-1072103413i32);
    pub const STATUS_LOG_RESTART_INVALID: NTSTATUS = NTSTATUS(-1072037880i32);
    pub const STATUS_LOG_SECTOR_INVALID: NTSTATUS = NTSTATUS(-1072037887i32);
    pub const STATUS_LOG_SECTOR_PARITY_INVALID: NTSTATUS = NTSTATUS(-1072037886i32);
    pub const STATUS_LOG_SECTOR_REMAPPED: NTSTATUS = NTSTATUS(-1072037885i32);
    pub const STATUS_LOG_SPACE_RESERVED_INVALID: NTSTATUS = NTSTATUS(-1072037861i32);
    pub const STATUS_LOG_START_OF_LOG: NTSTATUS = NTSTATUS(-1072037869i32);
    pub const STATUS_LOG_STATE_INVALID: NTSTATUS = NTSTATUS(-1072037845i32);
    pub const STATUS_LOG_TAIL_INVALID: NTSTATUS = NTSTATUS(-1072037860i32);
    pub const STATUS_LONGJUMP: NTSTATUS = NTSTATUS(-2147483610i32);
    pub const STATUS_LOST_MODE_LOGON_RESTRICTION: NTSTATUS = NTSTATUS(-1073741043i32);
    pub const STATUS_LOST_WRITEBEHIND_DATA: NTSTATUS = NTSTATUS(-1073741278i32);
    pub const STATUS_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR: NTSTATUS = NTSTATUS(-1073700734i32);
    pub const STATUS_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED: NTSTATUS = NTSTATUS(-1073700736i32);
    pub const STATUS_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR: NTSTATUS = NTSTATUS(-1073700735i32);
    pub const STATUS_LPAC_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073700349i32);
    pub const STATUS_LPC_HANDLE_COUNT_EXCEEDED: NTSTATUS = NTSTATUS(-1073739998i32);
    pub const STATUS_LPC_INVALID_CONNECTION_USAGE: NTSTATUS = NTSTATUS(-1073740026i32);
    pub const STATUS_LPC_RECEIVE_BUFFER_EXPECTED: NTSTATUS = NTSTATUS(-1073740027i32);
    pub const STATUS_LPC_REPLY_LOST: NTSTATUS = NTSTATUS(-1073741229i32);
    pub const STATUS_LPC_REQUESTS_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1073740025i32);
    pub const STATUS_LUIDS_EXHAUSTED: NTSTATUS = NTSTATUS(-1073741707i32);
    pub const STATUS_MAGAZINE_NOT_PRESENT: NTSTATUS = NTSTATUS(-1073741178i32);
    pub const STATUS_MAPPED_ALIGNMENT: NTSTATUS = NTSTATUS(-1073741280i32);
    pub const STATUS_MAPPED_FILE_SIZE_ZERO: NTSTATUS = NTSTATUS(-1073741538i32);
    pub const STATUS_MARKED_TO_DISALLOW_WRITES: NTSTATUS = NTSTATUS(-1073740659i32);
    pub const STATUS_MARSHALL_OVERFLOW: NTSTATUS = NTSTATUS(-1073741263i32);
    pub const STATUS_MAX_REFERRALS_EXCEEDED: NTSTATUS = NTSTATUS(-1073741068i32);
    pub const STATUS_MCA_EXCEPTION: NTSTATUS = NTSTATUS(-1073740013i32);
    pub const STATUS_MCA_OCCURED: NTSTATUS = NTSTATUS(-1073740950i32);
    pub const STATUS_MEDIA_CHANGED: NTSTATUS = NTSTATUS(-2147483620i32);
    pub const STATUS_MEDIA_CHECK: NTSTATUS = NTSTATUS(-2147483616i32);
    pub const STATUS_MEDIA_WRITE_PROTECTED: NTSTATUS = NTSTATUS(-1073741662i32);
    pub const STATUS_MEMBERS_PRIMARY_GROUP: NTSTATUS = NTSTATUS(-1073741529i32);
    pub const STATUS_MEMBER_IN_ALIAS: NTSTATUS = NTSTATUS(-1073741485i32);
    pub const STATUS_MEMBER_IN_GROUP: NTSTATUS = NTSTATUS(-1073741721i32);
    pub const STATUS_MEMBER_NOT_IN_ALIAS: NTSTATUS = NTSTATUS(-1073741486i32);
    pub const STATUS_MEMBER_NOT_IN_GROUP: NTSTATUS = NTSTATUS(-1073741720i32);
    pub const STATUS_MEMORY_NOT_ALLOCATED: NTSTATUS = NTSTATUS(-1073741664i32);
    pub const STATUS_MESSAGE_LOST: NTSTATUS = NTSTATUS(-1073740031i32);
    pub const STATUS_MESSAGE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741559i32);
    pub const STATUS_MESSAGE_RETRIEVED: NTSTATUS = NTSTATUS(1073741870i32);
    pub const STATUS_MFT_TOO_FRAGMENTED: NTSTATUS = NTSTATUS(-1073741052i32);
    pub const STATUS_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION: NTSTATUS = NTSTATUS(-1072103388i32);
    pub const STATUS_MISSING_SYSTEMFILE: NTSTATUS = NTSTATUS(-1073741501i32);
    pub const STATUS_MONITOR_INVALID_DESCRIPTOR_CHECKSUM: NTSTATUS = NTSTATUS(-1071841277i32);
    pub const STATUS_MONITOR_INVALID_DETAILED_TIMING_BLOCK: NTSTATUS = NTSTATUS(-1071841271i32);
    pub const STATUS_MONITOR_INVALID_MANUFACTURE_DATE: NTSTATUS = NTSTATUS(-1071841270i32);
    pub const STATUS_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK: NTSTATUS = NTSTATUS(-1071841274i32);
    pub const STATUS_MONITOR_INVALID_STANDARD_TIMING_BLOCK: NTSTATUS = NTSTATUS(-1071841276i32);
    pub const STATUS_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK: NTSTATUS = NTSTATUS(-1071841273i32);
    pub const STATUS_MONITOR_NO_DESCRIPTOR: NTSTATUS = NTSTATUS(-1071841279i32);
    pub const STATUS_MONITOR_NO_MORE_DESCRIPTOR_DATA: NTSTATUS = NTSTATUS(-1071841272i32);
    pub const STATUS_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT: NTSTATUS = NTSTATUS(-1071841278i32);
    pub const STATUS_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED: NTSTATUS = NTSTATUS(-1071841275i32);
    pub const STATUS_MORE_ENTRIES: NTSTATUS = NTSTATUS(261i32);
    pub const STATUS_MORE_PROCESSING_REQUIRED: NTSTATUS = NTSTATUS(-1073741802i32);
    pub const STATUS_MOUNT_POINT_NOT_RESOLVED: NTSTATUS = NTSTATUS(-1073740952i32);
    pub const STATUS_MP_PROCESSOR_MISMATCH: NTSTATUS = NTSTATUS(1073741865i32);
    pub const STATUS_MUI_FILE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073020927i32);
    pub const STATUS_MUI_FILE_NOT_LOADED: NTSTATUS = NTSTATUS(-1073020922i32);
    pub const STATUS_MUI_INVALID_FILE: NTSTATUS = NTSTATUS(-1073020926i32);
    pub const STATUS_MUI_INVALID_LOCALE_NAME: NTSTATUS = NTSTATUS(-1073020924i32);
    pub const STATUS_MUI_INVALID_RC_CONFIG: NTSTATUS = NTSTATUS(-1073020925i32);
    pub const STATUS_MUI_INVALID_ULTIMATEFALLBACK_NAME: NTSTATUS = NTSTATUS(-1073020923i32);
    pub const STATUS_MULTIPLE_FAULT_VIOLATION: NTSTATUS = NTSTATUS(-1073741080i32);
    pub const STATUS_MUST_BE_KDC: NTSTATUS = NTSTATUS(-1073741067i32);
    pub const STATUS_MUTANT_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073741423i32);
    pub const STATUS_MUTANT_NOT_OWNED: NTSTATUS = NTSTATUS(-1073741754i32);
    pub const STATUS_MUTUAL_AUTHENTICATION_FAILED: NTSTATUS = NTSTATUS(-1073741117i32);
    pub const STATUS_NAME_TOO_LONG: NTSTATUS = NTSTATUS(-1073741562i32);
    pub const STATUS_NDIS_ADAPTER_NOT_FOUND: NTSTATUS = NTSTATUS(-1071448058i32);
    pub const STATUS_NDIS_ADAPTER_NOT_READY: NTSTATUS = NTSTATUS(-1071448047i32);
    pub const STATUS_NDIS_ADAPTER_REMOVED: NTSTATUS = NTSTATUS(-1071448040i32);
    pub const STATUS_NDIS_ALREADY_MAPPED: NTSTATUS = NTSTATUS(-1071448035i32);
    pub const STATUS_NDIS_BAD_CHARACTERISTICS: NTSTATUS = NTSTATUS(-1071448059i32);
    pub const STATUS_NDIS_BAD_VERSION: NTSTATUS = NTSTATUS(-1071448060i32);
    pub const STATUS_NDIS_BUFFER_TOO_SHORT: NTSTATUS = NTSTATUS(-1071448042i32);
    pub const STATUS_NDIS_CLOSING: NTSTATUS = NTSTATUS(-1071448062i32);
    pub const STATUS_NDIS_DEVICE_FAILED: NTSTATUS = NTSTATUS(-1071448056i32);
    pub const STATUS_NDIS_DOT11_AP_BAND_CURRENTLY_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1071439866i32);
    pub const STATUS_NDIS_DOT11_AP_BAND_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1071439864i32);
    pub const STATUS_NDIS_DOT11_AP_CHANNEL_CURRENTLY_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1071439867i32);
    pub const STATUS_NDIS_DOT11_AP_CHANNEL_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1071439865i32);
    pub const STATUS_NDIS_DOT11_AUTO_CONFIG_ENABLED: NTSTATUS = NTSTATUS(-1071439872i32);
    pub const STATUS_NDIS_DOT11_MEDIA_IN_USE: NTSTATUS = NTSTATUS(-1071439871i32);
    pub const STATUS_NDIS_DOT11_POWER_STATE_INVALID: NTSTATUS = NTSTATUS(-1071439870i32);
    pub const STATUS_NDIS_ERROR_READING_FILE: NTSTATUS = NTSTATUS(-1071448036i32);
    pub const STATUS_NDIS_FILE_NOT_FOUND: NTSTATUS = NTSTATUS(-1071448037i32);
    pub const STATUS_NDIS_GROUP_ADDRESS_IN_USE: NTSTATUS = NTSTATUS(-1071448038i32);
    pub const STATUS_NDIS_INDICATION_REQUIRED: NTSTATUS = NTSTATUS(1076035585i32);
    pub const STATUS_NDIS_INTERFACE_NOT_FOUND: NTSTATUS = NTSTATUS(-1071448021i32);
    pub const STATUS_NDIS_INVALID_ADDRESS: NTSTATUS = NTSTATUS(-1071448030i32);
    pub const STATUS_NDIS_INVALID_DATA: NTSTATUS = NTSTATUS(-1071448043i32);
    pub const STATUS_NDIS_INVALID_DEVICE_REQUEST: NTSTATUS = NTSTATUS(-1071448048i32);
    pub const STATUS_NDIS_INVALID_LENGTH: NTSTATUS = NTSTATUS(-1071448044i32);
    pub const STATUS_NDIS_INVALID_OID: NTSTATUS = NTSTATUS(-1071448041i32);
    pub const STATUS_NDIS_INVALID_PACKET: NTSTATUS = NTSTATUS(-1071448049i32);
    pub const STATUS_NDIS_INVALID_PORT: NTSTATUS = NTSTATUS(-1071448019i32);
    pub const STATUS_NDIS_INVALID_PORT_STATE: NTSTATUS = NTSTATUS(-1071448018i32);
    pub const STATUS_NDIS_LOW_POWER_STATE: NTSTATUS = NTSTATUS(-1071448017i32);
    pub const STATUS_NDIS_MEDIA_DISCONNECTED: NTSTATUS = NTSTATUS(-1071448033i32);
    pub const STATUS_NDIS_MULTICAST_EXISTS: NTSTATUS = NTSTATUS(-1071448054i32);
    pub const STATUS_NDIS_MULTICAST_FULL: NTSTATUS = NTSTATUS(-1071448055i32);
    pub const STATUS_NDIS_MULTICAST_NOT_FOUND: NTSTATUS = NTSTATUS(-1071448053i32);
    pub const STATUS_NDIS_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071447877i32);
    pub const STATUS_NDIS_NO_QUEUES: NTSTATUS = NTSTATUS(-1071448015i32);
    pub const STATUS_NDIS_OFFLOAD_CONNECTION_REJECTED: NTSTATUS = NTSTATUS(-1071443950i32);
    pub const STATUS_NDIS_OFFLOAD_PATH_REJECTED: NTSTATUS = NTSTATUS(-1071443949i32);
    pub const STATUS_NDIS_OFFLOAD_POLICY: NTSTATUS = NTSTATUS(-1071443953i32);
    pub const STATUS_NDIS_OPEN_FAILED: NTSTATUS = NTSTATUS(-1071448057i32);
    pub const STATUS_NDIS_PAUSED: NTSTATUS = NTSTATUS(-1071448022i32);
    pub const STATUS_NDIS_PM_PROTOCOL_OFFLOAD_LIST_FULL: NTSTATUS = NTSTATUS(-1071439868i32);
    pub const STATUS_NDIS_PM_WOL_PATTERN_LIST_FULL: NTSTATUS = NTSTATUS(-1071439869i32);
    pub const STATUS_NDIS_REINIT_REQUIRED: NTSTATUS = NTSTATUS(-1071448016i32);
    pub const STATUS_NDIS_REQUEST_ABORTED: NTSTATUS = NTSTATUS(-1071448052i32);
    pub const STATUS_NDIS_RESET_IN_PROGRESS: NTSTATUS = NTSTATUS(-1071448051i32);
    pub const STATUS_NDIS_RESOURCE_CONFLICT: NTSTATUS = NTSTATUS(-1071448034i32);
    pub const STATUS_NDIS_UNSUPPORTED_MEDIA: NTSTATUS = NTSTATUS(-1071448039i32);
    pub const STATUS_NDIS_UNSUPPORTED_REVISION: NTSTATUS = NTSTATUS(-1071448020i32);
    pub const STATUS_ND_QUEUE_OVERFLOW: NTSTATUS = NTSTATUS(-1073700847i32);
    pub const STATUS_NEEDS_REGISTRATION: NTSTATUS = NTSTATUS(-1073740663i32);
    pub const STATUS_NEEDS_REMEDIATION: NTSTATUS = NTSTATUS(-1073740702i32);
    pub const STATUS_NETLOGON_NOT_STARTED: NTSTATUS = NTSTATUS(-1073741422i32);
    pub const STATUS_NETWORK_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073741622i32);
    pub const STATUS_NETWORK_ACCESS_DENIED_EDP: NTSTATUS = NTSTATUS(-1073740658i32);
    pub const STATUS_NETWORK_AUTHENTICATION_PROMPT_CANCELED: NTSTATUS = NTSTATUS(-1067646972i32);
    pub const STATUS_NETWORK_BUSY: NTSTATUS = NTSTATUS(-1073741633i32);
    pub const STATUS_NETWORK_CREDENTIAL_CONFLICT: NTSTATUS = NTSTATUS(-1073741419i32);
    pub const STATUS_NETWORK_NAME_DELETED: NTSTATUS = NTSTATUS(-1073741623i32);
    pub const STATUS_NETWORK_OPEN_RESTRICTION: NTSTATUS = NTSTATUS(-1073741311i32);
    pub const STATUS_NETWORK_SESSION_EXPIRED: NTSTATUS = NTSTATUS(-1073740964i32);
    pub const STATUS_NETWORK_UNREACHABLE: NTSTATUS = NTSTATUS(-1073741252i32);
    pub const STATUS_NET_WRITE_FAULT: NTSTATUS = NTSTATUS(-1073741614i32);
    pub const STATUS_NOINTERFACE: NTSTATUS = NTSTATUS(-1073741127i32);
    pub const STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT: NTSTATUS = NTSTATUS(-1073741416i32);
    pub const STATUS_NOLOGON_SERVER_TRUST_ACCOUNT: NTSTATUS = NTSTATUS(-1073741414i32);
    pub const STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT: NTSTATUS = NTSTATUS(-1073741415i32);
    pub const STATUS_NONCONTINUABLE_EXCEPTION: NTSTATUS = NTSTATUS(-1073741787i32);
    pub const STATUS_NONEXISTENT_EA_ENTRY: NTSTATUS = NTSTATUS(-1073741743i32);
    pub const STATUS_NONEXISTENT_SECTOR: NTSTATUS = NTSTATUS(-1073741803i32);
    pub const STATUS_NONE_MAPPED: NTSTATUS = NTSTATUS(-1073741709i32);
    pub const STATUS_NOTHING_TO_TERMINATE: NTSTATUS = NTSTATUS(290i32);
    pub const STATUS_NOTIFICATION_GUID_ALREADY_DEFINED: NTSTATUS = NTSTATUS(-1073741404i32);
    pub const STATUS_NOTIFY_CLEANUP: NTSTATUS = NTSTATUS(267i32);
    pub const STATUS_NOTIFY_ENUM_DIR: NTSTATUS = NTSTATUS(268i32);
    pub const STATUS_NOT_ALLOWED_ON_SYSTEM_FILE: NTSTATUS = NTSTATUS(-1073741401i32);
    pub const STATUS_NOT_ALL_ASSIGNED: NTSTATUS = NTSTATUS(262i32);
    pub const STATUS_NOT_APPCONTAINER: NTSTATUS = NTSTATUS(-1073700352i32);
    pub const STATUS_NOT_A_CLOUD_FILE: NTSTATUS = NTSTATUS(-1073688825i32);
    pub const STATUS_NOT_A_CLOUD_SYNC_ROOT: NTSTATUS = NTSTATUS(-1073688802i32);
    pub const STATUS_NOT_A_DAX_VOLUME: NTSTATUS = NTSTATUS(-1073740623i32);
    pub const STATUS_NOT_A_DIRECTORY: NTSTATUS = NTSTATUS(-1073741565i32);
    pub const STATUS_NOT_A_REPARSE_POINT: NTSTATUS = NTSTATUS(-1073741195i32);
    pub const STATUS_NOT_A_TIERED_VOLUME: NTSTATUS = NTSTATUS(-1073740531i32);
    pub const STATUS_NOT_CAPABLE: NTSTATUS = NTSTATUS(-1073740759i32);
    pub const STATUS_NOT_CLIENT_SESSION: NTSTATUS = NTSTATUS(-1073741289i32);
    pub const STATUS_NOT_COMMITTED: NTSTATUS = NTSTATUS(-1073741779i32);
    pub const STATUS_NOT_DAX_MAPPABLE: NTSTATUS = NTSTATUS(-1073740622i32);
    pub const STATUS_NOT_EXPORT_FORMAT: NTSTATUS = NTSTATUS(-1073741166i32);
    pub const STATUS_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741275i32);
    pub const STATUS_NOT_GUI_PROCESS: NTSTATUS = NTSTATUS(-1073740538i32);
    pub const STATUS_NOT_IMPLEMENTED: NTSTATUS = NTSTATUS(-1073741822i32);
    pub const STATUS_NOT_LOCKED: NTSTATUS = NTSTATUS(-1073741782i32);
    pub const STATUS_NOT_LOGON_PROCESS: NTSTATUS = NTSTATUS(-1073741587i32);
    pub const STATUS_NOT_MAPPED_DATA: NTSTATUS = NTSTATUS(-1073741688i32);
    pub const STATUS_NOT_MAPPED_VIEW: NTSTATUS = NTSTATUS(-1073741799i32);
    pub const STATUS_NOT_READ_FROM_COPY: NTSTATUS = NTSTATUS(-1073740694i32);
    pub const STATUS_NOT_REDUNDANT_STORAGE: NTSTATUS = NTSTATUS(-1073740679i32);
    pub const STATUS_NOT_REGISTRY_FILE: NTSTATUS = NTSTATUS(-1073741476i32);
    pub const STATUS_NOT_SAFE_MODE_DRIVER: NTSTATUS = NTSTATUS(-1073740961i32);
    pub const STATUS_NOT_SAME_DEVICE: NTSTATUS = NTSTATUS(-1073741612i32);
    pub const STATUS_NOT_SAME_OBJECT: NTSTATUS = NTSTATUS(-1073741396i32);
    pub const STATUS_NOT_SERVER_SESSION: NTSTATUS = NTSTATUS(-1073741290i32);
    pub const STATUS_NOT_SNAPSHOT_VOLUME: NTSTATUS = NTSTATUS(-1072103353i32);
    pub const STATUS_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073741637i32);
    pub const STATUS_NOT_SUPPORTED_IN_APPCONTAINER: NTSTATUS = NTSTATUS(-1073700351i32);
    pub const STATUS_NOT_SUPPORTED_ON_DAX: NTSTATUS = NTSTATUS(-1073740646i32);
    pub const STATUS_NOT_SUPPORTED_ON_SBS: NTSTATUS = NTSTATUS(-1073741056i32);
    pub const STATUS_NOT_SUPPORTED_WITH_AUDITING: NTSTATUS = NTSTATUS(-1073740595i32);
    pub const STATUS_NOT_SUPPORTED_WITH_BTT: NTSTATUS = NTSTATUS(-1073740619i32);
    pub const STATUS_NOT_SUPPORTED_WITH_BYPASSIO: NTSTATUS = NTSTATUS(-1073740601i32);
    pub const STATUS_NOT_SUPPORTED_WITH_CACHED_HANDLE: NTSTATUS = NTSTATUS(-1073740587i32);
    pub const STATUS_NOT_SUPPORTED_WITH_COMPRESSION: NTSTATUS = NTSTATUS(-1073740598i32);
    pub const STATUS_NOT_SUPPORTED_WITH_DEDUPLICATION: NTSTATUS = NTSTATUS(-1073740596i32);
    pub const STATUS_NOT_SUPPORTED_WITH_ENCRYPTION: NTSTATUS = NTSTATUS(-1073740599i32);
    pub const STATUS_NOT_SUPPORTED_WITH_MONITORING: NTSTATUS = NTSTATUS(-1073740594i32);
    pub const STATUS_NOT_SUPPORTED_WITH_REPLICATION: NTSTATUS = NTSTATUS(-1073740597i32);
    pub const STATUS_NOT_SUPPORTED_WITH_SNAPSHOT: NTSTATUS = NTSTATUS(-1073740593i32);
    pub const STATUS_NOT_SUPPORTED_WITH_VIRTUALIZATION: NTSTATUS = NTSTATUS(-1073740592i32);
    pub const STATUS_NOT_TINY_STREAM: NTSTATUS = NTSTATUS(-1073741274i32);
    pub const STATUS_NO_ACE_CONDITION: NTSTATUS = NTSTATUS(-2147483601i32);
    pub const STATUS_NO_APPLICABLE_APP_LICENSES_FOUND: NTSTATUS = NTSTATUS(-1058406399i32);
    pub const STATUS_NO_APPLICATION_PACKAGE: NTSTATUS = NTSTATUS(-1073741398i32);
    pub const STATUS_NO_BROWSER_SERVERS_FOUND: NTSTATUS = NTSTATUS(-1073741284i32);
    pub const STATUS_NO_BYPASSIO_DRIVER_SUPPORT: NTSTATUS = NTSTATUS(-1073740600i32);
    pub const STATUS_NO_CALLBACK_ACTIVE: NTSTATUS = NTSTATUS(-1073741224i32);
    pub const STATUS_NO_DATA_DETECTED: NTSTATUS = NTSTATUS(-2147483614i32);
    pub const STATUS_NO_EAS_ON_FILE: NTSTATUS = NTSTATUS(-1073741742i32);
    pub const STATUS_NO_EFS: NTSTATUS = NTSTATUS(-1073741170i32);
    pub const STATUS_NO_EVENT_PAIR: NTSTATUS = NTSTATUS(-1073741490i32);
    pub const STATUS_NO_GUID_TRANSLATION: NTSTATUS = NTSTATUS(-1073741556i32);
    pub const STATUS_NO_IMPERSONATION_TOKEN: NTSTATUS = NTSTATUS(-1073741732i32);
    pub const STATUS_NO_INHERITANCE: NTSTATUS = NTSTATUS(-2147483637i32);
    pub const STATUS_NO_IP_ADDRESSES: NTSTATUS = NTSTATUS(-1073741071i32);
    pub const STATUS_NO_KERB_KEY: NTSTATUS = NTSTATUS(-1073741022i32);
    pub const STATUS_NO_KEY: NTSTATUS = NTSTATUS(-1073739508i32);
    pub const STATUS_NO_LDT: NTSTATUS = NTSTATUS(-1073741545i32);
    pub const STATUS_NO_LINK_TRACKING_IN_TRANSACTION: NTSTATUS = NTSTATUS(-1072103335i32);
    pub const STATUS_NO_LOGON_SERVERS: NTSTATUS = NTSTATUS(-1073741730i32);
    pub const STATUS_NO_LOG_SPACE: NTSTATUS = NTSTATUS(-1073741443i32);
    pub const STATUS_NO_MATCH: NTSTATUS = NTSTATUS(-1073741198i32);
    pub const STATUS_NO_MEDIA: NTSTATUS = NTSTATUS(-1073741448i32);
    pub const STATUS_NO_MEDIA_IN_DEVICE: NTSTATUS = NTSTATUS(-1073741805i32);
    pub const STATUS_NO_MEMORY: NTSTATUS = NTSTATUS(-1073741801i32);
    pub const STATUS_NO_MORE_EAS: NTSTATUS = NTSTATUS(-2147483630i32);
    pub const STATUS_NO_MORE_ENTRIES: NTSTATUS = NTSTATUS(-2147483622i32);
    pub const STATUS_NO_MORE_FILES: NTSTATUS = NTSTATUS(-2147483642i32);
    pub const STATUS_NO_MORE_MATCHES: NTSTATUS = NTSTATUS(-1073741197i32);
    pub const STATUS_NO_PAGEFILE: NTSTATUS = NTSTATUS(-1073741497i32);
    pub const STATUS_NO_PA_DATA: NTSTATUS = NTSTATUS(-1073741064i32);
    pub const STATUS_NO_PHYSICALLY_ALIGNED_FREE_SPACE_FOUND: NTSTATUS = NTSTATUS(-1073740635i32);
    pub const STATUS_NO_QUOTAS_FOR_ACCOUNT: NTSTATUS = NTSTATUS(269i32);
    pub const STATUS_NO_RANGES_PROCESSED: NTSTATUS = NTSTATUS(-1073740704i32);
    pub const STATUS_NO_RECOVERY_POLICY: NTSTATUS = NTSTATUS(-1073741171i32);
    pub const STATUS_NO_S4U_PROT_SUPPORT: NTSTATUS = NTSTATUS(-1073740790i32);
    pub const STATUS_NO_SAVEPOINT_WITH_OPEN_FILES: NTSTATUS = NTSTATUS(-1072103352i32);
    pub const STATUS_NO_SECRETS: NTSTATUS = NTSTATUS(-1073740943i32);
    pub const STATUS_NO_SECURITY_CONTEXT: NTSTATUS = NTSTATUS(-1073740755i32);
    pub const STATUS_NO_SECURITY_ON_OBJECT: NTSTATUS = NTSTATUS(-1073741609i32);
    pub const STATUS_NO_SPOOL_SPACE: NTSTATUS = NTSTATUS(-1073741625i32);
    pub const STATUS_NO_SUCH_ALIAS: NTSTATUS = NTSTATUS(-1073741487i32);
    pub const STATUS_NO_SUCH_DEVICE: NTSTATUS = NTSTATUS(-1073741810i32);
    pub const STATUS_NO_SUCH_DOMAIN: NTSTATUS = NTSTATUS(-1073741601i32);
    pub const STATUS_NO_SUCH_FILE: NTSTATUS = NTSTATUS(-1073741809i32);
    pub const STATUS_NO_SUCH_GROUP: NTSTATUS = NTSTATUS(-1073741722i32);
    pub const STATUS_NO_SUCH_MEMBER: NTSTATUS = NTSTATUS(-1073741446i32);
    pub const STATUS_NO_SUCH_PACKAGE: NTSTATUS = NTSTATUS(-1073741570i32);
    pub const STATUS_NO_SUCH_PRIVILEGE: NTSTATUS = NTSTATUS(-1073741728i32);
    pub const STATUS_NO_TGT_REPLY: NTSTATUS = NTSTATUS(-1073741073i32);
    pub const STATUS_NO_TOKEN: NTSTATUS = NTSTATUS(-1073741700i32);
    pub const STATUS_NO_TRACKING_SERVICE: NTSTATUS = NTSTATUS(-1073741153i32);
    pub const STATUS_NO_TRUST_LSA_SECRET: NTSTATUS = NTSTATUS(-1073741430i32);
    pub const STATUS_NO_TRUST_SAM_ACCOUNT: NTSTATUS = NTSTATUS(-1073741429i32);
    pub const STATUS_NO_TXF_METADATA: NTSTATUS = NTSTATUS(-2145845207i32);
    pub const STATUS_NO_UNICODE_TRANSLATION: NTSTATUS = NTSTATUS(-1073740009i32);
    pub const STATUS_NO_USER_KEYS: NTSTATUS = NTSTATUS(-1073741168i32);
    pub const STATUS_NO_USER_SESSION_KEY: NTSTATUS = NTSTATUS(-1073741310i32);
    pub const STATUS_NO_WORK_DONE: NTSTATUS = NTSTATUS(-2147483598i32);
    pub const STATUS_NO_YIELD_PERFORMED: NTSTATUS = NTSTATUS(1073741860i32);
    pub const STATUS_NTLM_BLOCKED: NTSTATUS = NTSTATUS(-1073740776i32);
    pub const STATUS_NT_CROSS_ENCRYPTION_REQUIRED: NTSTATUS = NTSTATUS(-1073741475i32);
    pub const STATUS_NULL_LM_PASSWORD: NTSTATUS = NTSTATUS(1073741837i32);
    pub const STATUS_OBJECTID_EXISTS: NTSTATUS = NTSTATUS(-1073741269i32);
    pub const STATUS_OBJECTID_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741072i32);
    pub const STATUS_OBJECT_IS_IMMUTABLE: NTSTATUS = NTSTATUS(-1073740610i32);
    pub const STATUS_OBJECT_NAME_COLLISION: NTSTATUS = NTSTATUS(-1073741771i32);
    pub const STATUS_OBJECT_NAME_EXISTS: NTSTATUS = NTSTATUS(1073741824i32);
    pub const STATUS_OBJECT_NAME_INVALID: NTSTATUS = NTSTATUS(-1073741773i32);
    pub const STATUS_OBJECT_NAME_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741772i32);
    pub const STATUS_OBJECT_NOT_EXTERNALLY_BACKED: NTSTATUS = NTSTATUS(-1073740691i32);
    pub const STATUS_OBJECT_NO_LONGER_EXISTS: NTSTATUS = NTSTATUS(-1072103391i32);
    pub const STATUS_OBJECT_PATH_INVALID: NTSTATUS = NTSTATUS(-1073741767i32);
    pub const STATUS_OBJECT_PATH_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741766i32);
    pub const STATUS_OBJECT_PATH_SYNTAX_BAD: NTSTATUS = NTSTATUS(-1073741765i32);
    pub const STATUS_OBJECT_TYPE_MISMATCH: NTSTATUS = NTSTATUS(-1073741788i32);
    pub const STATUS_OFFLOAD_READ_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700189i32);
    pub const STATUS_OFFLOAD_READ_FLT_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700191i32);
    pub const STATUS_OFFLOAD_WRITE_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700188i32);
    pub const STATUS_OFFLOAD_WRITE_FLT_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700190i32);
    pub const STATUS_ONLY_IF_CONNECTED: NTSTATUS = NTSTATUS(-1073741108i32);
    pub const STATUS_OPEN_FAILED: NTSTATUS = NTSTATUS(-1073741514i32);
    pub const STATUS_OPERATION_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073740682i32);
    pub const STATUS_OPERATION_NOT_SUPPORTED_IN_TRANSACTION: NTSTATUS = NTSTATUS(-1072103334i32);
    pub const STATUS_OPLOCK_BREAK_IN_PROGRESS: NTSTATUS = NTSTATUS(264i32);
    pub const STATUS_OPLOCK_HANDLE_CLOSED: NTSTATUS = NTSTATUS(534i32);
    pub const STATUS_OPLOCK_NOT_GRANTED: NTSTATUS = NTSTATUS(-1073741598i32);
    pub const STATUS_OPLOCK_SWITCHED_TO_NEW_HANDLE: NTSTATUS = NTSTATUS(533i32);
    pub const STATUS_ORDINAL_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741512i32);
    pub const STATUS_ORPHAN_NAME_EXHAUSTED: NTSTATUS = NTSTATUS(-1073739762i32);
    pub const STATUS_PACKAGE_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073740649i32);
    pub const STATUS_PACKAGE_UPDATING: NTSTATUS = NTSTATUS(-1073740695i32);
    pub const STATUS_PAGEFILE_CREATE_FAILED: NTSTATUS = NTSTATUS(-1073741498i32);
    pub const STATUS_PAGEFILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740603i32);
    pub const STATUS_PAGEFILE_QUOTA: NTSTATUS = NTSTATUS(-1073741817i32);
    pub const STATUS_PAGEFILE_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073741524i32);
    pub const STATUS_PAGE_FAULT_COPY_ON_WRITE: NTSTATUS = NTSTATUS(274i32);
    pub const STATUS_PAGE_FAULT_DEMAND_ZERO: NTSTATUS = NTSTATUS(273i32);
    pub const STATUS_PAGE_FAULT_GUARD_PAGE: NTSTATUS = NTSTATUS(275i32);
    pub const STATUS_PAGE_FAULT_PAGING_FILE: NTSTATUS = NTSTATUS(276i32);
    pub const STATUS_PAGE_FAULT_RETRY: NTSTATUS = NTSTATUS(873i32);
    pub const STATUS_PAGE_FAULT_TRANSITION: NTSTATUS = NTSTATUS(272i32);
    pub const STATUS_PARAMETER_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073740784i32);
    pub const STATUS_PARITY_ERROR: NTSTATUS = NTSTATUS(-1073741781i32);
    pub const STATUS_PARTIAL_COPY: NTSTATUS = NTSTATUS(-2147483635i32);
    pub const STATUS_PARTITION_FAILURE: NTSTATUS = NTSTATUS(-1073741454i32);
    pub const STATUS_PARTITION_TERMINATING: NTSTATUS = NTSTATUS(-1073740640i32);
    pub const STATUS_PASSWORD_CHANGE_REQUIRED: NTSTATUS = NTSTATUS(-1073741044i32);
    pub const STATUS_PASSWORD_RESTRICTION: NTSTATUS = NTSTATUS(-1073741716i32);
    pub const STATUS_PATCH_CONFLICT: NTSTATUS = NTSTATUS(-1073740628i32);
    pub const STATUS_PATCH_DEFERRED: NTSTATUS = NTSTATUS(1073741879i32);
    pub const STATUS_PATCH_NOT_REGISTERED: NTSTATUS = NTSTATUS(-1073740588i32);
    pub const STATUS_PATH_NOT_COVERED: NTSTATUS = NTSTATUS(-1073741225i32);
    pub const STATUS_PCP_ATTESTATION_CHALLENGE_NOT_SET: NTSTATUS = NTSTATUS(-1071046638i32);
    pub const STATUS_PCP_AUTHENTICATION_FAILED: NTSTATUS = NTSTATUS(-1071046648i32);
    pub const STATUS_PCP_AUTHENTICATION_IGNORED: NTSTATUS = NTSTATUS(-1071046647i32);
    pub const STATUS_PCP_BUFFER_LENGTH_MISMATCH: NTSTATUS = NTSTATUS(-1071046626i32);
    pub const STATUS_PCP_BUFFER_TOO_SMALL: NTSTATUS = NTSTATUS(-1071046650i32);
    pub const STATUS_PCP_CLAIM_TYPE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071046628i32);
    pub const STATUS_PCP_DEVICE_NOT_FOUND: NTSTATUS = NTSTATUS(-1071046643i32);
    pub const STATUS_PCP_DEVICE_NOT_READY: NTSTATUS = NTSTATUS(-1071046655i32);
    pub const STATUS_PCP_ERROR_MASK: NTSTATUS = NTSTATUS(-1071046656i32);
    pub const STATUS_PCP_FLAG_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071046652i32);
    pub const STATUS_PCP_IFX_RSA_KEY_CREATION_BLOCKED: NTSTATUS = NTSTATUS(-1071046625i32);
    pub const STATUS_PCP_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071046649i32);
    pub const STATUS_PCP_INVALID_HANDLE: NTSTATUS = NTSTATUS(-1071046654i32);
    pub const STATUS_PCP_INVALID_PARAMETER: NTSTATUS = NTSTATUS(-1071046653i32);
    pub const STATUS_PCP_KEY_ALREADY_FINALIZED: NTSTATUS = NTSTATUS(-1071046636i32);
    pub const STATUS_PCP_KEY_HANDLE_INVALIDATED: NTSTATUS = NTSTATUS(-1071046622i32);
    pub const STATUS_PCP_KEY_NOT_AIK: NTSTATUS = NTSTATUS(-1071046631i32);
    pub const STATUS_PCP_KEY_NOT_AUTHENTICATED: NTSTATUS = NTSTATUS(-1071046632i32);
    pub const STATUS_PCP_KEY_NOT_FINALIZED: NTSTATUS = NTSTATUS(-1071046639i32);
    pub const STATUS_PCP_KEY_NOT_LOADED: NTSTATUS = NTSTATUS(-1071046641i32);
    pub const STATUS_PCP_KEY_NOT_SIGNING_KEY: NTSTATUS = NTSTATUS(-1071046630i32);
    pub const STATUS_PCP_KEY_USAGE_POLICY_INVALID: NTSTATUS = NTSTATUS(-1071046634i32);
    pub const STATUS_PCP_KEY_USAGE_POLICY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071046635i32);
    pub const STATUS_PCP_LOCKED_OUT: NTSTATUS = NTSTATUS(-1071046629i32);
    pub const STATUS_PCP_NOT_PCR_BOUND: NTSTATUS = NTSTATUS(-1071046637i32);
    pub const STATUS_PCP_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071046651i32);
    pub const STATUS_PCP_NO_KEY_CERTIFICATION: NTSTATUS = NTSTATUS(-1071046640i32);
    pub const STATUS_PCP_POLICY_NOT_FOUND: NTSTATUS = NTSTATUS(-1071046646i32);
    pub const STATUS_PCP_PROFILE_NOT_FOUND: NTSTATUS = NTSTATUS(-1071046645i32);
    pub const STATUS_PCP_RAW_POLICY_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071046623i32);
    pub const STATUS_PCP_SOFT_KEY_ERROR: NTSTATUS = NTSTATUS(-1071046633i32);
    pub const STATUS_PCP_TICKET_MISSING: NTSTATUS = NTSTATUS(-1071046624i32);
    pub const STATUS_PCP_TPM_VERSION_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1071046627i32);
    pub const STATUS_PCP_UNSUPPORTED_PSS_SALT: NTSTATUS = NTSTATUS(1076437027i32);
    pub const STATUS_PCP_VALIDATION_FAILED: NTSTATUS = NTSTATUS(-1071046644i32);
    pub const STATUS_PCP_WRONG_PARENT: NTSTATUS = NTSTATUS(-1071046642i32);
    pub const STATUS_PENDING: NTSTATUS = NTSTATUS(259i32);
    pub const STATUS_PER_USER_TRUST_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073740799i32);
    pub const STATUS_PIPE_BROKEN: NTSTATUS = NTSTATUS(-1073741493i32);
    pub const STATUS_PIPE_BUSY: NTSTATUS = NTSTATUS(-1073741650i32);
    pub const STATUS_PIPE_CLOSING: NTSTATUS = NTSTATUS(-1073741647i32);
    pub const STATUS_PIPE_CONNECTED: NTSTATUS = NTSTATUS(-1073741646i32);
    pub const STATUS_PIPE_DISCONNECTED: NTSTATUS = NTSTATUS(-1073741648i32);
    pub const STATUS_PIPE_EMPTY: NTSTATUS = NTSTATUS(-1073741607i32);
    pub const STATUS_PIPE_LISTENING: NTSTATUS = NTSTATUS(-1073741645i32);
    pub const STATUS_PIPE_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073741652i32);
    pub const STATUS_PKINIT_CLIENT_FAILURE: NTSTATUS = NTSTATUS(-1073740916i32);
    pub const STATUS_PKINIT_FAILURE: NTSTATUS = NTSTATUS(-1073741024i32);
    pub const STATUS_PKINIT_NAME_MISMATCH: NTSTATUS = NTSTATUS(-1073741063i32);
    pub const STATUS_PKU2U_CERT_FAILURE: NTSTATUS = NTSTATUS(-1073740753i32);
    pub const STATUS_PLATFORM_MANIFEST_BINARY_ID_NOT_FOUND: NTSTATUS = NTSTATUS(-1058340859i32);
    pub const STATUS_PLATFORM_MANIFEST_CATALOG_NOT_AUTHORIZED: NTSTATUS = NTSTATUS(-1058340860i32);
    pub const STATUS_PLATFORM_MANIFEST_FILE_NOT_AUTHORIZED: NTSTATUS = NTSTATUS(-1058340861i32);
    pub const STATUS_PLATFORM_MANIFEST_INVALID: NTSTATUS = NTSTATUS(-1058340862i32);
    pub const STATUS_PLATFORM_MANIFEST_NOT_ACTIVE: NTSTATUS = NTSTATUS(-1058340858i32);
    pub const STATUS_PLATFORM_MANIFEST_NOT_AUTHORIZED: NTSTATUS = NTSTATUS(-1058340863i32);
    pub const STATUS_PLATFORM_MANIFEST_NOT_SIGNED: NTSTATUS = NTSTATUS(-1058340857i32);
    pub const STATUS_PLUGPLAY_NO_DEVICE: NTSTATUS = NTSTATUS(-1073741218i32);
    pub const STATUS_PLUGPLAY_QUERY_VETOED: NTSTATUS = NTSTATUS(-2147483608i32);
    pub const STATUS_PNP_BAD_MPS_TABLE: NTSTATUS = NTSTATUS(-1073479627i32);
    pub const STATUS_PNP_DEVICE_CONFIGURATION_PENDING: NTSTATUS = NTSTATUS(-1073740651i32);
    pub const STATUS_PNP_DRIVER_CONFIGURATION_INCOMPLETE: NTSTATUS = NTSTATUS(-1073740653i32);
    pub const STATUS_PNP_DRIVER_CONFIGURATION_NOT_FOUND: NTSTATUS = NTSTATUS(-1073740654i32);
    pub const STATUS_PNP_DRIVER_PACKAGE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073740655i32);
    pub const STATUS_PNP_FUNCTION_DRIVER_REQUIRED: NTSTATUS = NTSTATUS(-1073740652i32);
    pub const STATUS_PNP_INVALID_ID: NTSTATUS = NTSTATUS(-1073479624i32);
    pub const STATUS_PNP_IRQ_TRANSLATION_FAILED: NTSTATUS = NTSTATUS(-1073479625i32);
    pub const STATUS_PNP_NO_COMPAT_DRIVERS: NTSTATUS = NTSTATUS(-1073740656i32);
    pub const STATUS_PNP_REBOOT_REQUIRED: NTSTATUS = NTSTATUS(-1073741102i32);
    pub const STATUS_PNP_RESTART_ENUMERATION: NTSTATUS = NTSTATUS(-1073741106i32);
    pub const STATUS_PNP_TRANSLATION_FAILED: NTSTATUS = NTSTATUS(-1073479626i32);
    pub const STATUS_POLICY_OBJECT_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741158i32);
    pub const STATUS_POLICY_ONLY_IN_DS: NTSTATUS = NTSTATUS(-1073741157i32);
    pub const STATUS_PORT_ALREADY_HAS_COMPLETION_LIST: NTSTATUS = NTSTATUS(-1073740006i32);
    pub const STATUS_PORT_ALREADY_SET: NTSTATUS = NTSTATUS(-1073741752i32);
    pub const STATUS_PORT_CLOSED: NTSTATUS = NTSTATUS(-1073740032i32);
    pub const STATUS_PORT_CONNECTION_REFUSED: NTSTATUS = NTSTATUS(-1073741759i32);
    pub const STATUS_PORT_DISCONNECTED: NTSTATUS = NTSTATUS(-1073741769i32);
    pub const STATUS_PORT_DO_NOT_DISTURB: NTSTATUS = NTSTATUS(-1073741770i32);
    pub const STATUS_PORT_MESSAGE_TOO_LONG: NTSTATUS = NTSTATUS(-1073741777i32);
    pub const STATUS_PORT_NOT_SET: NTSTATUS = NTSTATUS(-1073740973i32);
    pub const STATUS_PORT_UNREACHABLE: NTSTATUS = NTSTATUS(-1073741249i32);
    pub const STATUS_POSSIBLE_DEADLOCK: NTSTATUS = NTSTATUS(-1073741420i32);
    pub const STATUS_POWER_STATE_INVALID: NTSTATUS = NTSTATUS(-1073741101i32);
    pub const STATUS_PREDEFINED_HANDLE: NTSTATUS = NTSTATUS(1073741846i32);
    pub const STATUS_PRENT4_MACHINE_ACCOUNT: NTSTATUS = NTSTATUS(-1073740969i32);
    pub const STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED: NTSTATUS = NTSTATUS(270i32);
    pub const STATUS_PRINT_CANCELLED: NTSTATUS = NTSTATUS(-1073741624i32);
    pub const STATUS_PRINT_QUEUE_FULL: NTSTATUS = NTSTATUS(-1073741626i32);
    pub const STATUS_PRIVILEGED_INSTRUCTION: NTSTATUS = NTSTATUS(-1073741674i32);
    pub const STATUS_PRIVILEGE_NOT_HELD: NTSTATUS = NTSTATUS(-1073741727i32);
    pub const STATUS_PROACTIVE_SCAN_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073739761i32);
    pub const STATUS_PROCEDURE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741702i32);
    pub const STATUS_PROCESS_CLONED: NTSTATUS = NTSTATUS(297i32);
    pub const STATUS_PROCESS_IN_JOB: NTSTATUS = NTSTATUS(292i32);
    pub const STATUS_PROCESS_IS_PROTECTED: NTSTATUS = NTSTATUS(-1073740014i32);
    pub const STATUS_PROCESS_IS_TERMINATING: NTSTATUS = NTSTATUS(-1073741558i32);
    pub const STATUS_PROCESS_NOT_IN_JOB: NTSTATUS = NTSTATUS(291i32);
    pub const STATUS_PROFILING_AT_LIMIT: NTSTATUS = NTSTATUS(-1073741613i32);
    pub const STATUS_PROFILING_NOT_STARTED: NTSTATUS = NTSTATUS(-1073741641i32);
    pub const STATUS_PROFILING_NOT_STOPPED: NTSTATUS = NTSTATUS(-1073741640i32);
    pub const STATUS_PROPSET_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741264i32);
    pub const STATUS_PROTOCOL_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073700845i32);
    pub const STATUS_PROTOCOL_UNREACHABLE: NTSTATUS = NTSTATUS(-1073741250i32);
    pub const STATUS_PTE_CHANGED: NTSTATUS = NTSTATUS(-1073740748i32);
    pub const STATUS_PURGE_FAILED: NTSTATUS = NTSTATUS(-1073740747i32);
    pub const STATUS_PWD_HISTORY_CONFLICT: NTSTATUS = NTSTATUS(-1073741220i32);
    pub const STATUS_PWD_TOO_LONG: NTSTATUS = NTSTATUS(-1073741190i32);
    pub const STATUS_PWD_TOO_RECENT: NTSTATUS = NTSTATUS(-1073741221i32);
    pub const STATUS_PWD_TOO_SHORT: NTSTATUS = NTSTATUS(-1073741222i32);
    pub const STATUS_QUERY_STORAGE_ERROR: NTSTATUS = NTSTATUS(-2143682559i32);
    pub const STATUS_QUIC_ALPN_NEG_FAILURE: NTSTATUS = NTSTATUS(-1071382521i32);
    pub const STATUS_QUIC_CONNECTION_IDLE: NTSTATUS = NTSTATUS(-1071382523i32);
    pub const STATUS_QUIC_CONNECTION_TIMEOUT: NTSTATUS = NTSTATUS(-1071382522i32);
    pub const STATUS_QUIC_HANDSHAKE_FAILURE: NTSTATUS = NTSTATUS(-1071382528i32);
    pub const STATUS_QUIC_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1071382525i32);
    pub const STATUS_QUIC_PROTOCOL_VIOLATION: NTSTATUS = NTSTATUS(-1071382524i32);
    pub const STATUS_QUIC_USER_CANCELED: NTSTATUS = NTSTATUS(-1071382526i32);
    pub const STATUS_QUIC_VER_NEG_FAILURE: NTSTATUS = NTSTATUS(-1071382527i32);
    pub const STATUS_QUOTA_ACTIVITY: NTSTATUS = NTSTATUS(-1073740662i32);
    pub const STATUS_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073741756i32);
    pub const STATUS_QUOTA_LIST_INCONSISTENT: NTSTATUS = NTSTATUS(-1073741210i32);
    pub const STATUS_QUOTA_NOT_ENABLED: NTSTATUS = NTSTATUS(-1073741399i32);
    pub const STATUS_RANGE_LIST_CONFLICT: NTSTATUS = NTSTATUS(-1073741182i32);
    pub const STATUS_RANGE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741172i32);
    pub const STATUS_RANGE_NOT_LOCKED: NTSTATUS = NTSTATUS(-1073741698i32);
    pub const STATUS_RDBSS_CONTINUE_OPERATION: NTSTATUS = NTSTATUS(-1069481982i32);
    pub const STATUS_RDBSS_POST_OPERATION: NTSTATUS = NTSTATUS(-1069481981i32);
    pub const STATUS_RDBSS_RESTART_OPERATION: NTSTATUS = NTSTATUS(-1069481983i32);
    pub const STATUS_RDBSS_RETRY_LOOKUP: NTSTATUS = NTSTATUS(-1069481980i32);
    pub const STATUS_RDP_PROTOCOL_ERROR: NTSTATUS = NTSTATUS(-1073086414i32);
    pub const STATUS_RECEIVE_EXPEDITED: NTSTATUS = NTSTATUS(1073741840i32);
    pub const STATUS_RECEIVE_PARTIAL: NTSTATUS = NTSTATUS(1073741839i32);
    pub const STATUS_RECEIVE_PARTIAL_EXPEDITED: NTSTATUS = NTSTATUS(1073741841i32);
    pub const STATUS_RECOVERABLE_BUGCHECK: NTSTATUS = NTSTATUS(-2147483596i32);
    pub const STATUS_RECOVERY_FAILURE: NTSTATUS = NTSTATUS(-1073741273i32);
    pub const STATUS_RECOVERY_NOT_NEEDED: NTSTATUS = NTSTATUS(1075380276i32);
    pub const STATUS_RECURSIVE_DISPATCH: NTSTATUS = NTSTATUS(-1073740028i32);
    pub const STATUS_REDIRECTOR_HAS_OPEN_HANDLES: NTSTATUS = NTSTATUS(-2147483613i32);
    pub const STATUS_REDIRECTOR_NOT_STARTED: NTSTATUS = NTSTATUS(-1073741573i32);
    pub const STATUS_REDIRECTOR_PAUSED: NTSTATUS = NTSTATUS(-1073741615i32);
    pub const STATUS_REDIRECTOR_STARTED: NTSTATUS = NTSTATUS(-1073741572i32);
    pub const STATUS_REGISTRY_CORRUPT: NTSTATUS = NTSTATUS(-1073741492i32);
    pub const STATUS_REGISTRY_HIVE_RECOVERED: NTSTATUS = NTSTATUS(-2147483606i32);
    pub const STATUS_REGISTRY_IO_FAILED: NTSTATUS = NTSTATUS(-1073741491i32);
    pub const STATUS_REGISTRY_QUOTA_LIMIT: NTSTATUS = NTSTATUS(-1073741226i32);
    pub const STATUS_REGISTRY_RECOVERED: NTSTATUS = NTSTATUS(1073741833i32);
    pub const STATUS_REG_NAT_CONSUMPTION: NTSTATUS = NTSTATUS(-1073741111i32);
    pub const STATUS_REINITIALIZATION_NEEDED: NTSTATUS = NTSTATUS(-1073741177i32);
    pub const STATUS_REMOTE_DISCONNECT: NTSTATUS = NTSTATUS(-1073741508i32);
    pub const STATUS_REMOTE_FILE_VERSION_MISMATCH: NTSTATUS = NTSTATUS(-1072103412i32);
    pub const STATUS_REMOTE_NOT_LISTENING: NTSTATUS = NTSTATUS(-1073741636i32);
    pub const STATUS_REMOTE_RESOURCES: NTSTATUS = NTSTATUS(-1073741507i32);
    pub const STATUS_REMOTE_SESSION_LIMIT: NTSTATUS = NTSTATUS(-1073741418i32);
    pub const STATUS_REMOTE_STORAGE_MEDIA_ERROR: NTSTATUS = NTSTATUS(-1073741154i32);
    pub const STATUS_REMOTE_STORAGE_NOT_ACTIVE: NTSTATUS = NTSTATUS(-1073741155i32);
    pub const STATUS_REPAIR_NEEDED: NTSTATUS = NTSTATUS(-1073741400i32);
    pub const STATUS_REPARSE: NTSTATUS = NTSTATUS(260i32);
    pub const STATUS_REPARSE_ATTRIBUTE_CONFLICT: NTSTATUS = NTSTATUS(-1073741134i32);
    pub const STATUS_REPARSE_GLOBAL: NTSTATUS = NTSTATUS(872i32);
    pub const STATUS_REPARSE_OBJECT: NTSTATUS = NTSTATUS(280i32);
    pub const STATUS_REPARSE_POINT_ENCOUNTERED: NTSTATUS = NTSTATUS(-1073740533i32);
    pub const STATUS_REPARSE_POINT_NOT_RESOLVED: NTSTATUS = NTSTATUS(-1073741184i32);
    pub const STATUS_REPLY_MESSAGE_MISMATCH: NTSTATUS = NTSTATUS(-1073741281i32);
    pub const STATUS_REQUEST_ABORTED: NTSTATUS = NTSTATUS(-1073741248i32);
    pub const STATUS_REQUEST_CANCELED: NTSTATUS = NTSTATUS(-1073740029i32);
    pub const STATUS_REQUEST_NOT_ACCEPTED: NTSTATUS = NTSTATUS(-1073741616i32);
    pub const STATUS_REQUEST_OUT_OF_SEQUENCE: NTSTATUS = NTSTATUS(-1073740758i32);
    pub const STATUS_REQUEST_PAUSED: NTSTATUS = NTSTATUS(-1073740711i32);
    pub const STATUS_RESIDENT_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740678i32);
    pub const STATUS_RESOURCEMANAGER_NOT_FOUND: NTSTATUS = NTSTATUS(-1072103345i32);
    pub const STATUS_RESOURCEMANAGER_READ_ONLY: NTSTATUS = NTSTATUS(514i32);
    pub const STATUS_RESOURCE_DATA_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741687i32);
    pub const STATUS_RESOURCE_ENUM_USER_STOP: NTSTATUS = NTSTATUS(-1073020921i32);
    pub const STATUS_RESOURCE_IN_USE: NTSTATUS = NTSTATUS(-1073740024i32);
    pub const STATUS_RESOURCE_LANG_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741308i32);
    pub const STATUS_RESOURCE_NAME_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741685i32);
    pub const STATUS_RESOURCE_NOT_OWNED: NTSTATUS = NTSTATUS(-1073741212i32);
    pub const STATUS_RESOURCE_REQUIREMENTS_CHANGED: NTSTATUS = NTSTATUS(281i32);
    pub const STATUS_RESOURCE_TYPE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741686i32);
    pub const STATUS_RESTART_BOOT_APPLICATION: NTSTATUS = NTSTATUS(-1073740717i32);
    pub const STATUS_RESUME_HIBERNATION: NTSTATUS = NTSTATUS(1073741867i32);
    pub const STATUS_RETRY: NTSTATUS = NTSTATUS(-1073741267i32);
    pub const STATUS_RETURN_ADDRESS_HIJACK_ATTEMPT: NTSTATUS = NTSTATUS(-2147483597i32);
    pub const STATUS_REVISION_MISMATCH: NTSTATUS = NTSTATUS(-1073741735i32);
    pub const STATUS_REVOCATION_OFFLINE_C: NTSTATUS = NTSTATUS(-1073740917i32);
    pub const STATUS_REVOCATION_OFFLINE_KDC: NTSTATUS = NTSTATUS(-1073740788i32);
    pub const STATUS_RING_NEWLY_EMPTY: NTSTATUS = NTSTATUS(531i32);
    pub const STATUS_RING_PREVIOUSLY_ABOVE_QUOTA: NTSTATUS = NTSTATUS(530i32);
    pub const STATUS_RING_PREVIOUSLY_EMPTY: NTSTATUS = NTSTATUS(528i32);
    pub const STATUS_RING_PREVIOUSLY_FULL: NTSTATUS = NTSTATUS(529i32);
    pub const STATUS_RING_SIGNAL_OPPOSITE_ENDPOINT: NTSTATUS = NTSTATUS(532i32);
    pub const STATUS_RKF_ACTIVE_KEY: NTSTATUS = NTSTATUS(-1069547514i32);
    pub const STATUS_RKF_BLOB_FULL: NTSTATUS = NTSTATUS(-1069547517i32);
    pub const STATUS_RKF_DUPLICATE_KEY: NTSTATUS = NTSTATUS(-1069547518i32);
    pub const STATUS_RKF_FILE_BLOCKED: NTSTATUS = NTSTATUS(-1069547515i32);
    pub const STATUS_RKF_KEY_NOT_FOUND: NTSTATUS = NTSTATUS(-1069547519i32);
    pub const STATUS_RKF_STORE_FULL: NTSTATUS = NTSTATUS(-1069547516i32);
    pub const STATUS_RM_ALREADY_STARTED: NTSTATUS = NTSTATUS(1075380277i32);
    pub const STATUS_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT: NTSTATUS = NTSTATUS(-1072103331i32);
    pub const STATUS_RM_DISCONNECTED: NTSTATUS = NTSTATUS(-1072103374i32);
    pub const STATUS_RM_METADATA_CORRUPT: NTSTATUS = NTSTATUS(-1072103418i32);
    pub const STATUS_RM_NOT_ACTIVE: NTSTATUS = NTSTATUS(-1072103419i32);
    pub const STATUS_ROLLBACK_TIMER_EXPIRED: NTSTATUS = NTSTATUS(-1072103364i32);
    pub const STATUS_RTPM_CONTEXT_COMPLETE: NTSTATUS = NTSTATUS(2699265i32);
    pub const STATUS_RTPM_CONTEXT_CONTINUE: NTSTATUS = NTSTATUS(2699264i32);
    pub const STATUS_RTPM_INVALID_CONTEXT: NTSTATUS = NTSTATUS(-1071042556i32);
    pub const STATUS_RTPM_NO_RESULT: NTSTATUS = NTSTATUS(-1071042558i32);
    pub const STATUS_RTPM_PCR_READ_INCOMPLETE: NTSTATUS = NTSTATUS(-1071042557i32);
    pub const STATUS_RTPM_UNSUPPORTED_CMD: NTSTATUS = NTSTATUS(-1071042555i32);
    pub const STATUS_RUNLEVEL_SWITCH_AGENT_TIMEOUT: NTSTATUS = NTSTATUS(-1073700539i32);
    pub const STATUS_RUNLEVEL_SWITCH_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073700538i32);
    pub const STATUS_RUNLEVEL_SWITCH_TIMEOUT: NTSTATUS = NTSTATUS(-1073700541i32);
    pub const STATUS_RWRAW_ENCRYPTED_FILE_NOT_ENCRYPTED: NTSTATUS = NTSTATUS(-1073740633i32);
    pub const STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILEOFFSET: NTSTATUS = NTSTATUS(-1073740632i32);
    pub const STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_FILERANGE: NTSTATUS = NTSTATUS(-1073740631i32);
    pub const STATUS_RWRAW_ENCRYPTED_INVALID_EDATAINFO_PARAMETER: NTSTATUS = NTSTATUS(-1073740630i32);
    pub const STATUS_RXACT_COMMITTED: NTSTATUS = NTSTATUS(266i32);
    pub const STATUS_RXACT_COMMIT_FAILURE: NTSTATUS = NTSTATUS(-1073741539i32);
    pub const STATUS_RXACT_COMMIT_NECESSARY: NTSTATUS = NTSTATUS(-2147483624i32);
    pub const STATUS_RXACT_INVALID_STATE: NTSTATUS = NTSTATUS(-1073741540i32);
    pub const STATUS_RXACT_STATE_CREATED: NTSTATUS = NTSTATUS(1073741828i32);
    pub const STATUS_SAM_INIT_FAILURE: NTSTATUS = NTSTATUS(-1073741085i32);
    pub const STATUS_SAM_NEED_BOOTKEY_FLOPPY: NTSTATUS = NTSTATUS(-1073741088i32);
    pub const STATUS_SAM_NEED_BOOTKEY_PASSWORD: NTSTATUS = NTSTATUS(-1073741089i32);
    pub const STATUS_SCRUB_DATA_DISABLED: NTSTATUS = NTSTATUS(-1073740680i32);
    pub const STATUS_SECCORE_INVALID_COMMAND: NTSTATUS = NTSTATUS(-1058537472i32);
    pub const STATUS_SECONDARY_IC_PROVIDER_NOT_REGISTERED: NTSTATUS = NTSTATUS(-1073700575i32);
    pub const STATUS_SECRET_TOO_LONG: NTSTATUS = NTSTATUS(-1073741481i32);
    pub const STATUS_SECTION_DIRECT_MAP_ONLY: NTSTATUS = NTSTATUS(-1073739503i32);
    pub const STATUS_SECTION_NOT_EXTENDED: NTSTATUS = NTSTATUS(-1073741689i32);
    pub const STATUS_SECTION_NOT_IMAGE: NTSTATUS = NTSTATUS(-1073741751i32);
    pub const STATUS_SECTION_PROTECTION: NTSTATUS = NTSTATUS(-1073741746i32);
    pub const STATUS_SECTION_TOO_BIG: NTSTATUS = NTSTATUS(-1073741760i32);
    pub const STATUS_SECUREBOOT_FILE_REPLACED: NTSTATUS = NTSTATUS(-1069350905i32);
    pub const STATUS_SECUREBOOT_INVALID_POLICY: NTSTATUS = NTSTATUS(-1069350909i32);
    pub const STATUS_SECUREBOOT_NOT_BASE_POLICY: NTSTATUS = NTSTATUS(-1069350897i32);
    pub const STATUS_SECUREBOOT_NOT_ENABLED: NTSTATUS = NTSTATUS(-2143092730i32);
    pub const STATUS_SECUREBOOT_NOT_SUPPLEMENTAL_POLICY: NTSTATUS = NTSTATUS(-1069350896i32);
    pub const STATUS_SECUREBOOT_PLATFORM_ID_MISMATCH: NTSTATUS = NTSTATUS(-1069350901i32);
    pub const STATUS_SECUREBOOT_POLICY_MISSING_ANTIROLLBACKVERSION: NTSTATUS = NTSTATUS(-1069350902i32);
    pub const STATUS_SECUREBOOT_POLICY_NOT_AUTHORIZED: NTSTATUS = NTSTATUS(-1069350904i32);
    pub const STATUS_SECUREBOOT_POLICY_NOT_SIGNED: NTSTATUS = NTSTATUS(-1069350907i32);
    pub const STATUS_SECUREBOOT_POLICY_PUBLISHER_NOT_FOUND: NTSTATUS = NTSTATUS(-1069350908i32);
    pub const STATUS_SECUREBOOT_POLICY_ROLLBACK_DETECTED: NTSTATUS = NTSTATUS(-1069350900i32);
    pub const STATUS_SECUREBOOT_POLICY_UNKNOWN: NTSTATUS = NTSTATUS(-1069350903i32);
    pub const STATUS_SECUREBOOT_POLICY_UPGRADE_MISMATCH: NTSTATUS = NTSTATUS(-1069350899i32);
    pub const STATUS_SECUREBOOT_POLICY_VIOLATION: NTSTATUS = NTSTATUS(-1069350910i32);
    pub const STATUS_SECUREBOOT_REQUIRED_POLICY_FILE_MISSING: NTSTATUS = NTSTATUS(-1069350898i32);
    pub const STATUS_SECUREBOOT_ROLLBACK_DETECTED: NTSTATUS = NTSTATUS(-1069350911i32);
    pub const STATUS_SECURITY_STREAM_IS_INCONSISTENT: NTSTATUS = NTSTATUS(-1073741408i32);
    pub const STATUS_SEGMENT_NOTIFICATION: NTSTATUS = NTSTATUS(1073741829i32);
    pub const STATUS_SEMAPHORE_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1073741753i32);
    pub const STATUS_SERIAL_COUNTER_TIMEOUT: NTSTATUS = NTSTATUS(1073741836i32);
    pub const STATUS_SERIAL_MORE_WRITES: NTSTATUS = NTSTATUS(1073741832i32);
    pub const STATUS_SERIAL_NO_DEVICE_INITED: NTSTATUS = NTSTATUS(-1073741488i32);
    pub const STATUS_SERVER_DISABLED: NTSTATUS = NTSTATUS(-1073741696i32);
    pub const STATUS_SERVER_HAS_OPEN_HANDLES: NTSTATUS = NTSTATUS(-2147483612i32);
    pub const STATUS_SERVER_NOT_DISABLED: NTSTATUS = NTSTATUS(-1073741695i32);
    pub const STATUS_SERVER_SHUTDOWN_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073741057i32);
    pub const STATUS_SERVER_SID_MISMATCH: NTSTATUS = NTSTATUS(-1073741152i32);
    pub const STATUS_SERVER_TRANSPORT_CONFLICT: NTSTATUS = NTSTATUS(-1073741388i32);
    pub const STATUS_SERVER_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073740698i32);
    pub const STATUS_SERVICES_FAILED_AUTOSTART: NTSTATUS = NTSTATUS(1073783108i32);
    pub const STATUS_SERVICE_NOTIFICATION: NTSTATUS = NTSTATUS(1073741848i32);
    pub const STATUS_SESSION_KEY_TOO_SHORT: NTSTATUS = NTSTATUS(-1073740521i32);
    pub const STATUS_SETMARK_DETECTED: NTSTATUS = NTSTATUS(-2147483615i32);
    pub const STATUS_SET_CONTEXT_DENIED: NTSTATUS = NTSTATUS(-1073740278i32);
    pub const STATUS_SHARED_IRQ_BUSY: NTSTATUS = NTSTATUS(-1073741460i32);
    pub const STATUS_SHARED_POLICY: NTSTATUS = NTSTATUS(-1073741159i32);
    pub const STATUS_SHARE_UNAVAILABLE: NTSTATUS = NTSTATUS(-1073740672i32);
    pub const STATUS_SHARING_PAUSED: NTSTATUS = NTSTATUS(-1073741617i32);
    pub const STATUS_SHARING_VIOLATION: NTSTATUS = NTSTATUS(-1073741757i32);
    pub const STATUS_SHORT_NAMES_NOT_ENABLED_ON_VOLUME: NTSTATUS = NTSTATUS(-1073741409i32);
    pub const STATUS_SHUTDOWN_IN_PROGRESS: NTSTATUS = NTSTATUS(-1073741058i32);
    pub const STATUS_SINGLE_STEP: NTSTATUS = NTSTATUS(-2147483644i32);
    pub const STATUS_SMARTCARD_CARD_BLOCKED: NTSTATUS = NTSTATUS(-1073740927i32);
    pub const STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED: NTSTATUS = NTSTATUS(-1073740926i32);
    pub const STATUS_SMARTCARD_CERT_EXPIRED: NTSTATUS = NTSTATUS(-1073740915i32);
    pub const STATUS_SMARTCARD_CERT_REVOKED: NTSTATUS = NTSTATUS(-1073740919i32);
    pub const STATUS_SMARTCARD_IO_ERROR: NTSTATUS = NTSTATUS(-1073740921i32);
    pub const STATUS_SMARTCARD_LOGON_REQUIRED: NTSTATUS = NTSTATUS(-1073741062i32);
    pub const STATUS_SMARTCARD_NO_CARD: NTSTATUS = NTSTATUS(-1073740925i32);
    pub const STATUS_SMARTCARD_NO_CERTIFICATE: NTSTATUS = NTSTATUS(-1073740923i32);
    pub const STATUS_SMARTCARD_NO_KEYSET: NTSTATUS = NTSTATUS(-1073740922i32);
    pub const STATUS_SMARTCARD_NO_KEY_CONTAINER: NTSTATUS = NTSTATUS(-1073740924i32);
    pub const STATUS_SMARTCARD_SILENT_CONTEXT: NTSTATUS = NTSTATUS(-1073740913i32);
    pub const STATUS_SMARTCARD_SUBSYSTEM_FAILURE: NTSTATUS = NTSTATUS(-1073741023i32);
    pub const STATUS_SMARTCARD_WRONG_PIN: NTSTATUS = NTSTATUS(-1073740928i32);
    pub const STATUS_SMB1_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1073740525i32);
    pub const STATUS_SMB_BAD_CLUSTER_DIALECT: NTSTATUS = NTSTATUS(-1067646975i32);
    pub const STATUS_SMB_GUEST_LOGON_BLOCKED: NTSTATUS = NTSTATUS(-1067646974i32);
    pub const STATUS_SMB_NO_PREAUTH_INTEGRITY_HASH_OVERLAP: NTSTATUS = NTSTATUS(-1067646976i32);
    pub const STATUS_SMB_NO_SIGNING_ALGORITHM_OVERLAP: NTSTATUS = NTSTATUS(-1067646973i32);
    pub const STATUS_SMI_PRIMITIVE_INSTALLER_FAILED: NTSTATUS = NTSTATUS(-1072365531i32);
    pub const STATUS_SMR_GARBAGE_COLLECTION_REQUIRED: NTSTATUS = NTSTATUS(-1073740524i32);
    pub const STATUS_SOME_NOT_MAPPED: NTSTATUS = NTSTATUS(263i32);
    pub const STATUS_SOURCE_ELEMENT_EMPTY: NTSTATUS = NTSTATUS(-1073741181i32);
    pub const STATUS_SPACES_ALLOCATION_SIZE_INVALID: NTSTATUS = NTSTATUS(-1058602994i32);
    pub const STATUS_SPACES_CACHE_FULL: NTSTATUS = NTSTATUS(-1058602970i32);
    pub const STATUS_SPACES_COMPLETE: NTSTATUS = NTSTATUS(15138818i32);
    pub const STATUS_SPACES_CORRUPT_METADATA: NTSTATUS = NTSTATUS(-1058602986i32);
    pub const STATUS_SPACES_DRIVE_LOST_DATA: NTSTATUS = NTSTATUS(-1058602979i32);
    pub const STATUS_SPACES_DRIVE_NOT_READY: NTSTATUS = NTSTATUS(-1058602981i32);
    pub const STATUS_SPACES_DRIVE_OPERATIONAL_STATE_INVALID: NTSTATUS = NTSTATUS(-1058602990i32);
    pub const STATUS_SPACES_DRIVE_REDUNDANCY_INVALID: NTSTATUS = NTSTATUS(-1058603002i32);
    pub const STATUS_SPACES_DRIVE_SECTOR_SIZE_INVALID: NTSTATUS = NTSTATUS(-1058603004i32);
    pub const STATUS_SPACES_DRIVE_SPLIT: NTSTATUS = NTSTATUS(-1058602980i32);
    pub const STATUS_SPACES_DRT_FULL: NTSTATUS = NTSTATUS(-1058602985i32);
    pub const STATUS_SPACES_ENCLOSURE_AWARE_INVALID: NTSTATUS = NTSTATUS(-1058602993i32);
    pub const STATUS_SPACES_ENTRY_INCOMPLETE: NTSTATUS = NTSTATUS(-1058602978i32);
    pub const STATUS_SPACES_ENTRY_INVALID: NTSTATUS = NTSTATUS(-1058602977i32);
    pub const STATUS_SPACES_EXTENDED_ERROR: NTSTATUS = NTSTATUS(-1058602996i32);
    pub const STATUS_SPACES_FAULT_DOMAIN_TYPE_INVALID: NTSTATUS = NTSTATUS(-1058603007i32);
    pub const STATUS_SPACES_FLUSH_METADATA: NTSTATUS = NTSTATUS(-1058602971i32);
    pub const STATUS_SPACES_INCONSISTENCY: NTSTATUS = NTSTATUS(-1058602984i32);
    pub const STATUS_SPACES_INTERLEAVE_LENGTH_INVALID: NTSTATUS = NTSTATUS(-1058602999i32);
    pub const STATUS_SPACES_LOG_NOT_READY: NTSTATUS = NTSTATUS(-1058602983i32);
    pub const STATUS_SPACES_MAP_REQUIRED: NTSTATUS = NTSTATUS(-1058602988i32);
    pub const STATUS_SPACES_MARK_DIRTY: NTSTATUS = NTSTATUS(-1058602976i32);
    pub const STATUS_SPACES_NOT_ENOUGH_DRIVES: NTSTATUS = NTSTATUS(-1058602997i32);
    pub const STATUS_SPACES_NO_REDUNDANCY: NTSTATUS = NTSTATUS(-1058602982i32);
    pub const STATUS_SPACES_NUMBER_OF_COLUMNS_INVALID: NTSTATUS = NTSTATUS(-1058602998i32);
    pub const STATUS_SPACES_NUMBER_OF_DATA_COPIES_INVALID: NTSTATUS = NTSTATUS(-1058603001i32);
    pub const STATUS_SPACES_NUMBER_OF_GROUPS_INVALID: NTSTATUS = NTSTATUS(-1058602991i32);
    pub const STATUS_SPACES_PAUSE: NTSTATUS = NTSTATUS(15138817i32);
    pub const STATUS_SPACES_PD_INVALID_DATA: NTSTATUS = NTSTATUS(-1058602972i32);
    pub const STATUS_SPACES_PD_LENGTH_MISMATCH: NTSTATUS = NTSTATUS(-1058602974i32);
    pub const STATUS_SPACES_PD_NOT_FOUND: NTSTATUS = NTSTATUS(-1058602975i32);
    pub const STATUS_SPACES_PD_UNSUPPORTED_VERSION: NTSTATUS = NTSTATUS(-1058602973i32);
    pub const STATUS_SPACES_PROVISIONING_TYPE_INVALID: NTSTATUS = NTSTATUS(-1058602995i32);
    pub const STATUS_SPACES_REDIRECT: NTSTATUS = NTSTATUS(15138819i32);
    pub const STATUS_SPACES_REPAIRED: NTSTATUS = NTSTATUS(15138816i32);
    pub const STATUS_SPACES_REPAIR_IN_PROGRESS: NTSTATUS = NTSTATUS(-1058602969i32);
    pub const STATUS_SPACES_RESILIENCY_TYPE_INVALID: NTSTATUS = NTSTATUS(-1058603005i32);
    pub const STATUS_SPACES_UNSUPPORTED_VERSION: NTSTATUS = NTSTATUS(-1058602987i32);
    pub const STATUS_SPACES_UPDATE_COLUMN_STATE: NTSTATUS = NTSTATUS(-1058602989i32);
    pub const STATUS_SPACES_WRITE_CACHE_SIZE_INVALID: NTSTATUS = NTSTATUS(-1058602992i32);
    pub const STATUS_SPARSE_FILE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740604i32);
    pub const STATUS_SPARSE_NOT_ALLOWED_IN_TRANSACTION: NTSTATUS = NTSTATUS(-1072103351i32);
    pub const STATUS_SPECIAL_ACCOUNT: NTSTATUS = NTSTATUS(-1073741532i32);
    pub const STATUS_SPECIAL_GROUP: NTSTATUS = NTSTATUS(-1073741531i32);
    pub const STATUS_SPECIAL_USER: NTSTATUS = NTSTATUS(-1073741530i32);
    pub const STATUS_STACK_BUFFER_OVERRUN: NTSTATUS = NTSTATUS(-1073740791i32);
    pub const STATUS_STACK_OVERFLOW: NTSTATUS = NTSTATUS(-1073741571i32);
    pub const STATUS_STACK_OVERFLOW_READ: NTSTATUS = NTSTATUS(-1073741272i32);
    pub const STATUS_STOPPED_ON_SYMLINK: NTSTATUS = NTSTATUS(-2147483603i32);
    pub const STATUS_STORAGE_LOST_DATA_PERSISTENCE: NTSTATUS = NTSTATUS(-1073740642i32);
    pub const STATUS_STORAGE_RESERVE_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1073740625i32);
    pub const STATUS_STORAGE_RESERVE_DOES_NOT_EXIST: NTSTATUS = NTSTATUS(-1073740626i32);
    pub const STATUS_STORAGE_RESERVE_ID_INVALID: NTSTATUS = NTSTATUS(-1073740627i32);
    pub const STATUS_STORAGE_RESERVE_NOT_EMPTY: NTSTATUS = NTSTATUS(-1073740624i32);
    pub const STATUS_STORAGE_STACK_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073740607i32);
    pub const STATUS_STORAGE_TOPOLOGY_ID_MISMATCH: NTSTATUS = NTSTATUS(-1073740666i32);
    pub const STATUS_STOWED_EXCEPTION: NTSTATUS = NTSTATUS(-1073741189i32);
    pub const STATUS_STREAM_MINIVERSION_NOT_FOUND: NTSTATUS = NTSTATUS(-1072103390i32);
    pub const STATUS_STREAM_MINIVERSION_NOT_VALID: NTSTATUS = NTSTATUS(-1072103389i32);
    pub const STATUS_STRICT_CFG_VIOLATION: NTSTATUS = NTSTATUS(-1073740282i32);
    pub const STATUS_STRONG_CRYPTO_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073741066i32);
    pub const STATUS_SUCCESS: NTSTATUS = NTSTATUS(0i32);
    pub const STATUS_SUSPEND_COUNT_EXCEEDED: NTSTATUS = NTSTATUS(-1073741750i32);
    pub const STATUS_SVHDX_ERROR_NOT_AVAILABLE: NTSTATUS = NTSTATUS(-1067647232i32);
    pub const STATUS_SVHDX_ERROR_STORED: NTSTATUS = NTSTATUS(-1067712512i32);
    pub const STATUS_SVHDX_NO_INITIATOR: NTSTATUS = NTSTATUS(-1067647221i32);
    pub const STATUS_SVHDX_RESERVATION_CONFLICT: NTSTATUS = NTSTATUS(-1067647225i32);
    pub const STATUS_SVHDX_UNIT_ATTENTION_AVAILABLE: NTSTATUS = NTSTATUS(-1067647231i32);
    pub const STATUS_SVHDX_UNIT_ATTENTION_CAPACITY_DATA_CHANGED: NTSTATUS = NTSTATUS(-1067647230i32);
    pub const STATUS_SVHDX_UNIT_ATTENTION_OPERATING_DEFINITION_CHANGED: NTSTATUS = NTSTATUS(-1067647226i32);
    pub const STATUS_SVHDX_UNIT_ATTENTION_REGISTRATIONS_PREEMPTED: NTSTATUS = NTSTATUS(-1067647227i32);
    pub const STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_PREEMPTED: NTSTATUS = NTSTATUS(-1067647229i32);
    pub const STATUS_SVHDX_UNIT_ATTENTION_RESERVATIONS_RELEASED: NTSTATUS = NTSTATUS(-1067647228i32);
    pub const STATUS_SVHDX_VERSION_MISMATCH: NTSTATUS = NTSTATUS(-1067647223i32);
    pub const STATUS_SVHDX_WRONG_FILE_TYPE: NTSTATUS = NTSTATUS(-1067647224i32);
    pub const STATUS_SXS_ACTIVATION_CONTEXT_DISABLED: NTSTATUS = NTSTATUS(-1072365561i32);
    pub const STATUS_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT: NTSTATUS = NTSTATUS(-1072365538i32);
    pub const STATUS_SXS_ASSEMBLY_MISSING: NTSTATUS = NTSTATUS(-1072365556i32);
    pub const STATUS_SXS_ASSEMBLY_NOT_FOUND: NTSTATUS = NTSTATUS(-1072365564i32);
    pub const STATUS_SXS_CANT_GEN_ACTCTX: NTSTATUS = NTSTATUS(-1072365566i32);
    pub const STATUS_SXS_COMPONENT_STORE_CORRUPT: NTSTATUS = NTSTATUS(-1072365542i32);
    pub const STATUS_SXS_CORRUPTION: NTSTATUS = NTSTATUS(-1072365547i32);
    pub const STATUS_SXS_CORRUPT_ACTIVATION_STACK: NTSTATUS = NTSTATUS(-1072365548i32);
    pub const STATUS_SXS_EARLY_DEACTIVATION: NTSTATUS = NTSTATUS(-1072365553i32);
    pub const STATUS_SXS_FILE_HASH_MISMATCH: NTSTATUS = NTSTATUS(-1072365541i32);
    pub const STATUS_SXS_FILE_HASH_MISSING: NTSTATUS = NTSTATUS(-1072365529i32);
    pub const STATUS_SXS_FILE_NOT_PART_OF_ASSEMBLY: NTSTATUS = NTSTATUS(-1072365537i32);
    pub const STATUS_SXS_IDENTITIES_DIFFERENT: NTSTATUS = NTSTATUS(-1072365539i32);
    pub const STATUS_SXS_IDENTITY_DUPLICATE_ATTRIBUTE: NTSTATUS = NTSTATUS(-1072365544i32);
    pub const STATUS_SXS_IDENTITY_PARSE_ERROR: NTSTATUS = NTSTATUS(-1072365543i32);
    pub const STATUS_SXS_INVALID_ACTCTXDATA_FORMAT: NTSTATUS = NTSTATUS(-1072365565i32);
    pub const STATUS_SXS_INVALID_DEACTIVATION: NTSTATUS = NTSTATUS(-1072365552i32);
    pub const STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME: NTSTATUS = NTSTATUS(-1072365545i32);
    pub const STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE: NTSTATUS = NTSTATUS(-1072365546i32);
    pub const STATUS_SXS_KEY_NOT_FOUND: NTSTATUS = NTSTATUS(-1072365560i32);
    pub const STATUS_SXS_MANIFEST_FORMAT_ERROR: NTSTATUS = NTSTATUS(-1072365563i32);
    pub const STATUS_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT: NTSTATUS = NTSTATUS(-1072365540i32);
    pub const STATUS_SXS_MANIFEST_PARSE_ERROR: NTSTATUS = NTSTATUS(-1072365562i32);
    pub const STATUS_SXS_MANIFEST_TOO_BIG: NTSTATUS = NTSTATUS(-1072365534i32);
    pub const STATUS_SXS_MULTIPLE_DEACTIVATION: NTSTATUS = NTSTATUS(-1072365551i32);
    pub const STATUS_SXS_PROCESS_DEFAULT_ALREADY_SET: NTSTATUS = NTSTATUS(-1072365554i32);
    pub const STATUS_SXS_PROCESS_TERMINATION_REQUESTED: NTSTATUS = NTSTATUS(-1072365549i32);
    pub const STATUS_SXS_RELEASE_ACTIVATION_CONTEXT: NTSTATUS = NTSTATUS(1075118093i32);
    pub const STATUS_SXS_SECTION_NOT_FOUND: NTSTATUS = NTSTATUS(-1072365567i32);
    pub const STATUS_SXS_SETTING_NOT_REGISTERED: NTSTATUS = NTSTATUS(-1072365533i32);
    pub const STATUS_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY: NTSTATUS = NTSTATUS(-1072365550i32);
    pub const STATUS_SXS_THREAD_QUERIES_DISABLED: NTSTATUS = NTSTATUS(-1072365557i32);
    pub const STATUS_SXS_TRANSACTION_CLOSURE_INCOMPLETE: NTSTATUS = NTSTATUS(-1072365532i32);
    pub const STATUS_SXS_VERSION_CONFLICT: NTSTATUS = NTSTATUS(-1072365559i32);
    pub const STATUS_SXS_WRONG_SECTION_TYPE: NTSTATUS = NTSTATUS(-1072365558i32);
    pub const STATUS_SYMLINK_CLASS_DISABLED: NTSTATUS = NTSTATUS(-1073740011i32);
    pub const STATUS_SYNCHRONIZATION_REQUIRED: NTSTATUS = NTSTATUS(-1073741516i32);
    pub const STATUS_SYSTEM_DEVICE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073740718i32);
    pub const STATUS_SYSTEM_HIVE_TOO_LARGE: NTSTATUS = NTSTATUS(-1073740946i32);
    pub const STATUS_SYSTEM_IMAGE_BAD_SIGNATURE: NTSTATUS = NTSTATUS(-1073741103i32);
    pub const STATUS_SYSTEM_INTEGRITY_INVALID_POLICY: NTSTATUS = NTSTATUS(-1058471933i32);
    pub const STATUS_SYSTEM_INTEGRITY_POLICY_NOT_SIGNED: NTSTATUS = NTSTATUS(-1058471932i32);
    pub const STATUS_SYSTEM_INTEGRITY_POLICY_VIOLATION: NTSTATUS = NTSTATUS(-1058471934i32);
    pub const STATUS_SYSTEM_INTEGRITY_REPUTATION_DANGEROUS_EXT: NTSTATUS = NTSTATUS(-1058471927i32);
    pub const STATUS_SYSTEM_INTEGRITY_REPUTATION_MALICIOUS: NTSTATUS = NTSTATUS(-1058471929i32);
    pub const STATUS_SYSTEM_INTEGRITY_REPUTATION_OFFLINE: NTSTATUS = NTSTATUS(-1058471926i32);
    pub const STATUS_SYSTEM_INTEGRITY_REPUTATION_PUA: NTSTATUS = NTSTATUS(-1058471928i32);
    pub const STATUS_SYSTEM_INTEGRITY_REPUTATION_UNATTAINABLE: NTSTATUS = NTSTATUS(-1058471924i32);
    pub const STATUS_SYSTEM_INTEGRITY_REPUTATION_UNFRIENDLY_FILE: NTSTATUS = NTSTATUS(-1058471925i32);
    pub const STATUS_SYSTEM_INTEGRITY_ROLLBACK_DETECTED: NTSTATUS = NTSTATUS(-1058471935i32);
    pub const STATUS_SYSTEM_INTEGRITY_SUPPLEMENTAL_POLICY_NOT_AUTHORIZED: NTSTATUS = NTSTATUS(-1058471930i32);
    pub const STATUS_SYSTEM_INTEGRITY_TOO_MANY_POLICIES: NTSTATUS = NTSTATUS(-1058471931i32);
    pub const STATUS_SYSTEM_NEEDS_REMEDIATION: NTSTATUS = NTSTATUS(-1073740674i32);
    pub const STATUS_SYSTEM_POWERSTATE_COMPLEX_TRANSITION: NTSTATUS = NTSTATUS(1073741873i32);
    pub const STATUS_SYSTEM_POWERSTATE_TRANSITION: NTSTATUS = NTSTATUS(1073741871i32);
    pub const STATUS_SYSTEM_PROCESS_TERMINATED: NTSTATUS = NTSTATUS(-1073741286i32);
    pub const STATUS_SYSTEM_SHUTDOWN: NTSTATUS = NTSTATUS(-1073741077i32);
    pub const STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED: NTSTATUS = NTSTATUS(-1073740018i32);
    pub const STATUS_THREADPOOL_HANDLE_EXCEPTION: NTSTATUS = NTSTATUS(-1073740022i32);
    pub const STATUS_THREADPOOL_RELEASED_DURING_OPERATION: NTSTATUS = NTSTATUS(-1073740017i32);
    pub const STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED: NTSTATUS = NTSTATUS(-1073740019i32);
    pub const STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED: NTSTATUS = NTSTATUS(-1073740020i32);
    pub const STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED: NTSTATUS = NTSTATUS(-1073740021i32);
    pub const STATUS_THREAD_ALREADY_IN_SESSION: NTSTATUS = NTSTATUS(-1073740714i32);
    pub const STATUS_THREAD_ALREADY_IN_TASK: NTSTATUS = NTSTATUS(-1073740542i32);
    pub const STATUS_THREAD_IS_TERMINATING: NTSTATUS = NTSTATUS(-1073741749i32);
    pub const STATUS_THREAD_NOT_IN_PROCESS: NTSTATUS = NTSTATUS(-1073741526i32);
    pub const STATUS_THREAD_NOT_IN_SESSION: NTSTATUS = NTSTATUS(-1073740713i32);
    pub const STATUS_THREAD_NOT_RUNNING: NTSTATUS = NTSTATUS(-1073740522i32);
    pub const STATUS_THREAD_WAS_SUSPENDED: NTSTATUS = NTSTATUS(1073741825i32);
    pub const STATUS_TIMEOUT: NTSTATUS = NTSTATUS(258i32);
    pub const STATUS_TIMER_NOT_CANCELED: NTSTATUS = NTSTATUS(-1073741812i32);
    pub const STATUS_TIMER_RESOLUTION_NOT_SET: NTSTATUS = NTSTATUS(-1073741243i32);
    pub const STATUS_TIMER_RESUME_IGNORED: NTSTATUS = NTSTATUS(1073741861i32);
    pub const STATUS_TIME_DIFFERENCE_AT_DC: NTSTATUS = NTSTATUS(-1073741517i32);
    pub const STATUS_TM_IDENTITY_MISMATCH: NTSTATUS = NTSTATUS(-1072103350i32);
    pub const STATUS_TM_INITIALIZATION_FAILED: NTSTATUS = NTSTATUS(-1072103420i32);
    pub const STATUS_TM_VOLATILE: NTSTATUS = NTSTATUS(-1072103365i32);
    pub const STATUS_TOKEN_ALREADY_IN_USE: NTSTATUS = NTSTATUS(-1073741525i32);
    pub const STATUS_TOO_LATE: NTSTATUS = NTSTATUS(-1073741431i32);
    pub const STATUS_TOO_MANY_ADDRESSES: NTSTATUS = NTSTATUS(-1073741303i32);
    pub const STATUS_TOO_MANY_COMMANDS: NTSTATUS = NTSTATUS(-1073741631i32);
    pub const STATUS_TOO_MANY_CONTEXT_IDS: NTSTATUS = NTSTATUS(-1073741478i32);
    pub const STATUS_TOO_MANY_GUIDS_REQUESTED: NTSTATUS = NTSTATUS(-1073741694i32);
    pub const STATUS_TOO_MANY_LINKS: NTSTATUS = NTSTATUS(-1073741211i32);
    pub const STATUS_TOO_MANY_LUIDS_REQUESTED: NTSTATUS = NTSTATUS(-1073741708i32);
    pub const STATUS_TOO_MANY_NAMES: NTSTATUS = NTSTATUS(-1073741619i32);
    pub const STATUS_TOO_MANY_NODES: NTSTATUS = NTSTATUS(-1073741298i32);
    pub const STATUS_TOO_MANY_OPENED_FILES: NTSTATUS = NTSTATUS(-1073741537i32);
    pub const STATUS_TOO_MANY_PAGING_FILES: NTSTATUS = NTSTATUS(-1073741673i32);
    pub const STATUS_TOO_MANY_PRINCIPALS: NTSTATUS = NTSTATUS(-1073741065i32);
    pub const STATUS_TOO_MANY_SECRETS: NTSTATUS = NTSTATUS(-1073741482i32);
    pub const STATUS_TOO_MANY_SEGMENT_DESCRIPTORS: NTSTATUS = NTSTATUS(-1073740685i32);
    pub const STATUS_TOO_MANY_SESSIONS: NTSTATUS = NTSTATUS(-1073741618i32);
    pub const STATUS_TOO_MANY_SIDS: NTSTATUS = NTSTATUS(-1073741442i32);
    pub const STATUS_TOO_MANY_THREADS: NTSTATUS = NTSTATUS(-1073741527i32);
    pub const STATUS_TPM_20_E_ASYMMETRIC: NTSTATUS = NTSTATUS(-1071054719i32);
    pub const STATUS_TPM_20_E_ATTRIBUTES: NTSTATUS = NTSTATUS(-1071054718i32);
    pub const STATUS_TPM_20_E_AUTHSIZE: NTSTATUS = NTSTATUS(-1071054524i32);
    pub const STATUS_TPM_20_E_AUTH_CONTEXT: NTSTATUS = NTSTATUS(-1071054523i32);
    pub const STATUS_TPM_20_E_AUTH_FAIL: NTSTATUS = NTSTATUS(-1071054706i32);
    pub const STATUS_TPM_20_E_AUTH_MISSING: NTSTATUS = NTSTATUS(-1071054555i32);
    pub const STATUS_TPM_20_E_AUTH_TYPE: NTSTATUS = NTSTATUS(-1071054556i32);
    pub const STATUS_TPM_20_E_AUTH_UNAVAILABLE: NTSTATUS = NTSTATUS(-1071054545i32);
    pub const STATUS_TPM_20_E_BAD_AUTH: NTSTATUS = NTSTATUS(-1071054686i32);
    pub const STATUS_TPM_20_E_BAD_CONTEXT: NTSTATUS = NTSTATUS(-1071054512i32);
    pub const STATUS_TPM_20_E_BINDING: NTSTATUS = NTSTATUS(-1071054683i32);
    pub const STATUS_TPM_20_E_COMMAND_CODE: NTSTATUS = NTSTATUS(-1071054525i32);
    pub const STATUS_TPM_20_E_COMMAND_SIZE: NTSTATUS = NTSTATUS(-1071054526i32);
    pub const STATUS_TPM_20_E_CPHASH: NTSTATUS = NTSTATUS(-1071054511i32);
    pub const STATUS_TPM_20_E_CURVE: NTSTATUS = NTSTATUS(-1071054682i32);
    pub const STATUS_TPM_20_E_DISABLED: NTSTATUS = NTSTATUS(-1071054560i32);
    pub const STATUS_TPM_20_E_ECC_CURVE: NTSTATUS = NTSTATUS(-1071054557i32);
    pub const STATUS_TPM_20_E_ECC_POINT: NTSTATUS = NTSTATUS(-1071054681i32);
    pub const STATUS_TPM_20_E_EXCLUSIVE: NTSTATUS = NTSTATUS(-1071054559i32);
    pub const STATUS_TPM_20_E_EXPIRED: NTSTATUS = NTSTATUS(-1071054685i32);
    pub const STATUS_TPM_20_E_FAILURE: NTSTATUS = NTSTATUS(-1071054591i32);
    pub const STATUS_TPM_20_E_HANDLE: NTSTATUS = NTSTATUS(-1071054709i32);
    pub const STATUS_TPM_20_E_HASH: NTSTATUS = NTSTATUS(-1071054717i32);
    pub const STATUS_TPM_20_E_HIERARCHY: NTSTATUS = NTSTATUS(-1071054715i32);
    pub const STATUS_TPM_20_E_HMAC: NTSTATUS = NTSTATUS(-1071054567i32);
    pub const STATUS_TPM_20_E_INITIALIZE: NTSTATUS = NTSTATUS(-1071054592i32);
    pub const STATUS_TPM_20_E_INSUFFICIENT: NTSTATUS = NTSTATUS(-1071054694i32);
    pub const STATUS_TPM_20_E_INTEGRITY: NTSTATUS = NTSTATUS(-1071054689i32);
    pub const STATUS_TPM_20_E_KDF: NTSTATUS = NTSTATUS(-1071054708i32);
    pub const STATUS_TPM_20_E_KEY: NTSTATUS = NTSTATUS(-1071054692i32);
    pub const STATUS_TPM_20_E_KEY_SIZE: NTSTATUS = NTSTATUS(-1071054713i32);
    pub const STATUS_TPM_20_E_MGF: NTSTATUS = NTSTATUS(-1071054712i32);
    pub const STATUS_TPM_20_E_MODE: NTSTATUS = NTSTATUS(-1071054711i32);
    pub const STATUS_TPM_20_E_NEEDS_TEST: NTSTATUS = NTSTATUS(-1071054509i32);
    pub const STATUS_TPM_20_E_NONCE: NTSTATUS = NTSTATUS(-1071054705i32);
    pub const STATUS_TPM_20_E_NO_RESULT: NTSTATUS = NTSTATUS(-1071054508i32);
    pub const STATUS_TPM_20_E_NV_AUTHORIZATION: NTSTATUS = NTSTATUS(-1071054519i32);
    pub const STATUS_TPM_20_E_NV_DEFINED: NTSTATUS = NTSTATUS(-1071054516i32);
    pub const STATUS_TPM_20_E_NV_LOCKED: NTSTATUS = NTSTATUS(-1071054520i32);
    pub const STATUS_TPM_20_E_NV_RANGE: NTSTATUS = NTSTATUS(-1071054522i32);
    pub const STATUS_TPM_20_E_NV_SIZE: NTSTATUS = NTSTATUS(-1071054521i32);
    pub const STATUS_TPM_20_E_NV_SPACE: NTSTATUS = NTSTATUS(-1071054517i32);
    pub const STATUS_TPM_20_E_NV_UNINITIALIZED: NTSTATUS = NTSTATUS(-1071054518i32);
    pub const STATUS_TPM_20_E_PARENT: NTSTATUS = NTSTATUS(-1071054510i32);
    pub const STATUS_TPM_20_E_PCR: NTSTATUS = NTSTATUS(-1071054553i32);
    pub const STATUS_TPM_20_E_PCR_CHANGED: NTSTATUS = NTSTATUS(-1071054552i32);
    pub const STATUS_TPM_20_E_POLICY: NTSTATUS = NTSTATUS(-1071054554i32);
    pub const STATUS_TPM_20_E_POLICY_CC: NTSTATUS = NTSTATUS(-1071054684i32);
    pub const STATUS_TPM_20_E_POLICY_FAIL: NTSTATUS = NTSTATUS(-1071054691i32);
    pub const STATUS_TPM_20_E_PP: NTSTATUS = NTSTATUS(-1071054704i32);
    pub const STATUS_TPM_20_E_PRIVATE: NTSTATUS = NTSTATUS(-1071054581i32);
    pub const STATUS_TPM_20_E_RANGE: NTSTATUS = NTSTATUS(-1071054707i32);
    pub const STATUS_TPM_20_E_REBOOT: NTSTATUS = NTSTATUS(-1071054544i32);
    pub const STATUS_TPM_20_E_RESERVED_BITS: NTSTATUS = NTSTATUS(-1071054687i32);
    pub const STATUS_TPM_20_E_SCHEME: NTSTATUS = NTSTATUS(-1071054702i32);
    pub const STATUS_TPM_20_E_SELECTOR: NTSTATUS = NTSTATUS(-1071054696i32);
    pub const STATUS_TPM_20_E_SENSITIVE: NTSTATUS = NTSTATUS(-1071054507i32);
    pub const STATUS_TPM_20_E_SEQUENCE: NTSTATUS = NTSTATUS(-1071054589i32);
    pub const STATUS_TPM_20_E_SIGNATURE: NTSTATUS = NTSTATUS(-1071054693i32);
    pub const STATUS_TPM_20_E_SIZE: NTSTATUS = NTSTATUS(-1071054699i32);
    pub const STATUS_TPM_20_E_SYMMETRIC: NTSTATUS = NTSTATUS(-1071054698i32);
    pub const STATUS_TPM_20_E_TAG: NTSTATUS = NTSTATUS(-1071054697i32);
    pub const STATUS_TPM_20_E_TICKET: NTSTATUS = NTSTATUS(-1071054688i32);
    pub const STATUS_TPM_20_E_TOO_MANY_CONTEXTS: NTSTATUS = NTSTATUS(-1071054546i32);
    pub const STATUS_TPM_20_E_TYPE: NTSTATUS = NTSTATUS(-1071054710i32);
    pub const STATUS_TPM_20_E_UNBALANCED: NTSTATUS = NTSTATUS(-1071054543i32);
    pub const STATUS_TPM_20_E_UPGRADE: NTSTATUS = NTSTATUS(-1071054547i32);
    pub const STATUS_TPM_20_E_VALUE: NTSTATUS = NTSTATUS(-1071054716i32);
    pub const STATUS_TPM_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1071050748i32);
    pub const STATUS_TPM_AREA_LOCKED: NTSTATUS = NTSTATUS(-1071054788i32);
    pub const STATUS_TPM_AUDITFAILURE: NTSTATUS = NTSTATUS(-1071054844i32);
    pub const STATUS_TPM_AUDITFAIL_SUCCESSFUL: NTSTATUS = NTSTATUS(-1071054799i32);
    pub const STATUS_TPM_AUDITFAIL_UNSUCCESSFUL: NTSTATUS = NTSTATUS(-1071054800i32);
    pub const STATUS_TPM_AUTH2FAIL: NTSTATUS = NTSTATUS(-1071054819i32);
    pub const STATUS_TPM_AUTHFAIL: NTSTATUS = NTSTATUS(-1071054847i32);
    pub const STATUS_TPM_AUTH_CONFLICT: NTSTATUS = NTSTATUS(-1071054789i32);
    pub const STATUS_TPM_BADCONTEXT: NTSTATUS = NTSTATUS(-1071054758i32);
    pub const STATUS_TPM_BADINDEX: NTSTATUS = NTSTATUS(-1071054846i32);
    pub const STATUS_TPM_BADTAG: NTSTATUS = NTSTATUS(-1071054818i32);
    pub const STATUS_TPM_BAD_ATTRIBUTES: NTSTATUS = NTSTATUS(-1071054782i32);
    pub const STATUS_TPM_BAD_COUNTER: NTSTATUS = NTSTATUS(-1071054779i32);
    pub const STATUS_TPM_BAD_DATASIZE: NTSTATUS = NTSTATUS(-1071054805i32);
    pub const STATUS_TPM_BAD_DELEGATE: NTSTATUS = NTSTATUS(-1071054759i32);
    pub const STATUS_TPM_BAD_HANDLE: NTSTATUS = NTSTATUS(-1071054760i32);
    pub const STATUS_TPM_BAD_KEY_PROPERTY: NTSTATUS = NTSTATUS(-1071054808i32);
    pub const STATUS_TPM_BAD_LOCALITY: NTSTATUS = NTSTATUS(-1071054787i32);
    pub const STATUS_TPM_BAD_MIGRATION: NTSTATUS = NTSTATUS(-1071054807i32);
    pub const STATUS_TPM_BAD_MODE: NTSTATUS = NTSTATUS(-1071054804i32);
    pub const STATUS_TPM_BAD_ORDINAL: NTSTATUS = NTSTATUS(-1071054838i32);
    pub const STATUS_TPM_BAD_PARAMETER: NTSTATUS = NTSTATUS(-1071054845i32);
    pub const STATUS_TPM_BAD_PARAM_SIZE: NTSTATUS = NTSTATUS(-1071054823i32);
    pub const STATUS_TPM_BAD_PRESENCE: NTSTATUS = NTSTATUS(-1071054803i32);
    pub const STATUS_TPM_BAD_SCHEME: NTSTATUS = NTSTATUS(-1071054806i32);
    pub const STATUS_TPM_BAD_SIGNATURE: NTSTATUS = NTSTATUS(-1071054750i32);
    pub const STATUS_TPM_BAD_TYPE: NTSTATUS = NTSTATUS(-1071054796i32);
    pub const STATUS_TPM_BAD_VERSION: NTSTATUS = NTSTATUS(-1071054802i32);
    pub const STATUS_TPM_CLEAR_DISABLED: NTSTATUS = NTSTATUS(-1071054843i32);
    pub const STATUS_TPM_COMMAND_BLOCKED: NTSTATUS = NTSTATUS(-1071053824i32);
    pub const STATUS_TPM_COMMAND_CANCELED: NTSTATUS = NTSTATUS(-1071050751i32);
    pub const STATUS_TPM_CONTEXT_GAP: NTSTATUS = NTSTATUS(-1071054777i32);
    pub const STATUS_TPM_DAA_INPUT_DATA0: NTSTATUS = NTSTATUS(-1071054767i32);
    pub const STATUS_TPM_DAA_INPUT_DATA1: NTSTATUS = NTSTATUS(-1071054766i32);
    pub const STATUS_TPM_DAA_ISSUER_SETTINGS: NTSTATUS = NTSTATUS(-1071054765i32);
    pub const STATUS_TPM_DAA_ISSUER_VALIDITY: NTSTATUS = NTSTATUS(-1071054762i32);
    pub const STATUS_TPM_DAA_RESOURCES: NTSTATUS = NTSTATUS(-1071054768i32);
    pub const STATUS_TPM_DAA_STAGE: NTSTATUS = NTSTATUS(-1071054763i32);
    pub const STATUS_TPM_DAA_TPM_SETTINGS: NTSTATUS = NTSTATUS(-1071054764i32);
    pub const STATUS_TPM_DAA_WRONG_W: NTSTATUS = NTSTATUS(-1071054761i32);
    pub const STATUS_TPM_DEACTIVATED: NTSTATUS = NTSTATUS(-1071054842i32);
    pub const STATUS_TPM_DECRYPT_ERROR: NTSTATUS = NTSTATUS(-1071054815i32);
    pub const STATUS_TPM_DEFEND_LOCK_RUNNING: NTSTATUS = NTSTATUS(-1071052797i32);
    pub const STATUS_TPM_DELEGATE_ADMIN: NTSTATUS = NTSTATUS(-1071054771i32);
    pub const STATUS_TPM_DELEGATE_FAMILY: NTSTATUS = NTSTATUS(-1071054772i32);
    pub const STATUS_TPM_DELEGATE_LOCK: NTSTATUS = NTSTATUS(-1071054773i32);
    pub const STATUS_TPM_DISABLED: NTSTATUS = NTSTATUS(-1071054841i32);
    pub const STATUS_TPM_DISABLED_CMD: NTSTATUS = NTSTATUS(-1071054840i32);
    pub const STATUS_TPM_DOING_SELFTEST: NTSTATUS = NTSTATUS(-1071052798i32);
    pub const STATUS_TPM_DUPLICATE_VHANDLE: NTSTATUS = NTSTATUS(-1071053822i32);
    pub const STATUS_TPM_EMBEDDED_COMMAND_BLOCKED: NTSTATUS = NTSTATUS(-1071053821i32);
    pub const STATUS_TPM_EMBEDDED_COMMAND_UNSUPPORTED: NTSTATUS = NTSTATUS(-1071053820i32);
    pub const STATUS_TPM_ENCRYPT_ERROR: NTSTATUS = NTSTATUS(-1071054816i32);
    pub const STATUS_TPM_ERROR_MASK: NTSTATUS = NTSTATUS(-1071054848i32);
    pub const STATUS_TPM_FAIL: NTSTATUS = NTSTATUS(-1071054839i32);
    pub const STATUS_TPM_FAILEDSELFTEST: NTSTATUS = NTSTATUS(-1071054820i32);
    pub const STATUS_TPM_FAMILYCOUNT: NTSTATUS = NTSTATUS(-1071054784i32);
    pub const STATUS_TPM_INAPPROPRIATE_ENC: NTSTATUS = NTSTATUS(-1071054834i32);
    pub const STATUS_TPM_INAPPROPRIATE_SIG: NTSTATUS = NTSTATUS(-1071054809i32);
    pub const STATUS_TPM_INSTALL_DISABLED: NTSTATUS = NTSTATUS(-1071054837i32);
    pub const STATUS_TPM_INSUFFICIENT_BUFFER: NTSTATUS = NTSTATUS(-1071050747i32);
    pub const STATUS_TPM_INVALID_AUTHHANDLE: NTSTATUS = NTSTATUS(-1071054814i32);
    pub const STATUS_TPM_INVALID_FAMILY: NTSTATUS = NTSTATUS(-1071054793i32);
    pub const STATUS_TPM_INVALID_HANDLE: NTSTATUS = NTSTATUS(-1071053823i32);
    pub const STATUS_TPM_INVALID_KEYHANDLE: NTSTATUS = NTSTATUS(-1071054836i32);
    pub const STATUS_TPM_INVALID_KEYUSAGE: NTSTATUS = NTSTATUS(-1071054812i32);
    pub const STATUS_TPM_INVALID_PCR_INFO: NTSTATUS = NTSTATUS(-1071054832i32);
    pub const STATUS_TPM_INVALID_POSTINIT: NTSTATUS = NTSTATUS(-1071054810i32);
    pub const STATUS_TPM_INVALID_RESOURCE: NTSTATUS = NTSTATUS(-1071054795i32);
    pub const STATUS_TPM_INVALID_STRUCTURE: NTSTATUS = NTSTATUS(-1071054781i32);
    pub const STATUS_TPM_IOERROR: NTSTATUS = NTSTATUS(-1071054817i32);
    pub const STATUS_TPM_KEYNOTFOUND: NTSTATUS = NTSTATUS(-1071054835i32);
    pub const STATUS_TPM_KEY_NOTSUPPORTED: NTSTATUS = NTSTATUS(-1071054790i32);
    pub const STATUS_TPM_KEY_OWNER_CONTROL: NTSTATUS = NTSTATUS(-1071054780i32);
    pub const STATUS_TPM_MAXNVWRITES: NTSTATUS = NTSTATUS(-1071054776i32);
    pub const STATUS_TPM_MA_AUTHORITY: NTSTATUS = NTSTATUS(-1071054753i32);
    pub const STATUS_TPM_MA_DESTINATION: NTSTATUS = NTSTATUS(-1071054755i32);
    pub const STATUS_TPM_MA_SOURCE: NTSTATUS = NTSTATUS(-1071054754i32);
    pub const STATUS_TPM_MA_TICKET_SIGNATURE: NTSTATUS = NTSTATUS(-1071054756i32);
    pub const STATUS_TPM_MIGRATEFAIL: NTSTATUS = NTSTATUS(-1071054833i32);
    pub const STATUS_TPM_NEEDS_SELFTEST: NTSTATUS = NTSTATUS(-1071052799i32);
    pub const STATUS_TPM_NOCONTEXTSPACE: NTSTATUS = NTSTATUS(-1071054749i32);
    pub const STATUS_TPM_NOOPERATOR: NTSTATUS = NTSTATUS(-1071054775i32);
    pub const STATUS_TPM_NOSPACE: NTSTATUS = NTSTATUS(-1071054831i32);
    pub const STATUS_TPM_NOSRK: NTSTATUS = NTSTATUS(-1071054830i32);
    pub const STATUS_TPM_NOTFIPS: NTSTATUS = NTSTATUS(-1071054794i32);
    pub const STATUS_TPM_NOTLOCAL: NTSTATUS = NTSTATUS(-1071054797i32);
    pub const STATUS_TPM_NOTRESETABLE: NTSTATUS = NTSTATUS(-1071054798i32);
    pub const STATUS_TPM_NOTSEALED_BLOB: NTSTATUS = NTSTATUS(-1071054829i32);
    pub const STATUS_TPM_NOT_FOUND: NTSTATUS = NTSTATUS(-1071050749i32);
    pub const STATUS_TPM_NOT_FULLWRITE: NTSTATUS = NTSTATUS(-1071054778i32);
    pub const STATUS_TPM_NO_ENDORSEMENT: NTSTATUS = NTSTATUS(-1071054813i32);
    pub const STATUS_TPM_NO_NV_PERMISSION: NTSTATUS = NTSTATUS(-1071054792i32);
    pub const STATUS_TPM_NO_WRAP_TRANSPORT: NTSTATUS = NTSTATUS(-1071054801i32);
    pub const STATUS_TPM_OWNER_CONTROL: NTSTATUS = NTSTATUS(-1071054769i32);
    pub const STATUS_TPM_OWNER_SET: NTSTATUS = NTSTATUS(-1071054828i32);
    pub const STATUS_TPM_PERMANENTEK: NTSTATUS = NTSTATUS(-1071054751i32);
    pub const STATUS_TPM_PER_NOWRITE: NTSTATUS = NTSTATUS(-1071054785i32);
    pub const STATUS_TPM_PPI_FUNCTION_UNSUPPORTED: NTSTATUS = NTSTATUS(-1071050746i32);
    pub const STATUS_TPM_READ_ONLY: NTSTATUS = NTSTATUS(-1071054786i32);
    pub const STATUS_TPM_REQUIRES_SIGN: NTSTATUS = NTSTATUS(-1071054791i32);
    pub const STATUS_TPM_RESOURCEMISSING: NTSTATUS = NTSTATUS(-1071054774i32);
    pub const STATUS_TPM_RESOURCES: NTSTATUS = NTSTATUS(-1071054827i32);
    pub const STATUS_TPM_RETRY: NTSTATUS = NTSTATUS(-1071052800i32);
    pub const STATUS_TPM_SHA_ERROR: NTSTATUS = NTSTATUS(-1071054821i32);
    pub const STATUS_TPM_SHA_THREAD: NTSTATUS = NTSTATUS(-1071054822i32);
    pub const STATUS_TPM_SHORTRANDOM: NTSTATUS = NTSTATUS(-1071054826i32);
    pub const STATUS_TPM_SIZE: NTSTATUS = NTSTATUS(-1071054825i32);
    pub const STATUS_TPM_TOOMANYCONTEXTS: NTSTATUS = NTSTATUS(-1071054757i32);
    pub const STATUS_TPM_TOO_MANY_CONTEXTS: NTSTATUS = NTSTATUS(-1071050750i32);
    pub const STATUS_TPM_TRANSPORT_NOTEXCLUSIVE: NTSTATUS = NTSTATUS(-1071054770i32);
    pub const STATUS_TPM_WRITE_LOCKED: NTSTATUS = NTSTATUS(-1071054783i32);
    pub const STATUS_TPM_WRONGPCRVAL: NTSTATUS = NTSTATUS(-1071054824i32);
    pub const STATUS_TPM_WRONG_ENTITYTYPE: NTSTATUS = NTSTATUS(-1071054811i32);
    pub const STATUS_TPM_ZERO_EXHAUST_ENABLED: NTSTATUS = NTSTATUS(-1071038464i32);
    pub const STATUS_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE: NTSTATUS = NTSTATUS(-1072103360i32);
    pub const STATUS_TRANSACTIONAL_CONFLICT: NTSTATUS = NTSTATUS(-1072103423i32);
    pub const STATUS_TRANSACTIONAL_OPEN_NOT_ALLOWED: NTSTATUS = NTSTATUS(-1072103361i32);
    pub const STATUS_TRANSACTIONMANAGER_IDENTITY_MISMATCH: NTSTATUS = NTSTATUS(-1072103332i32);
    pub const STATUS_TRANSACTIONMANAGER_NOT_FOUND: NTSTATUS = NTSTATUS(-1072103343i32);
    pub const STATUS_TRANSACTIONMANAGER_NOT_ONLINE: NTSTATUS = NTSTATUS(-1072103342i32);
    pub const STATUS_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION: NTSTATUS = NTSTATUS(-1072103341i32);
    pub const STATUS_TRANSACTIONS_NOT_FROZEN: NTSTATUS = NTSTATUS(-1072103355i32);
    pub const STATUS_TRANSACTIONS_UNSUPPORTED_REMOTE: NTSTATUS = NTSTATUS(-1072103414i32);
    pub const STATUS_TRANSACTION_ABORTED: NTSTATUS = NTSTATUS(-1073741297i32);
    pub const STATUS_TRANSACTION_ALREADY_ABORTED: NTSTATUS = NTSTATUS(-1072103403i32);
    pub const STATUS_TRANSACTION_ALREADY_COMMITTED: NTSTATUS = NTSTATUS(-1072103402i32);
    pub const STATUS_TRANSACTION_FREEZE_IN_PROGRESS: NTSTATUS = NTSTATUS(-1072103354i32);
    pub const STATUS_TRANSACTION_INTEGRITY_VIOLATED: NTSTATUS = NTSTATUS(-1072103333i32);
    pub const STATUS_TRANSACTION_INVALID_ID: NTSTATUS = NTSTATUS(-1073741292i32);
    pub const STATUS_TRANSACTION_INVALID_MARSHALL_BUFFER: NTSTATUS = NTSTATUS(-1072103401i32);
    pub const STATUS_TRANSACTION_INVALID_TYPE: NTSTATUS = NTSTATUS(-1073741291i32);
    pub const STATUS_TRANSACTION_MUST_WRITETHROUGH: NTSTATUS = NTSTATUS(-1072103330i32);
    pub const STATUS_TRANSACTION_NOT_ACTIVE: NTSTATUS = NTSTATUS(-1072103421i32);
    pub const STATUS_TRANSACTION_NOT_ENLISTED: NTSTATUS = NTSTATUS(-1072103327i32);
    pub const STATUS_TRANSACTION_NOT_FOUND: NTSTATUS = NTSTATUS(-1072103346i32);
    pub const STATUS_TRANSACTION_NOT_JOINED: NTSTATUS = NTSTATUS(-1072103417i32);
    pub const STATUS_TRANSACTION_NOT_REQUESTED: NTSTATUS = NTSTATUS(-1072103404i32);
    pub const STATUS_TRANSACTION_NOT_ROOT: NTSTATUS = NTSTATUS(-1072103340i32);
    pub const STATUS_TRANSACTION_NO_MATCH: NTSTATUS = NTSTATUS(-1073741294i32);
    pub const STATUS_TRANSACTION_NO_RELEASE: NTSTATUS = NTSTATUS(-1073741295i32);
    pub const STATUS_TRANSACTION_NO_SUPERIOR: NTSTATUS = NTSTATUS(-1072103329i32);
    pub const STATUS_TRANSACTION_OBJECT_EXPIRED: NTSTATUS = NTSTATUS(-1072103339i32);
    pub const STATUS_TRANSACTION_PROPAGATION_FAILED: NTSTATUS = NTSTATUS(-1072103408i32);
    pub const STATUS_TRANSACTION_RECORD_TOO_LONG: NTSTATUS = NTSTATUS(-1072103336i32);
    pub const STATUS_TRANSACTION_REQUEST_NOT_VALID: NTSTATUS = NTSTATUS(-1072103405i32);
    pub const STATUS_TRANSACTION_REQUIRED_PROMOTION: NTSTATUS = NTSTATUS(-1072103357i32);
    pub const STATUS_TRANSACTION_RESPONDED: NTSTATUS = NTSTATUS(-1073741293i32);
    pub const STATUS_TRANSACTION_RESPONSE_NOT_ENLISTED: NTSTATUS = NTSTATUS(-1072103337i32);
    pub const STATUS_TRANSACTION_SCOPE_CALLBACKS_NOT_SET: NTSTATUS = NTSTATUS(-2145845182i32);
    pub const STATUS_TRANSACTION_SUPERIOR_EXISTS: NTSTATUS = NTSTATUS(-1072103406i32);
    pub const STATUS_TRANSACTION_TIMED_OUT: NTSTATUS = NTSTATUS(-1073741296i32);
    pub const STATUS_TRANSLATION_COMPLETE: NTSTATUS = NTSTATUS(288i32);
    pub const STATUS_TRANSPORT_FULL: NTSTATUS = NTSTATUS(-1073741110i32);
    pub const STATUS_TRIGGERED_EXECUTABLE_MEMORY_WRITE: NTSTATUS = NTSTATUS(-1073739994i32);
    pub const STATUS_TRIM_READ_ZERO_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740686i32);
    pub const STATUS_TRUSTED_DOMAIN_FAILURE: NTSTATUS = NTSTATUS(-1073741428i32);
    pub const STATUS_TRUSTED_RELATIONSHIP_FAILURE: NTSTATUS = NTSTATUS(-1073741427i32);
    pub const STATUS_TRUST_FAILURE: NTSTATUS = NTSTATUS(-1073741424i32);
    pub const STATUS_TS_INCOMPATIBLE_SESSIONS: NTSTATUS = NTSTATUS(-1073086407i32);
    pub const STATUS_TS_VIDEO_SUBSYSTEM_ERROR: NTSTATUS = NTSTATUS(-1073086406i32);
    pub const STATUS_TXF_ATTRIBUTE_CORRUPT: NTSTATUS = NTSTATUS(-1072103363i32);
    pub const STATUS_TXF_DIR_NOT_EMPTY: NTSTATUS = NTSTATUS(-1072103367i32);
    pub const STATUS_TXF_METADATA_ALREADY_PRESENT: NTSTATUS = NTSTATUS(-2145845183i32);
    pub const STATUS_UNABLE_TO_DECOMMIT_VM: NTSTATUS = NTSTATUS(-1073741780i32);
    pub const STATUS_UNABLE_TO_DELETE_SECTION: NTSTATUS = NTSTATUS(-1073741797i32);
    pub const STATUS_UNABLE_TO_FREE_VM: NTSTATUS = NTSTATUS(-1073741798i32);
    pub const STATUS_UNABLE_TO_LOCK_MEDIA: NTSTATUS = NTSTATUS(-1073741451i32);
    pub const STATUS_UNABLE_TO_UNLOAD_MEDIA: NTSTATUS = NTSTATUS(-1073741450i32);
    pub const STATUS_UNDEFINED_CHARACTER: NTSTATUS = NTSTATUS(-1073741469i32);
    pub const STATUS_UNDEFINED_SCOPE: NTSTATUS = NTSTATUS(-1073740540i32);
    pub const STATUS_UNEXPECTED_IO_ERROR: NTSTATUS = NTSTATUS(-1073741591i32);
    pub const STATUS_UNEXPECTED_MM_CREATE_ERR: NTSTATUS = NTSTATUS(-1073741590i32);
    pub const STATUS_UNEXPECTED_MM_EXTEND_ERR: NTSTATUS = NTSTATUS(-1073741588i32);
    pub const STATUS_UNEXPECTED_MM_MAP_ERROR: NTSTATUS = NTSTATUS(-1073741589i32);
    pub const STATUS_UNEXPECTED_NETWORK_ERROR: NTSTATUS = NTSTATUS(-1073741628i32);
    pub const STATUS_UNFINISHED_CONTEXT_DELETED: NTSTATUS = NTSTATUS(-1073741074i32);
    pub const STATUS_UNHANDLED_EXCEPTION: NTSTATUS = NTSTATUS(-1073741500i32);
    pub const STATUS_UNKNOWN_REVISION: NTSTATUS = NTSTATUS(-1073741736i32);
    pub const STATUS_UNMAPPABLE_CHARACTER: NTSTATUS = NTSTATUS(-1073741470i32);
    pub const STATUS_UNRECOGNIZED_MEDIA: NTSTATUS = NTSTATUS(-1073741804i32);
    pub const STATUS_UNRECOGNIZED_VOLUME: NTSTATUS = NTSTATUS(-1073741489i32);
    pub const STATUS_UNSATISFIED_DEPENDENCIES: NTSTATUS = NTSTATUS(-1073740615i32);
    pub const STATUS_UNSUCCESSFUL: NTSTATUS = NTSTATUS(-1073741823i32);
    pub const STATUS_UNSUPPORTED_COMPRESSION: NTSTATUS = NTSTATUS(-1073741217i32);
    pub const STATUS_UNSUPPORTED_PAGING_MODE: NTSTATUS = NTSTATUS(-1073740613i32);
    pub const STATUS_UNSUPPORTED_PREAUTH: NTSTATUS = NTSTATUS(-1073740975i32);
    pub const STATUS_UNTRUSTED_MOUNT_POINT: NTSTATUS = NTSTATUS(-1073740612i32);
    pub const STATUS_UNWIND: NTSTATUS = NTSTATUS(-1073741785i32);
    pub const STATUS_UNWIND_CONSOLIDATE: NTSTATUS = NTSTATUS(-2147483607i32);
    pub const STATUS_USER2USER_REQUIRED: NTSTATUS = NTSTATUS(-1073740792i32);
    pub const STATUS_USER_APC: NTSTATUS = NTSTATUS(192i32);
    pub const STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED: NTSTATUS = NTSTATUS(-1073740797i32);
    pub const STATUS_USER_EXISTS: NTSTATUS = NTSTATUS(-1073741725i32);
    pub const STATUS_USER_MAPPED_FILE: NTSTATUS = NTSTATUS(-1073741245i32);
    pub const STATUS_USER_SESSION_DELETED: NTSTATUS = NTSTATUS(-1073741309i32);
    pub const STATUS_VALIDATE_CONTINUE: NTSTATUS = NTSTATUS(-1073741199i32);
    pub const STATUS_VALID_CATALOG_HASH: NTSTATUS = NTSTATUS(301i32);
    pub const STATUS_VALID_IMAGE_HASH: NTSTATUS = NTSTATUS(300i32);
    pub const STATUS_VALID_STRONG_CODE_HASH: NTSTATUS = NTSTATUS(302i32);
    pub const STATUS_VARIABLE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741568i32);
    pub const STATUS_VDM_DISALLOWED: NTSTATUS = NTSTATUS(-1073740780i32);
    pub const STATUS_VDM_HARD_ERROR: NTSTATUS = NTSTATUS(-1073741283i32);
    pub const STATUS_VERIFIER_STOP: NTSTATUS = NTSTATUS(-1073740767i32);
    pub const STATUS_VERIFY_REQUIRED: NTSTATUS = NTSTATUS(-2147483626i32);
    pub const STATUS_VHDSET_BACKING_STORAGE_NOT_FOUND: NTSTATUS = NTSTATUS(-1067647220i32);
    pub const STATUS_VHD_ALREADY_AT_OR_BELOW_MINIMUM_VIRTUAL_SIZE: NTSTATUS = NTSTATUS(-1069940685i32);
    pub const STATUS_VHD_BITMAP_MISMATCH: NTSTATUS = NTSTATUS(-1069940724i32);
    pub const STATUS_VHD_BLOCK_ALLOCATION_FAILURE: NTSTATUS = NTSTATUS(-1069940727i32);
    pub const STATUS_VHD_BLOCK_ALLOCATION_TABLE_CORRUPT: NTSTATUS = NTSTATUS(-1069940726i32);
    pub const STATUS_VHD_CHANGE_TRACKING_DISABLED: NTSTATUS = NTSTATUS(-1069940694i32);
    pub const STATUS_VHD_CHILD_PARENT_ID_MISMATCH: NTSTATUS = NTSTATUS(-1069940722i32);
    pub const STATUS_VHD_CHILD_PARENT_SIZE_MISMATCH: NTSTATUS = NTSTATUS(-1069940713i32);
    pub const STATUS_VHD_CHILD_PARENT_TIMESTAMP_MISMATCH: NTSTATUS = NTSTATUS(-1069940721i32);
    pub const STATUS_VHD_COULD_NOT_COMPUTE_MINIMUM_VIRTUAL_SIZE: NTSTATUS = NTSTATUS(-1069940686i32);
    pub const STATUS_VHD_DIFFERENCING_CHAIN_CYCLE_DETECTED: NTSTATUS = NTSTATUS(-1069940712i32);
    pub const STATUS_VHD_DIFFERENCING_CHAIN_ERROR_IN_PARENT: NTSTATUS = NTSTATUS(-1069940711i32);
    pub const STATUS_VHD_DRIVE_FOOTER_CHECKSUM_MISMATCH: NTSTATUS = NTSTATUS(-1069940734i32);
    pub const STATUS_VHD_DRIVE_FOOTER_CORRUPT: NTSTATUS = NTSTATUS(-1069940733i32);
    pub const STATUS_VHD_DRIVE_FOOTER_MISSING: NTSTATUS = NTSTATUS(-1069940735i32);
    pub const STATUS_VHD_FORMAT_UNKNOWN: NTSTATUS = NTSTATUS(-1069940732i32);
    pub const STATUS_VHD_FORMAT_UNSUPPORTED_VERSION: NTSTATUS = NTSTATUS(-1069940731i32);
    pub const STATUS_VHD_INVALID_BLOCK_SIZE: NTSTATUS = NTSTATUS(-1069940725i32);
    pub const STATUS_VHD_INVALID_CHANGE_TRACKING_ID: NTSTATUS = NTSTATUS(-1069940695i32);
    pub const STATUS_VHD_INVALID_FILE_SIZE: NTSTATUS = NTSTATUS(-1069940717i32);
    pub const STATUS_VHD_INVALID_SIZE: NTSTATUS = NTSTATUS(-1069940718i32);
    pub const STATUS_VHD_INVALID_STATE: NTSTATUS = NTSTATUS(-1069940708i32);
    pub const STATUS_VHD_INVALID_TYPE: NTSTATUS = NTSTATUS(-1069940709i32);
    pub const STATUS_VHD_METADATA_FULL: NTSTATUS = NTSTATUS(-1069940696i32);
    pub const STATUS_VHD_METADATA_READ_FAILURE: NTSTATUS = NTSTATUS(-1069940720i32);
    pub const STATUS_VHD_METADATA_WRITE_FAILURE: NTSTATUS = NTSTATUS(-1069940719i32);
    pub const STATUS_VHD_MISSING_CHANGE_TRACKING_INFORMATION: NTSTATUS = NTSTATUS(-1069940688i32);
    pub const STATUS_VHD_PARENT_VHD_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1069940714i32);
    pub const STATUS_VHD_PARENT_VHD_NOT_FOUND: NTSTATUS = NTSTATUS(-1069940723i32);
    pub const STATUS_VHD_RESIZE_WOULD_TRUNCATE_DATA: NTSTATUS = NTSTATUS(-1069940687i32);
    pub const STATUS_VHD_SHARED: NTSTATUS = NTSTATUS(-1067647222i32);
    pub const STATUS_VHD_SPARSE_HEADER_CHECKSUM_MISMATCH: NTSTATUS = NTSTATUS(-1069940730i32);
    pub const STATUS_VHD_SPARSE_HEADER_CORRUPT: NTSTATUS = NTSTATUS(-1069940728i32);
    pub const STATUS_VHD_SPARSE_HEADER_UNSUPPORTED_VERSION: NTSTATUS = NTSTATUS(-1069940729i32);
    pub const STATUS_VIDEO_DRIVER_DEBUG_REPORT_REQUEST: NTSTATUS = NTSTATUS(1075511532i32);
    pub const STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD: NTSTATUS = NTSTATUS(-1071972118i32);
    pub const STATUS_VIDEO_HUNG_DISPLAY_DRIVER_THREAD_RECOVERED: NTSTATUS = NTSTATUS(-2145713941i32);
    pub const STATUS_VID_CHILD_GPA_PAGE_SET_CORRUPTED: NTSTATUS = NTSTATUS(-1070137330i32);
    pub const STATUS_VID_DUPLICATE_HANDLER: NTSTATUS = NTSTATUS(-1070137343i32);
    pub const STATUS_VID_EXCEEDED_KM_CONTEXT_COUNT_LIMIT: NTSTATUS = NTSTATUS(-1070137314i32);
    pub const STATUS_VID_EXCEEDED_MBP_ENTRY_MAP_LIMIT: NTSTATUS = NTSTATUS(-1070137332i32);
    pub const STATUS_VID_HANDLER_NOT_PRESENT: NTSTATUS = NTSTATUS(-1070137340i32);
    pub const STATUS_VID_INSUFFICIENT_RESOURCES_HV_DEPOSIT: NTSTATUS = NTSTATUS(-1070137299i32);
    pub const STATUS_VID_INSUFFICIENT_RESOURCES_PHYSICAL_BUFFER: NTSTATUS = NTSTATUS(-1070137300i32);
    pub const STATUS_VID_INSUFFICIENT_RESOURCES_RESERVE: NTSTATUS = NTSTATUS(-1070137301i32);
    pub const STATUS_VID_INSUFFICIENT_RESOURCES_WITHDRAW: NTSTATUS = NTSTATUS(-1070137297i32);
    pub const STATUS_VID_INVALID_CHILD_GPA_PAGE_SET: NTSTATUS = NTSTATUS(-1070137310i32);
    pub const STATUS_VID_INVALID_GPA_RANGE_HANDLE: NTSTATUS = NTSTATUS(-1070137323i32);
    pub const STATUS_VID_INVALID_MEMORY_BLOCK_HANDLE: NTSTATUS = NTSTATUS(-1070137326i32);
    pub const STATUS_VID_INVALID_MESSAGE_QUEUE_HANDLE: NTSTATUS = NTSTATUS(-1070137324i32);
    pub const STATUS_VID_INVALID_NUMA_NODE_INDEX: NTSTATUS = NTSTATUS(-1070137328i32);
    pub const STATUS_VID_INVALID_NUMA_SETTINGS: NTSTATUS = NTSTATUS(-1070137329i32);
    pub const STATUS_VID_INVALID_OBJECT_NAME: NTSTATUS = NTSTATUS(-1070137339i32);
    pub const STATUS_VID_INVALID_PPM_HANDLE: NTSTATUS = NTSTATUS(-1070137320i32);
    pub const STATUS_VID_INVALID_PROCESSOR_STATE: NTSTATUS = NTSTATUS(-1070137315i32);
    pub const STATUS_VID_KM_INTERFACE_ALREADY_INITIALIZED: NTSTATUS = NTSTATUS(-1070137313i32);
    pub const STATUS_VID_MBPS_ARE_LOCKED: NTSTATUS = NTSTATUS(-1070137319i32);
    pub const STATUS_VID_MBP_ALREADY_LOCKED_USING_RESERVED_PAGE: NTSTATUS = NTSTATUS(-1070137307i32);
    pub const STATUS_VID_MBP_COUNT_EXCEEDED_LIMIT: NTSTATUS = NTSTATUS(-1070137306i32);
    pub const STATUS_VID_MB_PROPERTY_ALREADY_SET_RESET: NTSTATUS = NTSTATUS(-1070137312i32);
    pub const STATUS_VID_MB_STILL_REFERENCED: NTSTATUS = NTSTATUS(-1070137331i32);
    pub const STATUS_VID_MEMORY_BLOCK_LOCK_COUNT_EXCEEDED: NTSTATUS = NTSTATUS(-1070137321i32);
    pub const STATUS_VID_MEMORY_TYPE_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1070137298i32);
    pub const STATUS_VID_MESSAGE_QUEUE_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1070137333i32);
    pub const STATUS_VID_MESSAGE_QUEUE_CLOSED: NTSTATUS = NTSTATUS(-1070137318i32);
    pub const STATUS_VID_MESSAGE_QUEUE_NAME_TOO_LONG: NTSTATUS = NTSTATUS(-1070137337i32);
    pub const STATUS_VID_MMIO_RANGE_DESTROYED: NTSTATUS = NTSTATUS(-1070137311i32);
    pub const STATUS_VID_NOTIFICATION_QUEUE_ALREADY_ASSOCIATED: NTSTATUS = NTSTATUS(-1070137327i32);
    pub const STATUS_VID_NO_MEMORY_BLOCK_NOTIFICATION_QUEUE: NTSTATUS = NTSTATUS(-1070137322i32);
    pub const STATUS_VID_PAGE_RANGE_OVERFLOW: NTSTATUS = NTSTATUS(-1070137325i32);
    pub const STATUS_VID_PARTITION_ALREADY_EXISTS: NTSTATUS = NTSTATUS(-1070137336i32);
    pub const STATUS_VID_PARTITION_DOES_NOT_EXIST: NTSTATUS = NTSTATUS(-1070137335i32);
    pub const STATUS_VID_PARTITION_NAME_NOT_FOUND: NTSTATUS = NTSTATUS(-1070137334i32);
    pub const STATUS_VID_PARTITION_NAME_TOO_LONG: NTSTATUS = NTSTATUS(-1070137338i32);
    pub const STATUS_VID_PROCESS_ALREADY_SET: NTSTATUS = NTSTATUS(-1070137296i32);
    pub const STATUS_VID_QUEUE_FULL: NTSTATUS = NTSTATUS(-1070137341i32);
    pub const STATUS_VID_REMOTE_NODE_PARENT_GPA_PAGES_USED: NTSTATUS = NTSTATUS(-2143879167i32);
    pub const STATUS_VID_RESERVE_PAGE_SET_IS_BEING_USED: NTSTATUS = NTSTATUS(-1070137309i32);
    pub const STATUS_VID_RESERVE_PAGE_SET_TOO_SMALL: NTSTATUS = NTSTATUS(-1070137308i32);
    pub const STATUS_VID_SAVED_STATE_CORRUPT: NTSTATUS = NTSTATUS(-1070137305i32);
    pub const STATUS_VID_SAVED_STATE_INCOMPATIBLE: NTSTATUS = NTSTATUS(-1070137303i32);
    pub const STATUS_VID_SAVED_STATE_UNRECOGNIZED_ITEM: NTSTATUS = NTSTATUS(-1070137304i32);
    pub const STATUS_VID_STOP_PENDING: NTSTATUS = NTSTATUS(-1070137316i32);
    pub const STATUS_VID_TOO_MANY_HANDLERS: NTSTATUS = NTSTATUS(-1070137342i32);
    pub const STATUS_VID_VIRTUAL_PROCESSOR_LIMIT_EXCEEDED: NTSTATUS = NTSTATUS(-1070137317i32);
    pub const STATUS_VID_VTL_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1070137302i32);
    pub const STATUS_VIRTDISK_DISK_ALREADY_OWNED: NTSTATUS = NTSTATUS(-1069940706i32);
    pub const STATUS_VIRTDISK_DISK_ONLINE_AND_WRITABLE: NTSTATUS = NTSTATUS(-1069940705i32);
    pub const STATUS_VIRTDISK_NOT_VIRTUAL_DISK: NTSTATUS = NTSTATUS(-1069940715i32);
    pub const STATUS_VIRTDISK_PROVIDER_NOT_FOUND: NTSTATUS = NTSTATUS(-1069940716i32);
    pub const STATUS_VIRTDISK_UNSUPPORTED_DISK_SECTOR_SIZE: NTSTATUS = NTSTATUS(-1069940707i32);
    pub const STATUS_VIRTUAL_CIRCUIT_CLOSED: NTSTATUS = NTSTATUS(-1073741610i32);
    pub const STATUS_VIRTUAL_DISK_LIMITATION: NTSTATUS = NTSTATUS(-1069940710i32);
    pub const STATUS_VIRUS_DELETED: NTSTATUS = NTSTATUS(-1073739513i32);
    pub const STATUS_VIRUS_INFECTED: NTSTATUS = NTSTATUS(-1073739514i32);
    pub const STATUS_VOLMGR_ALL_DISKS_FAILED: NTSTATUS = NTSTATUS(-1070071767i32);
    pub const STATUS_VOLMGR_BAD_BOOT_DISK: NTSTATUS = NTSTATUS(-1070071729i32);
    pub const STATUS_VOLMGR_DATABASE_FULL: NTSTATUS = NTSTATUS(-1070071807i32);
    pub const STATUS_VOLMGR_DIFFERENT_SECTOR_SIZE: NTSTATUS = NTSTATUS(-1070071730i32);
    pub const STATUS_VOLMGR_DISK_CONFIGURATION_CORRUPTED: NTSTATUS = NTSTATUS(-1070071806i32);
    pub const STATUS_VOLMGR_DISK_CONFIGURATION_NOT_IN_SYNC: NTSTATUS = NTSTATUS(-1070071805i32);
    pub const STATUS_VOLMGR_DISK_CONTAINS_NON_SIMPLE_VOLUME: NTSTATUS = NTSTATUS(-1070071803i32);
    pub const STATUS_VOLMGR_DISK_DUPLICATE: NTSTATUS = NTSTATUS(-1070071802i32);
    pub const STATUS_VOLMGR_DISK_DYNAMIC: NTSTATUS = NTSTATUS(-1070071801i32);
    pub const STATUS_VOLMGR_DISK_ID_INVALID: NTSTATUS = NTSTATUS(-1070071800i32);
    pub const STATUS_VOLMGR_DISK_INVALID: NTSTATUS = NTSTATUS(-1070071799i32);
    pub const STATUS_VOLMGR_DISK_LAST_VOTER: NTSTATUS = NTSTATUS(-1070071798i32);
    pub const STATUS_VOLMGR_DISK_LAYOUT_INVALID: NTSTATUS = NTSTATUS(-1070071797i32);
    pub const STATUS_VOLMGR_DISK_LAYOUT_NON_BASIC_BETWEEN_BASIC_PARTITIONS: NTSTATUS = NTSTATUS(-1070071796i32);
    pub const STATUS_VOLMGR_DISK_LAYOUT_NOT_CYLINDER_ALIGNED: NTSTATUS = NTSTATUS(-1070071795i32);
    pub const STATUS_VOLMGR_DISK_LAYOUT_PARTITIONS_TOO_SMALL: NTSTATUS = NTSTATUS(-1070071794i32);
    pub const STATUS_VOLMGR_DISK_LAYOUT_PRIMARY_BETWEEN_LOGICAL_PARTITIONS: NTSTATUS = NTSTATUS(-1070071793i32);
    pub const STATUS_VOLMGR_DISK_LAYOUT_TOO_MANY_PARTITIONS: NTSTATUS = NTSTATUS(-1070071792i32);
    pub const STATUS_VOLMGR_DISK_MISSING: NTSTATUS = NTSTATUS(-1070071791i32);
    pub const STATUS_VOLMGR_DISK_NOT_EMPTY: NTSTATUS = NTSTATUS(-1070071790i32);
    pub const STATUS_VOLMGR_DISK_NOT_ENOUGH_SPACE: NTSTATUS = NTSTATUS(-1070071789i32);
    pub const STATUS_VOLMGR_DISK_REVECTORING_FAILED: NTSTATUS = NTSTATUS(-1070071788i32);
    pub const STATUS_VOLMGR_DISK_SECTOR_SIZE_INVALID: NTSTATUS = NTSTATUS(-1070071787i32);
    pub const STATUS_VOLMGR_DISK_SET_NOT_CONTAINED: NTSTATUS = NTSTATUS(-1070071786i32);
    pub const STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_MEMBERS: NTSTATUS = NTSTATUS(-1070071785i32);
    pub const STATUS_VOLMGR_DISK_USED_BY_MULTIPLE_PLEXES: NTSTATUS = NTSTATUS(-1070071784i32);
    pub const STATUS_VOLMGR_DYNAMIC_DISK_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1070071783i32);
    pub const STATUS_VOLMGR_EXTENT_ALREADY_USED: NTSTATUS = NTSTATUS(-1070071782i32);
    pub const STATUS_VOLMGR_EXTENT_NOT_CONTIGUOUS: NTSTATUS = NTSTATUS(-1070071781i32);
    pub const STATUS_VOLMGR_EXTENT_NOT_IN_PUBLIC_REGION: NTSTATUS = NTSTATUS(-1070071780i32);
    pub const STATUS_VOLMGR_EXTENT_NOT_SECTOR_ALIGNED: NTSTATUS = NTSTATUS(-1070071779i32);
    pub const STATUS_VOLMGR_EXTENT_OVERLAPS_EBR_PARTITION: NTSTATUS = NTSTATUS(-1070071778i32);
    pub const STATUS_VOLMGR_EXTENT_VOLUME_LENGTHS_DO_NOT_MATCH: NTSTATUS = NTSTATUS(-1070071777i32);
    pub const STATUS_VOLMGR_FAULT_TOLERANT_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1070071776i32);
    pub const STATUS_VOLMGR_INCOMPLETE_DISK_MIGRATION: NTSTATUS = NTSTATUS(-2143813630i32);
    pub const STATUS_VOLMGR_INCOMPLETE_REGENERATION: NTSTATUS = NTSTATUS(-2143813631i32);
    pub const STATUS_VOLMGR_INTERLEAVE_LENGTH_INVALID: NTSTATUS = NTSTATUS(-1070071775i32);
    pub const STATUS_VOLMGR_MAXIMUM_REGISTERED_USERS: NTSTATUS = NTSTATUS(-1070071774i32);
    pub const STATUS_VOLMGR_MEMBER_INDEX_DUPLICATE: NTSTATUS = NTSTATUS(-1070071772i32);
    pub const STATUS_VOLMGR_MEMBER_INDEX_INVALID: NTSTATUS = NTSTATUS(-1070071771i32);
    pub const STATUS_VOLMGR_MEMBER_IN_SYNC: NTSTATUS = NTSTATUS(-1070071773i32);
    pub const STATUS_VOLMGR_MEMBER_MISSING: NTSTATUS = NTSTATUS(-1070071770i32);
    pub const STATUS_VOLMGR_MEMBER_NOT_DETACHED: NTSTATUS = NTSTATUS(-1070071769i32);
    pub const STATUS_VOLMGR_MEMBER_REGENERATING: NTSTATUS = NTSTATUS(-1070071768i32);
    pub const STATUS_VOLMGR_MIRROR_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1070071717i32);
    pub const STATUS_VOLMGR_NOTIFICATION_RESET: NTSTATUS = NTSTATUS(-1070071764i32);
    pub const STATUS_VOLMGR_NOT_PRIMARY_PACK: NTSTATUS = NTSTATUS(-1070071726i32);
    pub const STATUS_VOLMGR_NO_REGISTERED_USERS: NTSTATUS = NTSTATUS(-1070071766i32);
    pub const STATUS_VOLMGR_NO_SUCH_USER: NTSTATUS = NTSTATUS(-1070071765i32);
    pub const STATUS_VOLMGR_NO_VALID_LOG_COPIES: NTSTATUS = NTSTATUS(-1070071720i32);
    pub const STATUS_VOLMGR_NUMBER_OF_DISKS_INVALID: NTSTATUS = NTSTATUS(-1070071718i32);
    pub const STATUS_VOLMGR_NUMBER_OF_DISKS_IN_MEMBER_INVALID: NTSTATUS = NTSTATUS(-1070071723i32);
    pub const STATUS_VOLMGR_NUMBER_OF_DISKS_IN_PLEX_INVALID: NTSTATUS = NTSTATUS(-1070071724i32);
    pub const STATUS_VOLMGR_NUMBER_OF_EXTENTS_INVALID: NTSTATUS = NTSTATUS(-1070071731i32);
    pub const STATUS_VOLMGR_NUMBER_OF_MEMBERS_INVALID: NTSTATUS = NTSTATUS(-1070071763i32);
    pub const STATUS_VOLMGR_NUMBER_OF_PLEXES_INVALID: NTSTATUS = NTSTATUS(-1070071762i32);
    pub const STATUS_VOLMGR_PACK_CONFIG_OFFLINE: NTSTATUS = NTSTATUS(-1070071728i32);
    pub const STATUS_VOLMGR_PACK_CONFIG_ONLINE: NTSTATUS = NTSTATUS(-1070071727i32);
    pub const STATUS_VOLMGR_PACK_CONFIG_UPDATE_FAILED: NTSTATUS = NTSTATUS(-1070071804i32);
    pub const STATUS_VOLMGR_PACK_DUPLICATE: NTSTATUS = NTSTATUS(-1070071761i32);
    pub const STATUS_VOLMGR_PACK_HAS_QUORUM: NTSTATUS = NTSTATUS(-1070071756i32);
    pub const STATUS_VOLMGR_PACK_ID_INVALID: NTSTATUS = NTSTATUS(-1070071760i32);
    pub const STATUS_VOLMGR_PACK_INVALID: NTSTATUS = NTSTATUS(-1070071759i32);
    pub const STATUS_VOLMGR_PACK_LOG_UPDATE_FAILED: NTSTATUS = NTSTATUS(-1070071725i32);
    pub const STATUS_VOLMGR_PACK_NAME_INVALID: NTSTATUS = NTSTATUS(-1070071758i32);
    pub const STATUS_VOLMGR_PACK_OFFLINE: NTSTATUS = NTSTATUS(-1070071757i32);
    pub const STATUS_VOLMGR_PACK_WITHOUT_QUORUM: NTSTATUS = NTSTATUS(-1070071755i32);
    pub const STATUS_VOLMGR_PARTITION_STYLE_INVALID: NTSTATUS = NTSTATUS(-1070071754i32);
    pub const STATUS_VOLMGR_PARTITION_UPDATE_FAILED: NTSTATUS = NTSTATUS(-1070071753i32);
    pub const STATUS_VOLMGR_PLEX_INDEX_DUPLICATE: NTSTATUS = NTSTATUS(-1070071751i32);
    pub const STATUS_VOLMGR_PLEX_INDEX_INVALID: NTSTATUS = NTSTATUS(-1070071750i32);
    pub const STATUS_VOLMGR_PLEX_IN_SYNC: NTSTATUS = NTSTATUS(-1070071752i32);
    pub const STATUS_VOLMGR_PLEX_LAST_ACTIVE: NTSTATUS = NTSTATUS(-1070071749i32);
    pub const STATUS_VOLMGR_PLEX_MISSING: NTSTATUS = NTSTATUS(-1070071748i32);
    pub const STATUS_VOLMGR_PLEX_NOT_RAID5: NTSTATUS = NTSTATUS(-1070071745i32);
    pub const STATUS_VOLMGR_PLEX_NOT_SIMPLE: NTSTATUS = NTSTATUS(-1070071744i32);
    pub const STATUS_VOLMGR_PLEX_NOT_SIMPLE_SPANNED: NTSTATUS = NTSTATUS(-1070071721i32);
    pub const STATUS_VOLMGR_PLEX_REGENERATING: NTSTATUS = NTSTATUS(-1070071747i32);
    pub const STATUS_VOLMGR_PLEX_TYPE_INVALID: NTSTATUS = NTSTATUS(-1070071746i32);
    pub const STATUS_VOLMGR_PRIMARY_PACK_PRESENT: NTSTATUS = NTSTATUS(-1070071719i32);
    pub const STATUS_VOLMGR_RAID5_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1070071716i32);
    pub const STATUS_VOLMGR_STRUCTURE_SIZE_INVALID: NTSTATUS = NTSTATUS(-1070071743i32);
    pub const STATUS_VOLMGR_TOO_MANY_NOTIFICATION_REQUESTS: NTSTATUS = NTSTATUS(-1070071742i32);
    pub const STATUS_VOLMGR_TRANSACTION_IN_PROGRESS: NTSTATUS = NTSTATUS(-1070071741i32);
    pub const STATUS_VOLMGR_UNEXPECTED_DISK_LAYOUT_CHANGE: NTSTATUS = NTSTATUS(-1070071740i32);
    pub const STATUS_VOLMGR_VOLUME_CONTAINS_MISSING_DISK: NTSTATUS = NTSTATUS(-1070071739i32);
    pub const STATUS_VOLMGR_VOLUME_ID_INVALID: NTSTATUS = NTSTATUS(-1070071738i32);
    pub const STATUS_VOLMGR_VOLUME_LENGTH_INVALID: NTSTATUS = NTSTATUS(-1070071737i32);
    pub const STATUS_VOLMGR_VOLUME_LENGTH_NOT_SECTOR_SIZE_MULTIPLE: NTSTATUS = NTSTATUS(-1070071736i32);
    pub const STATUS_VOLMGR_VOLUME_MIRRORED: NTSTATUS = NTSTATUS(-1070071722i32);
    pub const STATUS_VOLMGR_VOLUME_NOT_MIRRORED: NTSTATUS = NTSTATUS(-1070071735i32);
    pub const STATUS_VOLMGR_VOLUME_NOT_RETAINED: NTSTATUS = NTSTATUS(-1070071734i32);
    pub const STATUS_VOLMGR_VOLUME_OFFLINE: NTSTATUS = NTSTATUS(-1070071733i32);
    pub const STATUS_VOLMGR_VOLUME_RETAINED: NTSTATUS = NTSTATUS(-1070071732i32);
    pub const STATUS_VOLSNAP_ACTIVATION_TIMEOUT: NTSTATUS = NTSTATUS(-1068498940i32);
    pub const STATUS_VOLSNAP_BOOTFILE_NOT_VALID: NTSTATUS = NTSTATUS(-1068498941i32);
    pub const STATUS_VOLSNAP_HIBERNATE_READY: NTSTATUS = NTSTATUS(293i32);
    pub const STATUS_VOLSNAP_NO_BYPASSIO_WITH_SNAPSHOT: NTSTATUS = NTSTATUS(-1068498939i32);
    pub const STATUS_VOLSNAP_PREPARE_HIBERNATE: NTSTATUS = NTSTATUS(-1073740793i32);
    pub const STATUS_VOLUME_DIRTY: NTSTATUS = NTSTATUS(-1073739770i32);
    pub const STATUS_VOLUME_DISMOUNTED: NTSTATUS = NTSTATUS(-1073741202i32);
    pub const STATUS_VOLUME_MOUNTED: NTSTATUS = NTSTATUS(265i32);
    pub const STATUS_VOLUME_NOT_CLUSTER_ALIGNED: NTSTATUS = NTSTATUS(-1073740636i32);
    pub const STATUS_VOLUME_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073740602i32);
    pub const STATUS_VOLUME_NOT_UPGRADED: NTSTATUS = NTSTATUS(-1073741156i32);
    pub const STATUS_VOLUME_WRITE_ACCESS_DENIED: NTSTATUS = NTSTATUS(-1073740589i32);
    pub const STATUS_VRF_VOLATILE_CFG_AND_IO_ENABLED: NTSTATUS = NTSTATUS(-1073738744i32);
    pub const STATUS_VRF_VOLATILE_NMI_REGISTERED: NTSTATUS = NTSTATUS(-1073738738i32);
    pub const STATUS_VRF_VOLATILE_NOT_RUNNABLE_SYSTEM: NTSTATUS = NTSTATUS(-1073738741i32);
    pub const STATUS_VRF_VOLATILE_NOT_STOPPABLE: NTSTATUS = NTSTATUS(-1073738743i32);
    pub const STATUS_VRF_VOLATILE_NOT_SUPPORTED_RULECLASS: NTSTATUS = NTSTATUS(-1073738740i32);
    pub const STATUS_VRF_VOLATILE_PROTECTED_DRIVER: NTSTATUS = NTSTATUS(-1073738739i32);
    pub const STATUS_VRF_VOLATILE_SAFE_MODE: NTSTATUS = NTSTATUS(-1073738742i32);
    pub const STATUS_VRF_VOLATILE_SETTINGS_CONFLICT: NTSTATUS = NTSTATUS(-1073738737i32);
    pub const STATUS_VSM_DMA_PROTECTION_NOT_IN_USE: NTSTATUS = NTSTATUS(-1069219839i32);
    pub const STATUS_VSM_NOT_INITIALIZED: NTSTATUS = NTSTATUS(-1069219840i32);
    pub const STATUS_WAIT_0: NTSTATUS = NTSTATUS(0i32);
    pub const STATUS_WAIT_1: NTSTATUS = NTSTATUS(1i32);
    pub const STATUS_WAIT_2: NTSTATUS = NTSTATUS(2i32);
    pub const STATUS_WAIT_3: NTSTATUS = NTSTATUS(3i32);
    pub const STATUS_WAIT_63: NTSTATUS = NTSTATUS(63i32);
    pub const STATUS_WAIT_FOR_OPLOCK: NTSTATUS = NTSTATUS(871i32);
    pub const STATUS_WAKE_SYSTEM: NTSTATUS = NTSTATUS(1073742484i32);
    pub const STATUS_WAKE_SYSTEM_DEBUGGER: NTSTATUS = NTSTATUS(-2147483641i32);
    pub const STATUS_WAS_LOCKED: NTSTATUS = NTSTATUS(1073741849i32);
    pub const STATUS_WAS_UNLOCKED: NTSTATUS = NTSTATUS(1073741847i32);
    pub const STATUS_WEAK_WHFBKEY_BLOCKED: NTSTATUS = NTSTATUS(-1073741389i32);
    pub const STATUS_WIM_NOT_BOOTABLE: NTSTATUS = NTSTATUS(-1073740665i32);
    pub const STATUS_WMI_ALREADY_DISABLED: NTSTATUS = NTSTATUS(-1073741054i32);
    pub const STATUS_WMI_ALREADY_ENABLED: NTSTATUS = NTSTATUS(-1073741053i32);
    pub const STATUS_WMI_GUID_DISCONNECTED: NTSTATUS = NTSTATUS(-1073741055i32);
    pub const STATUS_WMI_GUID_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741163i32);
    pub const STATUS_WMI_INSTANCE_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741162i32);
    pub const STATUS_WMI_ITEMID_NOT_FOUND: NTSTATUS = NTSTATUS(-1073741161i32);
    pub const STATUS_WMI_NOT_SUPPORTED: NTSTATUS = NTSTATUS(-1073741091i32);
    pub const STATUS_WMI_READ_ONLY: NTSTATUS = NTSTATUS(-1073741114i32);
    pub const STATUS_WMI_SET_FAILURE: NTSTATUS = NTSTATUS(-1073741113i32);
    pub const STATUS_WMI_TRY_AGAIN: NTSTATUS = NTSTATUS(-1073741160i32);
    pub const STATUS_WOF_FILE_RESOURCE_TABLE_CORRUPT: NTSTATUS = NTSTATUS(-1073700185i32);
    pub const STATUS_WOF_WIM_HEADER_CORRUPT: NTSTATUS = NTSTATUS(-1073700187i32);
    pub const STATUS_WOF_WIM_RESOURCE_TABLE_CORRUPT: NTSTATUS = NTSTATUS(-1073700186i32);
    pub const STATUS_WORKING_SET_LIMIT_RANGE: NTSTATUS = NTSTATUS(1073741826i32);
    pub const STATUS_WORKING_SET_QUOTA: NTSTATUS = NTSTATUS(-1073741663i32);
    pub const STATUS_WOW_ASSERTION: NTSTATUS = NTSTATUS(-1073702760i32);
    pub const STATUS_WRONG_COMPARTMENT: NTSTATUS = NTSTATUS(-1073700731i32);
    pub const STATUS_WRONG_CREDENTIAL_HANDLE: NTSTATUS = NTSTATUS(-1073741070i32);
    pub const STATUS_WRONG_EFS: NTSTATUS = NTSTATUS(-1073741169i32);
    pub const STATUS_WRONG_PASSWORD_CORE: NTSTATUS = NTSTATUS(-1073741495i32);
    pub const STATUS_WRONG_VOLUME: NTSTATUS = NTSTATUS(-1073741806i32);
    pub const STATUS_WX86_BREAKPOINT: NTSTATUS = NTSTATUS(1073741855i32);
    pub const STATUS_WX86_CONTINUE: NTSTATUS = NTSTATUS(1073741853i32);
    pub const STATUS_WX86_CREATEWX86TIB: NTSTATUS = NTSTATUS(1073741864i32);
    pub const STATUS_WX86_EXCEPTION_CHAIN: NTSTATUS = NTSTATUS(1073741858i32);
    pub const STATUS_WX86_EXCEPTION_CONTINUE: NTSTATUS = NTSTATUS(1073741856i32);
    pub const STATUS_WX86_EXCEPTION_LASTCHANCE: NTSTATUS = NTSTATUS(1073741857i32);
    pub const STATUS_WX86_FLOAT_STACK_CHECK: NTSTATUS = NTSTATUS(-1073741200i32);
    pub const STATUS_WX86_INTERNAL_ERROR: NTSTATUS = NTSTATUS(-1073741201i32);
    pub const STATUS_WX86_SINGLE_STEP: NTSTATUS = NTSTATUS(1073741854i32);
    pub const STATUS_WX86_UNSIMULATE: NTSTATUS = NTSTATUS(1073741852i32);
    pub const STATUS_XMLDSIG_ERROR: NTSTATUS = NTSTATUS(-1073700732i32);
    pub const STATUS_XML_ENCODING_MISMATCH: NTSTATUS = NTSTATUS(-1072365535i32);
    pub const STATUS_XML_PARSE_ERROR: NTSTATUS = NTSTATUS(-1073700733i32);
}
